Package: declared
Title: Functions for Declared Missing Values
Version: 0.26
Authors@R: 
    c(
        person(
            "Adrian", "Dusa",
            email = "dusa.adrian@unibuc.ro",
            role = c("aut", "cre", "cph"),
            comment = c(ORCID = "0000-0002-3525-9253")
        )
    )
URL: https://github.com/dusadrian/declared
BugReports: https://github.com/dusadrian/declared/issues
Description: A zero dependency package containing functions to declare labels
    and missing values, coupled with associated functions to create (weighted)
    tables of frequencies and various other summary measures. 
    Some of the base functions have been rewritten to make use of the specific 
    information about the missing values, most importantly to distinguish
    between empty NA and declared NA values.
    Some functions have similar functionality with the corresponding ones from
    packages "haven" and "labelled". The aim is to ensure as much compatibility
    as possible with these packages, while offering an alternative in the
    objects of class "declared".
License: GPL (>= 3)
Language: en-US
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-04-01 16:01:14 UTC; dusadrian
Author: Adrian Dusa [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-3525-9253>)
Maintainer: Adrian Dusa <dusa.adrian@unibuc.ro>
Repository: CRAN
Date/Publication: 2026-04-02 10:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 13:32:21 UTC; unix
Archs: declared.so.dSYM

Package: declared
Title: Functions for Declared Missing Values
Version: 0.27
Authors@R: 
    c(
        person(
            "Adrian", "Dusa",
            email = "dusa.adrian@unibuc.ro",
            role = c("aut", "cre", "cph"),
            comment = c(ORCID = "0000-0002-3525-9253")
        )
    )
URL: https://dusadrian.github.io/declared/,
        https://github.com/dusadrian/declared
BugReports: https://github.com/dusadrian/declared/issues
Description: A zero dependency package containing functions to declare labels
    and missing values, coupled with associated functions to create (weighted)
    tables of frequencies and various other summary measures.
    Some of the base functions have been rewritten to make use of the specific
    information about the missing values, most importantly to distinguish
    between empty and declared NA values.
    Some functions have similar functionality with the corresponding ones from
    packages "haven" and "labelled". The aim is to ensure as much compatibility
    as possible with these packages, while offering an alternative in the
    objects of class "declared".
License: GPL (>= 3)
Imports: stats, utils
Language: en-US
Encoding: UTF-8
NeedsCompilation: yes
Packaged: 2026-08-22 02:49:27 UTC; dusadrian
Author: Adrian Dusa [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-3525-9253>)
Maintainer: Adrian Dusa <dusa.adrian@unibuc.ro>
Repository: CRAN
Date/Publication: 2026-08-22 03:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("!",declared)
S3method("!=",declared)
S3method("%%",declared)
S3method("%/%",declared)
S3method("&",declared)
S3method("*",declared)
S3method("+",declared)
S3method("-",declared)
S3method("/",declared)
S3method("<",declared)
S3method("<=",declared)
S3method("==",declared)
S3method(">",declared)
S3method(">=",declared)
S3method("[",declared)
S3method("[",labels_df)
S3method("[<-",declared)
S3method("^",declared)
S3method("label<-",data.frame)
S3method("label<-",declared)
S3method("label<-",default)
S3method("label<-",haven_labelled)
S3method("labels<-",declared)
S3method("labels<-",default)
S3method("labels<-",haven_labelled)
S3method("measurement<-",declared)
S3method("measurement<-",default)
S3method("missing_range<-",declared)
S3method("missing_range<-",default)
S3method("missing_values<-",declared)
S3method("missing_values<-",default)
S3method("names<-",declared)
S3method("|",declared)
S3method(Arg,declared)
S3method(Conj,declared)
S3method(Im,declared)
S3method(Mod,declared)
S3method(Re,declared)
S3method(Summary,declared)
S3method(abs,declared)
S3method(acos,declared)
S3method(acosh,declared)
S3method(all.equal,declared)
S3method(anyDuplicated,declared)
S3method(as.Date,declared)
S3method(as.character,declared)
S3method(as.declared,data.frame)
S3method(as.declared,default)
S3method(as.declared,factor)
S3method(as.declared,haven_labelled)
S3method(as.haven,data.frame)
S3method(as.haven,declared)
S3method(as.haven,default)
S3method(as.haven,haven_labelled)
S3method(as.haven,haven_labelled_spss)
S3method(asin,declared)
S3method(asinh,declared)
S3method(atan,declared)
S3method(atanh,declared)
S3method(c,declared)
S3method(ceiling,declared)
S3method(cos,declared)
S3method(cosh,declared)
S3method(cospi,declared)
S3method(cummax,declared)
S3method(cummin,declared)
S3method(cumprod,declared)
S3method(cumsum,declared)
S3method(declared,default)
S3method(digamma,declared)
S3method(drop_na,data.frame)
S3method(drop_na,declared)
S3method(drop_na,default)
S3method(drop_na,haven_labelled_spss)
S3method(duplicated,declared)
S3method(exp,declared)
S3method(expm1,declared)
S3method(floor,declared)
S3method(format,declared)
S3method(format,pillar_shaft_declared_chr)
S3method(format,pillar_shaft_declared_num)
S3method(gamma,declared)
S3method(head,declared)
S3method(head,labels_df)
S3method(label,data.frame)
S3method(label,declared)
S3method(label,default)
S3method(label,haven_labelled)
S3method(labels,data.frame)
S3method(labels,declared)
S3method(labels,haven_labelled)
S3method(lgamma,declared)
S3method(log,declared)
S3method(log1p,declared)
S3method(mean,declared)
S3method(measurement,declared)
S3method(measurement,default)
S3method(median,declared)
S3method(missing_range,data.frame)
S3method(missing_range,declared)
S3method(missing_range,default)
S3method(missing_range,haven_labelled_spss)
S3method(missing_values,data.frame)
S3method(missing_values,declared)
S3method(missing_values,default)
S3method(missing_values,haven_labelled_spss)
S3method(na.exclude,declared)
S3method(na.fail,declared)
S3method(na.omit,declared)
S3method(print,declared)
S3method(print,fobject)
S3method(print,labels_df)
S3method(print,wmeasures)
S3method(print,wtable)
S3method(rep,declared)
S3method(round,declared)
S3method(sign,declared)
S3method(signif,declared)
S3method(sin,declared)
S3method(sinh,declared)
S3method(sinpi,declared)
S3method(sort,declared)
S3method(sqrt,declared)
S3method(summary,declared)
S3method(tail,declared)
S3method(tail,labels_df)
S3method(tan,declared)
S3method(tanh,declared)
S3method(tanpi,declared)
S3method(trigamma,declared)
S3method(trunc,declared)
S3method(undeclare,data.frame)
S3method(undeclare,declared)
S3method(undeclare,default)
S3method(unique,declared)
S3method(weighted.mean,declared)
S3method(xtfrm,declared)
export("label<-")
export("labels<-")
export("measurement<-")
export("missing_range<-")
export("missing_values<-")
export(anyNAdeclared)
export(anyNAempty)
export(anyTagged_)
export(as.declared)
export(as.haven)
export(declared)
export(direct_declared)
export(drop_na)
export(format_declared)
export(getTag_)
export(hasTag_)
export(is.declared)
export(is.empty)
export(label)
export(makeTag_)
export(measurement)
export(missing_range)
export(missing_values)
export(names_values)
export(order_declared)
export(undeclare)
export(valid_na_index)
export(value_labels)
export(variable_label)
export(wIQR)
export(w_IQR)
export(w_fivenum)
export(w_mean)
export(w_median)
export(w_mode)
export(w_quantile)
export(w_sd)
export(w_standardize)
export(w_summary)
export(w_table)
export(w_var)
export(wfivenum)
export(wmean)
export(wmeasures)
export(wmedian)
export(wmode)
export(wquantile)
export(wsd)
export(wstandardize)
export(wsummary)
export(wtable)
export(wvar)
importFrom(stats,median)
importFrom(stats,na.exclude)
importFrom(stats,na.fail)
importFrom(stats,na.omit)
importFrom(stats,weighted.mean)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(declared, .registration = TRUE)

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Dynamically exported, see onLoad.R
# using eval(parse()) to avoid the huge dependency tree of vctrs, haven,
# labelled and pillar; these functions will be registered when or if the
# package vctrs is loaded

`vec_ptype_abbr.declared` <- function (x, ...) {
    command <- "vctrs::vec_ptype_abbr(vctrs::vec_data(unclass (undeclare (x))))"
    return (
        paste0 (eval (parse (text = command)), "+lbl")
    )
}

`vec_ptype_full.declared` <- function (x, ...) {
    command <- "vctrs::vec_ptype_full(vctrs::vec_data(unclass (undeclare (x))))"
    return (
        paste0 (
            "declared<",
            eval (parse (text = command)),
            ">"
        )
    )
}

`vec_proxy.declared` <- function (x, ...) {
    return (undeclare (x, drop = TRUE))
}

`vec_restore.declared` <- function(x, to, ...) {
    to <- sanitize_na_index_ (to)
    attrs <- attributes(to)
    todate <- isTRUE(attrs$date)

    misvals <- all_missing_values (
        x,
        attrs$na_values,
        attrs$na_range,
        attrs$labels
    )

    na_index <- which (is.element (x, misvals))

    if (length(na_index) > 0) {
        declared_nas <- x[na_index]

        if (todate) {
            declared_nas <- as.numeric (declared_nas)
        }

        x[na_index] <- NA
        names (na_index) <- declared_nas
    } else {
        na_index <- NULL
    }

    attrs$na_index <- na_index
    if (possibleNumeric_ (x)) {
        x <- as.numeric (x)
        if (inherits (to, "integer")) {
            x <- as.integer (x)
        }
    }
    attributes(x) <- attrs
    return (x)
}

# `vec_ptype2.declared` <- function (x, y, ...) {
#     command <- paste (
#         "vctrs::vec_ptype2(unclass (undeclare (x)),",
#         "vctrs::vec_data(unclass (undeclare (y))), ...)"
#     )
#     eval (parse (text = command))
# }

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

`.onLoad` <- function (...) {
    loc <- dirname(getNamespaceInfo("stats", "path"))
    suppressPackageStartupMessages(
        do.call (
            "library",
            list (
                "stats",
                lib.loc = loc,
                character.only = TRUE,
                warn.conflicts = FALSE
            )
        )
    )

    loc <- dirname(getNamespaceInfo("graphics", "path"))
    suppressPackageStartupMessages(
        do.call (
            "library",
            list (
                "graphics",
                lib.loc = loc,
                character.only = TRUE,
                warn.conflicts = FALSE
            )
        )
    )

    # if (unlockEnvironment_ (asNamespace("base"))) {

        env <- as.environment("package:base")

        do.call ("unlockBinding", list (sym = "rbind.data.frame", env = env))

        env$`rbind.data.frame` <- function (
            ...,
            deparse.level = 1,
            make.row.names = TRUE,
            stringsAsFactors = FALSE,
            factor.exclude = TRUE
        ) {
            match.names <- function (clabs, nmi) {
                if (identical (clabs, nmi)) NULL
                else if (
                    length (nmi) == length (clabs) && all (nmi %in% clabs)
                ) {
                        ## we need 1-1 matches here
                    m <- pmatch(nmi, clabs, 0L)
                        if (any (m == 0L))
                            stop ("names do not match previous names")
                        m
                } else stop ("names do not match previous names")
            }

            allargs <- list (...)
            allargs <- allargs[lengths(allargs) > 0L]
            if (length (allargs)) {
                ## drop any zero-row data frames, as they may not have proper
                ## column types (e.g. NULL).
                nr <- vapply (allargs, function (x)
                            if (is.data.frame(x)) .row_names_info(x, 2L)
                            else if (is.list (x)) length (x[[1L]])
                            # mismatched lists are checked later
                            else length (x), 1L)
            if (any (n0 <- nr == 0L)) {
                if (all (n0)) return (allargs[[1L]]) # pretty arbitrary
                allargs <- allargs[!n0]
            }
            }
            n <- length (allargs)
            if (n == 0L)
            return (list2DF())
            nms <- names (allargs)
            if (is.null (nms))
            nms <- character (n)
            cl <- NULL
            perm <- rows <- vector("list", n)
            if (make.row.names) {
                rlabs <- rows
                env <- new.env()
                # result with 1:nrow (.) row names? [efficiency!]
                env$autoRnms <- TRUE
                Make.row.names <- function (nmi, ri, ni, nrow) {
                    if (nzchar(nmi)) {
                    if (env$autoRnms) env$autoRnms <- FALSE
                    if (ni == 0L) character ()  # PR#8506
                    else if (ni > 1L) paste (nmi, ri, sep = ".")
                    else nmi
                    }
                    else if (
                        env$autoRnms && nrow > 0L && identical (ri, seq_len(ni))
                    )
                    as.integer (seq.int(from = nrow + 1L, length.out = ni))
                    else {
                        if (
                            env$autoRnms && (nrow > 0L ||
                            !identical (ri, seq_len(ni)))
                        ) {
                            env$autoRnms <- FALSE
                        }
                        ri
                    }
                }
            }
            smartX <- isTRUE (factor.exclude)

            ## check the arguments, develop row and column labels
            nrow <- 0L
            value <- clabs <- NULL
            all.levs <- list ()
            for(i in seq_len(n)) { ## check and treat arg [[ i ]]  -- part 1
                xi <- allargs[[i]]
                nmi <- nms[i]
                ## coerce matrix to data frame
                if (is.matrix(xi)) allargs[[i]] <- xi <-
                    as.data.frame(xi, stringsAsFactors = stringsAsFactors)
                if (inherits (xi, "data.frame")) {
                    declared_columns <- vapply (xi, is.declared, TRUE)
                    if (any (declared_columns)) {
                        xi[declared_columns] <- lapply (
                            xi[declared_columns], sanitize_na_index_
                        )
                        allargs[[i]] <- xi
                    }
                    if (is.null (cl))
                    cl <- oldClass (xi)
                    ri <- attr (xi, "row.names")
                    ni <- length (ri)
                    if (is.null (clabs)) ## first time
                    clabs <- names (xi)
                    else {
                            if (length (xi) != length (clabs))
                                stop (
                                  "numbers of columns of arguments do not match"
                                )
                        pi <- match.names (clabs, names (xi))
                        if ( !is.null (pi) ) perm[[i]] <- pi
                    }
                    rows[[i]] <- seq.int(from = nrow + 1L, length.out = ni)
                    if (make.row.names) rlabs[[i]] <- Make.row.names (
                        nmi, ri, ni, nrow
                    )
                    nrow <- nrow + ni
                    if (is.null (value)) { ## first time ==> setup once:
                        value <- unclass (xi)
                        nvar <- length (value)

                        ## code for declared
                        lxi <- vector("list", nvar)
                        xideclared <- vapply (xi, is.declared, TRUE)
                        names (lxi) <- names (xideclared)
                        lxi[xideclared] <- lapply(xi[xideclared], function (x) {
                            list (
                                labels = attr (x, "labels", exact = TRUE),
                                na_values = attr (x, "na_values"),
                                na_range = attr (x, "na_range"),
                                decimals = attr (x, "decimals", exact = TRUE)
                            )
                        })
                        ## code for declared

                        all.levs <- vector("list", nvar)
                        has.dim <- facCol <- ordCol <- logical (nvar)
                        if (smartX) NA.lev <- ordCol
                        for(j in seq_len(nvar)) {
                        xj <- value[[j]]
                                facCol[j] <- fac <-
                                    if (!is.null (lj <- levels (xj))) {
                                        all.levs[[j]] <- lj
                                        TRUE # turn categories into factors
                                    } else
                                        is.factor (xj)
                        if (fac) {
                        ordCol[j] <- is.ordered(xj)
                        if (smartX && !NA.lev[j])
                            NA.lev[j] <- anyNA(lj)
                        }
                        has.dim[j] <- length (dim(xj)) == 2L
                        }
                    }
                    else for(j in seq_len(nvar)) {
                        xij <- xi[[j]]
                        if (is.null (pi) || is.na (jj <- pi[[j]])) jj <- j
                        if (facCol[jj]) {
                            if (length (lij <- levels (xij))) {
                                all.levs[[jj]] <- unique(c(all.levs[[jj]], lij))
                                if (ordCol[jj])
                                    ordCol[jj] <- is.ordered(xij)
                                if (smartX && !NA.lev[jj])
                                    NA.lev[jj] <- anyNA(lij)
                            } else if (is.character (xij))
                                all.levs[[jj]] <- unique(c(all.levs[[jj]], xij))
                        }

                        ## code for declared
                        if (xideclared[jj]) {
                            labels <- c (
                                lxi[[jj]]$labels,
                                attr (xij, "labels", exact = TRUE)
                            )
                            labels <- labels[!duplicated (labels)]

                            lxi[[jj]]$labels <- sort (labels)

                            lxi[[jj]]$na_values <- sort (
                                unique (
                                    c (
                                        lxi[[jj]]$na_values,
                                        attr (xij, "na_values")
                                    )
                                )
                            )

                            na_range <- sort (
                                unique (
                                    c (
                                        lxi[[jj]]$na_range,
                                        attr (xij, "na_range")
                                    )
                                )
                            )

                            if (!is.null (na_range)) {
                                lxi[[jj]]$na_range <- range(na_range)
                            }

                            decimals <- unique (unlist (c (
                                lxi[[jj]]$decimals,
                                attr (xij, "decimals", exact = TRUE)
                            )))
                            if (length (decimals) > 0) {
                                lxi[[jj]]$decimals <- max (decimals)
                            }
                        }
                        ## code for declared
                    }
                } ## end{data.frame}
                else if (is.list (xi)) {
                    ni <- range(lengths(xi))
                    if (ni[1L] == ni[2L])
                    ni <- ni[1L]
                    else stop (
                        paste (
                            "invalid list argument: all variables",
                            "should have the same length"
                        )
                    )
                        ri <- seq_len(ni)
                    rows[[i]] <- seq.int(from = nrow + 1L, length.out = ni)
                    if (make.row.names) rlabs[[i]] <- Make.row.names (
                        nmi, ri, ni, nrow
                    )
                    nrow <- nrow + ni
                    if (length (nmi <- names (xi)) > 0L) {
                        if (is.null (clabs))
                            clabs <- nmi
                        else {
                            if (length (xi) != length (clabs)) {
                                stop (
                                  "numbers of columns of arguments do not match"
                                )
                            }
                            pi <- match.names (clabs, nmi)
                            if ( !is.null (pi) ) perm[[i]] <- pi
                        }
                    }
                }
                else if (length (xi)) { # 1 new row
                    rows[[i]] <- nrow <- nrow + 1L
                        if (make.row.names)
                    rlabs[[i]] <- if (nzchar(nmi)) nmi else as.integer (nrow)
                }
            } # for(i .)

            nvar <- length (clabs)
            if (nvar == 0L)
            nvar <- max (lengths(allargs)) # only vector args
            if (nvar == 0L)
            return (list2DF())
            pseq <- seq_len(nvar)
            if (is.null(value)) { # this happens if there has been no data frame
            value <- list ()
            value[pseq] <- list(logical(nrow)) # OK for coercion except to raw.
                all.levs <- vector("list", nvar)
            has.dim <- facCol <- ordCol <- logical (nvar)
            if (smartX) NA.lev <- ordCol
            }
            names (value) <- clabs
            for(j in pseq)
            if (length (lij <- all.levs[[j]]))
                    value[[j]] <-
                factor (as.vector (value[[j]]), levels = lij,
                    exclude = if (smartX) {
                            if (!NA.lev[j]) NA # else NULL
                        } else factor.exclude,
                    ordered = ordCol[j])

            if (any (has.dim)) { # some col's are matrices or d.frame's
                jdim <- pseq[has.dim]
                if (!all (df <- vapply (
                    jdim, function (j) inherits (value[[j]],"data.frame"), NA
                ))) {
                    ## Ensure matrix columns can be filled in  for(i ...) below
                    rmax <- max (unlist (rows))
                    for(j in jdim[!df]) {
                dn <- dimnames (vj <- value[[j]])
                rn <- dn[[1L]]
                if (length (rn) > 0L) length (rn) <- rmax
                pj <- dim(vj)[2L]
                length (vj) <- rmax * pj
                value[[j]] <- array(vj, c (rmax, pj), list (rn, dn[[2L]]))
                }
                }
            }

            for(i in seq_len(n)) { ## add arg [[i]] to result  (part 2)
            xi <- unclass (allargs[[i]])
            if (!is.list (xi))
                if ((ni <- length (xi)) != nvar) {
                if (ni && nvar %% ni != 0)
                    warning (
                        gettextf(
                            paste (
                                "number of columns of result, %d, is not a",
                                "multiple of vector length %d of arg %d"
                            ),
                            nvar, ni, i),
                        domain = NA
                    )
                xi <- rep_len(xi, nvar)
                    }
            ri <- rows[[i]]
            pi <- perm[[i]]
            if (is.null (pi)) pi <- pseq
            for(j in pseq) {
                jj <- pi[j]
                    xij <- xi[[j]]
                if (has.dim[jj]) {
                value[[jj]][ri,	 ] <- xij
                        ## copy rownames
                        if (!is.null (r <- rownames (xij)) &&
                        !(inherits (xij, "data.frame") &&
                            .row_names_info(xij) <= 0))
                            rownames (value[[jj]])[ri] <- r
                } else {
                        ## coerce factors to vectors, in case lhs is character
                        ## or level set has changed
                        value[[jj]][ri] <- if (is.factor (xij)) {
                            as.vector (xij)
                        } else {
                            xij
                        }
                        ## copy names if any
                        if (!is.null (nm <- names (xij))) {
                            names (value[[jj]])[ri] <- nm
                        }
                    }
            }
            }
            rlabs <- if (make.row.names && !env$autoRnms) {
                rlabs <- unlist (rlabs)
                if (anyDuplicated (rlabs))
                    make.unique (as.character (rlabs), sep = "")
                else
                    rlabs
                } # else NULL

            ## code for declared
            if (any (xideclared)) {
                for (i in which (xideclared)) {
                    value[[i]] <- declared (
                        undeclare (value[[i]], drop = TRUE),
                        label = attr (value[[i]], "label", exact = TRUE),
                        labels = lxi[[i]]$labels,
                        na_values = lxi[[i]]$na_values,
                        na_range = lxi[[i]]$na_range,
                        decimals = lxi[[i]]$decimals
                    )
                }
            }
            ## code for declared

            if (is.null (cl)) {
            as.data.frame(value, row.names = rlabs, fix.empty.names = TRUE,
                    stringsAsFactors = stringsAsFactors)
            } else {
            structure (value, class = cl,
                row.names = if (is.null (rlabs)) {
                        .set_row_names (nrow)
                    } else { rlabs }
                )
            }
        }

        do.call ("unlockBinding", list (sym = "order", env = env))

        env$order <- function (..., na.last = TRUE, decreasing = FALSE,
            method = c ("auto", "shell", "radix"), empty.last = na.last) {

            z <- list (...)
            decreasing <- as.logical (decreasing)

            if (
                length (z) == 1L &&
                is.numeric (x <- z[[1L]]) &&
                !is.object(x) &&
                length (x) > 0
            ) {
                if (
                    eval (parse (
                        text = ".Internal(sorted_fpass(x, decreasing, na.last))"
                    ))
                ) {
                    return (seq_along(x))
                }
            }

            method <- match.arg (method)

            has_declared <- any (vapply (z, is.declared, logical (1L)))
            if (has_declared && length (z) > 1L) {
                z <- lapply (z, function (x) {
                    if (is.declared (x)) {
                        return (xtfrm_declared (
                            x,
                            decreasing = decreasing,
                            na.last = na.last,
                            empty.last = empty.last
                        ))
                    }
                    else if (is.object (x)) {
                        return (as.vector (xtfrm (x)))
                    }
                    else {
                        return (x)
                    }
                })
                return (do.call ("order", c (z, list (na.last = na.last,
                    decreasing = decreasing, method = method))))
            }

            if (any (vapply (z, function (x) {
                    is.object(x) && !is.declared (x)
                }, logical (1L)))) {
                z <- lapply (z, function (x) if (is.object(x))
                    as.vector (xtfrm(x))
                else x)
                return (do.call ("order", c (z, list (na.last = na.last,
                    decreasing = decreasing, method = method))))
            }

            if (method == "auto") {
                useRadix <- all (vapply (z, function (x) {
                    (
                        is.numeric (x) ||
                        is.factor (x) ||
                        is.logical (x)
                    ) &&
                    is.integer (length (x))
                }, logical (1L)))
                method <- ifelse (useRadix, "radix", "shell")
            }


            if (length (z) == 1L && is.declared (x)) {
                return (order_declared (
                    x, na.last = na.last, decreasing = decreasing,
                    method = method, empty.last = empty.last
                ))
            }

            if (method != "radix" && !is.na (na.last)) {
                return (eval (parse (
                    text = ".Internal(order (na.last, decreasing, ...))"
                )))
            }

            if (method == "radix") {
                decreasing <- rep_len(as.logical (decreasing), length (z))
                return (eval (parse (
                    text = paste (
                        ".Internal(radixsort (",
                        "na.last, decreasing, FALSE, TRUE, ...",
                        "))"
                    )
                )))
            }

            if (any (diff((l.z <- lengths(z)) != 0L))) {
                stop ("argument lengths differ")
            }

            na <- vapply (z, is.na, rep.int(NA, l.z[1L]))

            ok <- if (is.matrix(na)) {
                rowSums(na) == 0L
            }
            else {
                !any (na)
            }

            if (all (!ok)) {
                return (integer())
            }

            z[[1L]][!ok] <- NA

            ans <- do.call ("order", c (z, list (decreasing = decreasing)))

            ans[ok[ans]]
        }

        do.call ("unlockBinding", list (sym = "as.factor", env = env))

        env$as.factor <- function (
            x,
            levels = c ("labels", "values", "both"),
            drop_na = TRUE,
            nolabels = FALSE,
            ordered = FALSE,
            ...
        ) {
            if (is.declared (x)) {
                x <- sanitize_na_index_ (x)
                levels <- match.arg (levels)
                labels <- attr (x, "labels", exact = TRUE)
                nv <- names_values (x, drop_na = drop_na)

                if (isFALSE (drop_na)) {
                    x <- undeclare (x)
                }

                if (levels == "labels") {
                    if (isTRUE (nolabels)) {
                        nv <- nv[is.element (nv, labels)]
                        x[!is.element (drop(undeclare (x)), labels)] <- NA
                    }

                    x <- factor (
                        as.character (x),
                        levels = names (nv),
                        ordered = ordered
                    )
                }
                else if (levels == "values") {
                    attributes (x) <- NULL
                    x <- factor (
                        x,
                        levels = unname (nv),
                        ordered = ordered
                    )
                }
                else if (levels == "both") {
                    nms <- paste0 ("[", nv, "] ", names (nv))

                    x <- factor (
                        nms[match(x, nv)],
                        levels = nms,
                        ordered = ordered
                    )
                }

                return (x)
            }
            else {
                if (is.factor (x))
                    x
                else if (!is.object(x) && is.integer (x)) {
                    levels <- sort.int(unique.default(x))
                    f <- match(x, levels)
                    levels (f) <- as.character (levels)
                    if (!is.null (nx <- names (x)))
                        names (f) <- nx
                    class (f) <- "factor"
                    f
                }
                else factor (x)
            }
        }
        do.call ("unlockBinding", list (sym = "drop", env = env))

        env$drop <- function (x) {
            if (is.declared (x)) {
                attributes (x) <- NULL
                return (x)
            }

            eval (parse (text = ".Internal(drop(x))"))
        }

        do.call ("unlockBinding", list (sym = "match", env = env))

        env$match <- function (
            x, table, nomatch = NA_integer_, incomparables = NULL
        ) {
            x <- undeclare (x, drop = TRUE)
            table <- undeclare (table, drop = TRUE)

            eval (parse (
                text = ".Internal(match(x, table, nomatch, incomparables))"
            ))
        }

        do.call ("unlockBinding", list (sym = "%in%", env = env))

        env$`%in%` <- function (x, table) {
            x <- undeclare (x, drop = TRUE)
            table <- undeclare (table, drop = TRUE)

            eval (parse (
                text = ".Internal(match(x, table, 0L, NULL))"
            )) > 0L
        }

        do.call ("unlockBinding", list (sym = "is.element", env = env))

        env$is.element <- function (el, set) {
            match(
                as.vector (undeclare (el, drop = TRUE)),
                as.vector (undeclare (set, drop = TRUE)),
                0L
            ) > 0L
        }
    # }

    # if (unlockEnvironment_ (asNamespace("stats"))) {

        env <- as.environment("package:stats")

        do.call ("unlockBinding", list (sym = "sd", env = env))

        env$sd <- function (x, na.rm = FALSE) {
            if (is.declared (x)) {
                x <- sanitize_na_index_ (x)
                na_index <- attr (x, "na_index")
                if (!is.null (na_index)) {
                    x <- x[-na_index]
                }
                class (x) <- setdiff (class (x), "declared")
            }

            sqrt (
                var (
                    if (is.vector(x) || is.factor (x)) x else as.double(x),
                    na.rm = na.rm
                )
            )
        }

        do.call ("unlockBinding", list (sym = "var", env = env))

        env$var <- function (x, y = NULL, na.rm = FALSE, use) {
            if (is.declared (x)) {
                x <- sanitize_na_index_ (x)
                na_index <- attr (x, "na_index")
                if (!is.null (na_index)) {
                    x <- x[-na_index]
                }
                class (x) <- setdiff (class (x), "declared")
            }

            if (missing(use)) {
                use <- ifelse (na.rm, "na.or.complete", "everything")
            }

            na.method <- pmatch(
                use,
                c (
                    "all.obs", "complete.obs", "pairwise.complete.obs",
                    "everything", "na.or.complete"
                )
            )

            if (is.na (na.method)) {
                stop ("invalid 'use' argument")
            }

            if (!is.null (x)) {
                if (is.data.frame (x)) {
                    x <- as.matrix (x)
                }
                else {
                    stopifnot(is.atomic (x))
                }
            }

            if (!is.null(y)) {
                if (is.data.frame (y)) {
                    y <- as.matrix (y)
                }
                else {
                    stopifnot(is.atomic (y))
                }
            }

            eval (parse (
                text = '.Call (stats:::C_cov, x, y, na.method, FALSE)'
            ))
            # .Call (C_cov, x, y, na.method, FALSE)
        }

        do.call ("unlockBinding", list (sym = "fivenum", env = env))

        env$fivenum <- function (x, na.rm = FALSE) {
            if (is.declared (x)) {
                x <- sanitize_na_index_ (x)
                na_index <- attr (x, "na_index")
                if (!is.null (na_index)) {
                    x <- x[-na_index]
                }
                class (x) <- setdiff (class (x), "declared")
            }

            xna <- is.na (x)
            if (any (xna)) {
                if (na.rm)
                    x <- x[!xna]
                else return (rep.int(NA, 5))
            }
            x <- sort (x)
            n <- length (x)
            if (n == 0)
                rep.int(NA, 5)
            else {
                n4 <- floor ((n + 3)/2)/2
                d <- c (1, n4, (n + 1)/2, n + 1 - n4, n)
                0.5 * (x[floor (d)] + x[ceiling(d)])
            }
        }
    # }


    env <- as.environment("package:graphics")

    do.call ("unlockBinding", list (sym = "dotchart", env = env))

    env$dotchart <- function(
        x, labels = NULL, groups = NULL, gdata = NULL, offset = 1/8,
        ann = par("ann"), xaxt = par("xaxt"), frame.plot = TRUE, log = "",
        cex = par("cex"), pt.cex = cex,
        pch = 21, gpch = 21, bg = par("bg"),
        color = par("fg"), gcolor = par("fg"), lcolor = "gray",
        xlim = range(x[is.finite(x)]),
        main = NULL, xlab = NULL, ylab = NULL, ...
    ) {
        ## old-style "graphics" design-bug: ("mar"), ("mai"), ("mar", "mai")
        ##			    all fail, just the following, ("mai", "mar") is ok:
        opar <- par("mai", "mar", "mgp", "cex", "yaxs")
        on.exit(par(opar))
        par(cex = cex, yaxs = "i")

        if(!is.numeric(x))
            stop("'x' must be a numeric vector or matrix")
        n <- length(x)
        if (is.matrix(x)) {
            if (is.null(labels))
                labels <- rownames(x)
            if (is.null(labels))
                labels <- as.character(seq_len(nrow(x)))
            labels <- rep_len(labels, n)
            if (is.null(groups))
                groups <- col(x, as.factor = TRUE)
            glabels <- levels(groups)
        } else {
            if (is.null(labels)) labels <- names(x)
            if (inherits(x, "wtable")) {
                x <- as.numeric(x)
            }
            glabels <- if(!is.null(groups)) levels(groups)
            if (!is.vector(x)) { # e.g. a table
                warning("'x' is neither a vector nor a matrix: using as.numeric(x)")
                x <- as.numeric(x)
            }
        }

        plot.new() # for strwidth()

        linch <-
        if(!is.null(labels)) max(strwidth(labels, "inch"), na.rm = TRUE) else 0
        if (is.null(glabels)) {
            ginch <- 0
            goffset <- 0
        }
        else {
            ginch <- max(strwidth(glabels, "inch"), na.rm = TRUE)
            goffset <- offset
        }
        nmai <- opar[["mai"]]
        if(ann)
            nm.2 <- nmai[2L]
        if (!(is.null(labels) && is.null(glabels))) {
            ## The intention seems to be to balance the whitespace
            ## on each side (2 & 4) of the labels+plot.
            yi <- if(is.null(ylab) || !ann) 0 else offset
            nm.2 <- nmai[4L] + max(yi + linch + goffset, ginch) + 1/16
            if (nmai[2L] <	nm.2) { ## add space for ylab + glabels on left margin
                nmai[2L] <- nm.2
                par(mai = nmai)
            }
        }

        if (is.null(groups)) {
            o <- seq_len(n)
            y <- o
            ylim <- c(0, n + 1)
        }
        else {
            o <- sort.list(as.numeric(groups), decreasing = TRUE)
            x      <- x     [o]
            groups <- groups[o]
            color  <- rep_len(color,  length(groups))[o]
            lcolor <- rep_len(lcolor, length(groups))[o]
            pch    <- rep_len(pch,    length(groups))[o]
            of.1 <- cumsum(c(0, diff(as.numeric(groups)) != 0))
            y <- seq_len(n) + 2 * of.1
            ylim <- range(0, y + 2)
        }

        plot.window(xlim = xlim, ylim = ylim, log = log)
        #    xmin <- par("usr")[1L]
        lheight <- par("csi")
        if (!is.null(labels)) {
            loffset <- (linch + 0.1)/lheight
            mtext(labels[o], side = 2, line = loffset, at = y, adj = 0,
                    col = color, las = 2, cex = cex, ...)
        }
        abline(h = y, lty = "dotted", col = lcolor)
        points(x, y, pch = pch, col = color, bg = bg, cex = pt.cex/cex)
        if (!is.null(groups)) {
            gpos <- rev(cumsum(rev(tapply(groups, groups, length)) + 2) - 1)
            ginch <- max(strwidth(glabels, "inch"), na.rm = TRUE)
            goffset <- (max(linch+offset, ginch, na.rm = TRUE) + 1/16)/lheight
                mtext(glabels, side = 2, line = goffset, at = gpos,
                        adj = 0, col = gcolor, las = 2, cex = cex, ...)
            if (!is.null(gdata)) {
                abline(h = gpos, lty = "dotted")
                points(gdata, gpos, pch = gpch, col = gcolor, bg = bg,
                            cex = pt.cex/cex, ...)
            }
        }
        axis(1, xaxt=xaxt) # FIXME? add '...' or use localAxis() as plot.default()
        if(frame.plot)
            box()
        if(ann) {
            title(main=main, xlab=xlab, ...) # with default "mgp"
            ## y-axis label must be left of the (regular + group) labels:
            mgp <- par("mgp")
            par(mgp = c(max(mgp[1], nm.2 / lheight - 1.5), mgp[-1]))
            title(ylab=ylab, ...)
        }
        invisible()
    }



    register_S3_method("labelled", "na_values", "declared")
    register_S3_method("labelled", "na_values<-", "declared")
    register_S3_method("labelled", "na_range", "declared")
    register_S3_method("labelled", "na_range<-", "declared")
    register_S3_method("labelled", "val_labels", "declared")
    register_S3_method("labelled", "val_labels<-", "declared")
    register_S3_method("labelled", "var_label", "declared")
    register_S3_method("labelled", "var_label<-", "declared")
    register_S3_method("labelled", "drop_unused_value_labels", "declared")
    register_S3_method("labelled", "val_label", "declared")
    register_S3_method("labelled", "val_label<-", "declared")
    register_S3_method("labelled", "sort_val_labels", "declared")
    register_S3_method("labelled", "nolabel_to_na", "declared")
    register_S3_method("labelled", "val_labels_to_na", "declared")
    register_S3_method("labelled", "remove_labels", "declared")
    register_S3_method("labelled", "remove_user_na", "declared")
    register_S3_method("labelled", "to_factor", "declared")
    register_S3_method("labelled", "to_character", "declared")
    register_S3_method("labelled", "copy_labels", "declared")

    register_S3_method("haven", "as_factor", "declared")
    register_S3_method("haven", "zap_labels", "declared")
    register_S3_method("haven", "zap_missing", "declared")

    register_S3_method("pillar", "pillar_shaft", "declared")
    register_S3_method("pillar", "format", "pillar_shaft_declared_num")
    register_S3_method("pillar", "format", "pillar_shaft_declared_chr")

    register_S3_method("vctrs", "vec_ptype_abbr", "declared")
    register_S3_method("vctrs", "vec_ptype_full", "declared")
    register_S3_method("vctrs", "vec_proxy", "declared")
    register_S3_method("vctrs", "vec_restore", "declared")
    # register_S3_method("vctrs", "vec_ptype2", "declared")

    register_S3_method("vroom", "output_column", "declared")

    invisible()
}

# function copied from package haven
`register_S3_method` <- function (pkg, generic, class, fun = NULL) {
    stopifnot(is.character (pkg), length (pkg) == 1)
    stopifnot(is.character (generic), length (generic) == 1)
    stopifnot(is.character (class), length (class) == 1)

    if (is.null (fun)) {
        fun <- get(paste0 (generic, ".", class), envir = parent.frame())
    }
    else {
        stopifnot(is.function (fun))
    }

    if (is.element (pkg, loadedNamespaces())) {
        registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }

    # Always register hook in case package is later unloaded & reloaded
    setHook(
        packageEvent(pkg, "onLoad"),
        function (...) {
            registerS3method(generic, class, fun, envir = asNamespace(pkg))
        }
    )
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

`validate_declared` <- function (x = double(), labels = NULL, label = NULL,
                                na_values = NULL, na_range = NULL, ...) {

  if (
    !is.numeric (x) &&
    !is.character (x) &&
    !inherits (x, "Date") &&
    !all (is.na (x))
  ) {
    stopError_ ("`x` must be a numeric, character or a Date vector.")
  }

  if (!is.null (labels)) {
    if (is.null (names (labels))) {
      stopError_ ("`labels` must have names.")
    }

    if (any (duplicated (stats::na.omit (labels)))) {
      stopError_ ("Value `labels` must be unique.")
    }
  }

  if (
    !is.null (label) && !(
      is.atomic (label) && is.character (label) && length (label) == 1
    )
  ) {
    stopError_ ("`label` must be a character vector of length one.")
  }

  if (!is.null (na_values)) {
    if (any (is.na (na_values))) {
      stopError_ ("`na_values` should not contain NA values.")
    }
  }

  if (!is.null (na_range)) {
    type_ok <-  all (is.na (x)) ||
      (is.character (x) && is.character (na_range)) ||
      (is.numeric (x) && is.numeric (na_range))

    if (!type_ok || length (unique (na_range)) != 2) {
      stopError_ (
        "`na_range` must have two unique values of the same type as `x`."
      )
    }

    if (any (is.na (na_range))) {
      stopError_ ("`na_range` can not contain missing values.")
    }
  }
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @title Labelled vectors with declared missing values
#'
#' @description
#' The labelled vectors are mainly used to analyse social science data,
#' and the missing values declaration is an important step in the analysis.
#'
#' @details
#' The `declared` objects are very similar to the `haven_labelled_spss` objects
#' from package \bold{haven}. It has exactly the same arguments, but it features
#' a fundamental difference in the treatment of (declared) missing values.
#'
#' In package \bold{haven}, existing values are treated as if they were missing.
#' By contrast, in package \bold{declared} the NA values are treated as if they
#' were existing values.
#'
#' This difference is fundamental and points to an inconsistency in package
#' \bold{haven}: while existing values can be identified as missing using the
#' function \code{is.na()}, they are in fact present in the vector and other
#' packages (most importantly the base ones) do not know these values should be
#' treated as missing.
#'
#' Consequently, the existing values are interpreted as missing only by package
#' \bold{haven}. Statistical procedures will use those values as if they were
#' valid values.
#'
#' Package \bold{declared} approaches the problem in exactly the opposite way:
#' instead of treating existing values as missing, it treats (certain) NA values
#' as existing. It does that by storing an attribute containing the indices of
#' those NA values which are to be treated as declared missing values, and it
#' refreshes this attribute each time the declared object is changed.
#'
#' This is a trade off and has important implications when subsetting datasets:
#' all declared variables get this attribute refreshed, which consumes some time
#' depending on the number of variables in the data.
#'
#' The generic function `as.declared()` attempts to coerce only the compatible
#' types of objects, namely `haven_labelled` and `factor`s. Dedicated class
#' methods can be written for any other type of object, and users are free to
#' write their own. To end of with a declared object, additional metadata is
#' needed such as value labels, which values should be treated as missing etc.
#'
#' The measurement level is optional and, for the moment, purely aesthetic. It
#' might however be useful to (automatically) determine if a declared object is
#' suitable for a certain statistical analysis, for instance regression requires
#' quantitative variables, while some declared objects are certainly categorical
#' despite using numbers to denote categories.
#'
#' It distinguishes between `"categorical"` and `"quantitative"` types of
#' variables, and additionally recognizes `"nominal"` and `"ordinal"` as
#' categorical, and similarly recognizes `"interval"`, `"ratio"`,
#' `"discrete"` and `"continuous"` as quantitative.
#'
#' @return `declared()` and `as.declared()` return labelled vector of class
#' "declared". When applied to a data frame, `as.declared()` will return a
#' corresponding data frame with declared variables. `is.declared()` and
#' `anyNAdeclared()` return a logical value.
#'
#' @examples
#'
#' x <- declared(
#'     c(1:5, -1),
#'     labels = c(Good = 1, Bad = 5, DK = -1),
#'     na_values = -1
#' )
#'
#' x
#'
#' is.na(x)
#'
#' x > 0
#'
#' x == -1
#'
#' # Values are actually placeholder for categories,
#' # so labels work as if they were factors:
#' x == "DK"
#'
#'
#' # when newly added values are already declared as missing,
#' # they are automatically coerced
#' c(x, 2, -1)
#'
#' # switch NAs with their original values
#' undeclare(x)
#'
#' as.character(x)
#'
#' # Returning values instead of categories
#' as.character(x, values = TRUE)
#'
#' anyNAdeclared(x) # contains declared missing values
#'
#' anyNAdeclared(c(1:5, NA)) # no declared missing values
#' @param x A numeric vector to label, or a declared labelled vector
#' (for `undeclare`)
#' @param labels A named vector or `NULL`. The vector should be the same type
#' as `x`. Unlike factors, labels don't need to be exhaustive: only a fraction
#' of the values might be labelled
#' @param na_values A vector of values that should also be considered as missing
#' @param na_range A numeric vector of length two giving the (inclusive) extents
#' of the range. Use \code{-Inf} and \code{Inf} if you want the range to be
#' open ended
#' @param label A short, human-readable description of the vector
#' @param measurement Optional, user specified measurement level
#' @param decimals Optional, maximum number of decimal places used for display
#' only
#' @param llevels Logical, when `x` is a factor only use those levels that have
#' labels
#' @param ... Other arguments used by various other methods
#' @name declared
NULL


#' @rdname declared
#' @export
declared <- function (
    x, labels = NULL, na_values = NULL, na_range = NULL, label = NULL,
    measurement = NULL, decimals = NULL, llevels = FALSE, ...
) {
  UseMethod ("declared")
}


#' @export
declared.default <- function (
    x, labels = NULL, na_values = NULL, na_range = NULL, label = NULL,
    measurement = NULL, decimals = NULL, llevels = FALSE, ...
) {

  xdate <- inherits(x, "Date")
  if (is.factor (x)) {
    nms <- levels (x)
    if (is.null (labels)) {
      labels <- seq (length (nms))
      names (labels) <- nms
    }

    if (isTRUE (llevels)) {
      labels <- labels[!possibleNumeric_ (names (labels), each = TRUE)]
    }

    wnms <- which (is.element (na_values, nms))

    if (length (wnms) > 0) {
      for (i in wnms) {
        na_values[i] <- which (nms == na_values[i])
      }
      if (possibleNumeric_ (na_values)) {
        na_values <- asNumeric_ (na_values)
      }
    }

    x <- as.numeric (x)
  }

  xchar <- FALSE

  if (is.null (labels)) {
    labels <- attr (x, "labels", exact = TRUE)
  }

  if (!is.null (labels)) {
    nms <- names (labels)
    if (xdate && !possibleNumeric_ (labels)) {
      stopError_ ("For date objects, the labels should be numeric.")
    }

    if (
      possibleNumeric_ (labels) && (
        xdate | possibleNumeric_ (x) | all (is.na (x))
      )
    ) {
      labels <- asNumeric_ (labels)
    }
    else {
      labels <- as.character (labels, values = TRUE)

      if (length(setdiff(labels, na_values)) > 0) {
        xchar <- !possibleNumeric_ (setdiff(labels, na_values))
      }

      if (xchar) {
        x <- as.character (x, values = TRUE)
        na_range <- NULL
      }
    }

    names (labels) <- nms

    # 2023.05.08 rationale, ex. ESS10 (integrated data file) has a variable
    # called region, which has multiple duplicated values and labels:
    # deprecated codes
    #        "DEP"
    # this is a mistake of course, but it nevertheless prevents subsetting,
    # because of validate_declared()
    labels <- labels[!duplicated(labels)]
  }

  if (!is.null (na_values)) {
    if (xdate && !possibleNumeric_ (na_values)) {
      stopError_ ("For date objects, the declared NA values should be numeric.")
    }

    if (possibleNumeric_ (na_values) & !xchar) {
      na_values <- asNumeric_ (na_values)
    }
    else {
      na_values <- as.character (na_values, values = TRUE)
    }
  }

  na_range <- sort (na_range)
  if (!is.null (na_range)) {
    if (xdate && !possibleNumeric_ (na_range)) {
      stopError_ ("For date objects, the declared NA range should be numeric.")
    }
  }
  if (!xdate) {
    attributes (x) <- NULL
    validate_declared (x, labels, label, na_values, na_range)
  }

  misvals <- all_missing_values (x, na_values, na_range, labels)

  # attr (x, "xdate") <- xdate
  na_index <- which(is.element(x, misvals))

  if (length(na_index) > 0) {
    declared_nas <- x[na_index]
    if (xdate) {
      declared_nas <- as.numeric (declared_nas)
    }

    x[na_index] <- NA
    names(na_index) <- declared_nas
  }
  else {
    na_index <- NULL
  }

  if ((possibleNumeric_ (x) || all (is.na (x))) & !xchar) {
    x <- asNumeric_ (x)
  }
  else if (!xdate) {
    x <- as.character (x, values = TRUE)
  }

  attr (x, "na_index") <- na_index
  attr (x, "na_values") <- na_values
  attr (x, "na_range") <- na_range
  attr (x, "labels") <- labels
  attr (x, "label") <- label
  attr (x, "date") <- xdate

  attr (x, "measurement") <- check_measurement (measurement)
  attr (x, "decimals") <- check_decimals (decimals)
  class(x) <- unique (c ("declared", class (x)))
  return (x)
}


#' @rdname declared_internal
#' @keywords internal
#
# @description
# `direct_declared()` is a low-level constructor for callers that already know
# the final storage mode and metadata of the vector. It does not validate,
# coerce, scan for missing codes, or replace values with `NA`.
#
# Callers are expected to pass a vector already stored in its final form, with
# declared missing values already represented as `NA`, and with `na_index`
# containing the original declared missing codes as names.
#
# @param x An atomic vector already in its final storage mode.
# @param na_index An optional named numeric vector indicating which `NA`
# positions correspond to declared missing values. The names should contain the
# original missing codes.
# @param na_values A vector of declared missing values metadata.
# @param na_range A vector of length two describing the declared missing range.
# @param labels A named vector of value labels.
# @param label A short, human-readable description of the vector.
# @param measurement Optional, user specified measurement level.
# @param decimals Optional, maximum number of decimal places used for display
# only.
# @param date Logical, whether `x` should be treated as a date vector.
# @return A vector of class `"declared"`.
#' @export
direct_declared <- function (
    x, na_index = NULL, na_values = NULL, na_range = NULL, labels = NULL,
    label = NULL, measurement = NULL, decimals = NULL, date = inherits(x, "Date")
) {
  if (!is.atomic(x)) {
    stopError_("`x` must be an atomic vector.")
  }

  if (!is.null(na_index) && !is.numeric(na_index)) {
    stopError_("`na_index` must be a numeric vector of indices.")
  }

  if (!is.null(na_index) && is.null(names(na_index))) {
    stopError_("`na_index` must have names.")
  }

  if (!is.null(label) && !(
    is.atomic(label) && is.character(label) && length(label) == 1
  )) {
    stopError_("`label` must be a character vector of length one.")
  }

  if (!is.logical(date) || length(date) != 1) {
    stopError_("`date` must be a logical vector of length one.")
  }

  .Call(
    "_directDeclared",
    x,
    na_index,
    na_values,
    na_range,
    labels,
    label,
    check_measurement(measurement),
    check_decimals(decimals),
    date,
    unique(c("declared", class(x)))
  )
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @title
#' Get / Declare missing values
#'
#' @description
#' Functions to extract information about the declared missing values, or to
#' declare such values if they are present in the data.
#'
#' @return
#'
#' \code{missing_values()} will return a vector of one or more values.
#'
#' \code{missing_range()} will return a numeric vector of length 2.
#'
#' @examples
#' x <- declared(c(-2, 1:5, -1),
#'     labels = c(Good = 1, Bad = 5, DK = -1, NotApplicable = -2),
#'     na_values = c(-1, -2)
#' )
#' x
#'
#' missing_values(x)
#'
#' missing_range(x) <- c(-10, -7)
#'
#' missing_range(x)
#'
#' @param x A vector.
#'
#' @param value Any vector of values that should be declared as missing
#' (for `missing_values`) or a numeric vector of length two giving the
#' (inclusive) extents of the range of missing values (for `missing_range`).
#'
#' @export
`missing_values` <- function (x) {
  UseMethod ("missing_values")
}


#' @export
`missing_values.default` <- function (x) {
  # return nothing
  NULL
}


#' @export
`missing_values.haven_labelled_spss` <- function (x) {
  attr (x, "na_values", exact = TRUE)
}


#' @export
`missing_values.declared` <- function (x) {
  x <- sanitize_na_index_ (x)
  attr (x, "na_values", exact = TRUE)
}


#' @export
`missing_values.data.frame` <- function (x) {
  lapply (x, missing_values)
}


#' @rdname missing_values
#'
#' @export
`missing_values<-` <- function (x, value) {
  UseMethod ("missing_values<-")
}


#' @export
`missing_values<-.default` <- function (x, value) {
  # do nothing
  x
}


#' @export
`missing_values<-.declared` <- function (x, value) {
  x <- sanitize_na_index_ (x)
  declared (undeclare (x),
           labels = attr (x, "labels", exact = TRUE),
           na_values = value,
           na_range = attr (x, "na_range", exact = TRUE),
           label = attr (x, "label", exact = TRUE),
           decimals = attr (x, "decimals", exact = TRUE)
  )
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Dynamically exported, see onLoad.R
# this functions will be registered when or if the package pillar is loaded
# using eval (parse ()) to avoid the huge dependency tree of vctrs, haven,
# labelled and pillar

`pillar_shaft.declared` <- function (
    x,
    show_labels = getOption ("declared.show_pillar_labels", TRUE),
    ...) {

    x <- sanitize_na_index_ (x)
    dots <- list (...)

    if (!isFALSE (dots$use_haven) && is.null (attr (x, "decimals"))) {
      if (eval (parse (text = "requireNamespace('haven', quietly = TRUE)"))) {
          return (eval (parse (
            text = "pillar::pillar_shaft(as.haven (x))"
          )))
      }
    }

    if (!isTRUE (show_labels) | !pillar_print_pkgs_available ()) {
        return (eval (parse (
            text = "pillar::pillar_shaft(drop(undeclare (x)))"
        )))
    }

    if (is.numeric (x)) {
        val <- val_num_pillar_info(x)
        lbl <- lbl_pillar_info(x)

        mw <- max (
            val$disp_short$lhs_ws + val$disp_short$main_wid + lbl$wid_short
        )
        w <- max (val$disp_full$lhs_ws + val$disp_full$main_wid + lbl$wid_full)

        eval (parse (text = paste (
            "pillar::new_pillar_shaft(list (val = val, lbl = lbl),",
            "min_width = mw, width = w, class = 'pillar_shaft_declared_num')"
        )))
    }
    else {
        val <- val_chr_pillar_info(x)
        lbl <- lbl_pillar_info(x)

        mw <- max (val$wid_short + lbl$wid_short)
        w <- max (val$wid_full + lbl$wid_full)

        eval (parse (text = paste (
            "pillar::new_pillar_shaft(list (val = val, lbl = lbl),",
            "min_width = mw, width = w, class = 'pillar_shaft_declared_chr')"
        )))
    }
}

`val_num_pillar_info` <- function (x) {
    val_pillar <- eval (parse (
        text = "pillar::pillar_shaft(haven::zap_labels (undeclare (x)))"
    ))

    disp_short <- num_disp_components (
        x, val_pillar, attr (val_pillar, "min_width")
    )
    disp_full <- num_disp_components (x, val_pillar, attr (val_pillar, "width"))

    list (
        disp_short = disp_short,
        disp_full = disp_full
    )
}

`num_disp_components` <- function (x, pillar, width) {
    if (!is.null (attr (x, "decimals"))) {
        display <- format (
            format_decimals_ (unclass (undeclare (x)), attr (x, "decimals")),
            justify = "right"
        )
    }
    else {
        display <- format (pillar, width)
    }
    # Sometimes there's an extra leading space from pillar
    display <- trim_ws_lhs (display)
    # exponent notation formatting hinders stripping white space in NAs
    display[is.na (unclass (x))] <- eval (parse (
        text = "crayon::strip_style(display[is.na (unclass (x))])"
    ))

    display_untrimmed_wid <- eval (parse (text = "pillar::get_extent(display)"))
    display_max_wid <- max (display_untrimmed_wid)
    display <- trim_ws_rhs (display)
    main_wid <- eval (parse (text = "pillar::get_extent(display)"))
    display_trimmed_rhs <- display_untrimmed_wid - main_wid

    # display[is.na (unclass (x))] <- eval (parse (
    #     text = "pillar::style_na(display[is.na (unclass (x))])"
    # ))

    list (
        lhs_ws = max (main_wid + display_trimmed_rhs) -
                    (main_wid + display_trimmed_rhs),
        main_wid = main_wid,
        main_txt = display,
        rhs_ws = display_trimmed_rhs
    )
}

`val_chr_pillar_info` <- function (x) {
    MIN_CHR_DISPLAY <- 4
    val_pillar <- eval (parse (
        text = "pillar::pillar_shaft(haven::zap_labels (x))"
    ))
    disp_full <- trim_ws_rhs (format (val_pillar, attr (val_pillar, "width")))
    wid_full <- eval (parse (text = "pillar::get_extent(disp_full)"))

    list (
        val_pillar = val_pillar,
        wid_short = pmin (MIN_CHR_DISPLAY, wid_full),
        disp_full = disp_full,
        wid_full = wid_full
    )
}

`lbl_pillar_info` <- function (x) {
    x <- sanitize_na_index_ (x)
    MIN_LBL_DISPLAY <- 6
    labels <- attr (x, "labels")
    if (length (labels) > 0) {
        names (labels) <- eval (parse (
            text = "pillar::style_subtle(paste0 (' [', names (labels), ']'))"
        ))
        attr (x, "labels") <- labels
        label_display <- eval (parse (
            text = "as.character (x, drop_na = FALSE, nolabels = TRUE)"
        ))
        label_display[is.na (label_display)] <- ""
    } else {
        label_display <- character (length (x))
    }
    label_widths <- eval (parse (text = "pillar::get_extent(label_display)"))
    label_min_widths <- ifelse (
        label_widths > 0, pmin (MIN_LBL_DISPLAY, label_widths), 0
    )

    if (inherits (x, "declared")) {
        MIN_NA_DISPLAY <- 4
        na_display <- character (length (x))
        na_index <- attr (x, "na_index")
        eval (parse (
            text = "na_display[na_index] <- pillar::style_na(' (NA)')"
        ))
        na_widths <- eval (parse (
            text = "pillar::get_extent(na_display)"
        ))

        label_display <- paste0 (na_display, label_display)
        label_widths <- label_widths + na_widths
        label_min_widths <- label_min_widths + ifelse (
            label_widths > 0,
            pmin (MIN_NA_DISPLAY, label_widths),
            0
        )
    }

    ret <- list (
        wid_short = label_min_widths,
        disp_full = label_display,
        wid_full = label_widths
    )
    ret
}

#' @export
`format.pillar_shaft_declared_num` <- function (x, width, ...) {
    vshort <- x$val$disp_short
    vfull <- x$val$disp_full
    lbl_wid <- pmax (0, x$lbl$wid_short - vfull$rhs_ws)

    if (width >= max (vfull$lhs_ws +vfull$main_wid + lbl_wid)) {
        lbl_width <- width - (vfull$lhs_ws + vfull$main_wid)
        lbl <- str_trunc (x$lbl$disp_full, lbl_width, subtle = TRUE)
        out <- paste_with_align (
            vfull$main_txt, lbl, vfull$lhs_ws, vfull$rhs_ws
        )
    } else {
        lbl_width <- width - (vshort$lhs_ws + vshort$main_wid)
        lbl <- str_trunc (x$lbl$disp_full, lbl_width, subtle = TRUE)
        out <- paste_with_align (
            vshort$main_txt, lbl, vshort$lhs_ws, vshort$rhs_ws
        )
    }
    eval (parse (
        text = "pillar::new_ornament(out, width = width, align = 'right')"
    ))
}

#' @export
`format.pillar_shaft_declared_chr` <- function (x, width, ...) {
    if (width >= max (x$val$wid_full + x$lbl$wid_short)) {
        lbl_width <- width - x$val$wid_full
        lbl <- str_trunc (x$lbl$disp_full, lbl_width, subtle = TRUE)
        out <- paste0 (x$val$disp_full, lbl)
    } else {
        val_widths <- pmin (x$val$wid_full, width - x$lbl$wid_short)
        val_display <- str_trunc (x$val$disp_full, val_widths)
        lbl <- str_trunc (x$lbl$disp_full, width - val_widths, subtle = TRUE)
        out <- paste0 (val_display, lbl)
    }
    eval (parse (
        text = "pillar::new_ornament(out, width = width, align = 'left')"
    ))
}


# Helpers -----------------------------------------------------------------

`str_trunc` <- function (x, widths, subtle = FALSE) {
    str_width <- eval (parse (text = "pillar::get_extent(x)"))
    too_wide <- which (!is.na (x) & str_width > widths)

    continue_symbol <- eval (parse (text = "cli::symbol$continue"))
    if (subtle) continue_symbol <- eval (parse (
        text = "pillar::style_subtle(continue_symbol)"
    ))

    truncated <- Map(x[too_wide], widths[too_wide], f = function (item, wid) {
        aa <- eval (parse (text = "crayon::col_substr(item, 1, wid - 1)"))
        paste0 (aa, continue_symbol)
    })
    truncated <- as.vector (truncated, "character")
    x[too_wide] <- truncated

    x
}

`trim_ws_rhs` <- function (x) {
    sub ("[ \t\r\n]+$", "", x)
}

`trim_ws_lhs` <- function (x) {
    sub ("^[ \t\r\n]+", "", x)
}

`pad_space` <- function (n) {
    vapply (n, function (x) paste (rep (" ", x), collapse = ""), "")
}

`paste_with_align` <- function (x, y, lhs_ws, rhs_ws) {
    y_wid <- eval (parse (text = "pillar::get_extent(y)"))
    added_chars <- max (y_wid - rhs_ws)
    rhs_ws <- added_chars - (y_wid - rhs_ws)

    paste0 (pad_space (lhs_ws), x, y, pad_space (rhs_ws))
}

`pillar_print_pkgs_available` <- function () {
    eval (parse (text = "requireNamespace('crayon', quietly = TRUE)")) &
    eval (parse (text = "requireNamespace('cli', quietly = TRUE)"))
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @name labels
#' @family labelling functions
#' @title Get / Declare value labels
#'
#' @description
#' Functions to extract information about the declared variable / value labels,
#' or to declare such values if they are present in the data.
#'
#' @details
#' The function `labels()` is a adaptation of the base function to the objects
#' of class `declared`. In addition to the regular arguments, it has two
#' additional (logical) arguments called `prefixed` (FALSE by default), to 
#' retrieve the value labels prefixed with their values, and `print_as_df`
#' (TRUE by default) to print the result as a data frame.
#'
#' @return
#' \code{labels()} will return a named vector.
#'
#' \code{label()} will return a single character string.
#'
#' @examples
#' x <- declared(
#'     c(-2, 1:5, -1),
#'     labels = c("Good" = 1, "Bad" = 5, "DK" = -1),
#'     na_values = c(-1, -2),
#'     label = "Test variable"
#' )
#' x
#'
#' labels(x)
#'
#' labels(x, prefixed = TRUE)
#'
#' labels(x) <- c("Good" = 1, "Bad" = 5, "DK" = -1, "Not applicable" = -2)
#'
#' label(x)
#'
#' label(x) <- "This is a proper label"
#'
#' x
#'
#' @param x Any vector of values that should be declared as missing
#' (for `labels`) or a numeric vector of length two giving the (inclusive)
#' extents of the range of missing values (for `label`).
#'
#' @export
label <- function (x) {
  UseMethod ("label")
}


#' @export
label.default <- function (x) {
  attr (x, "label", exact = TRUE)
}


#' @export
label.haven_labelled <- function (x) {
  attr (x, "label", exact = TRUE)
}


#' @export
label.declared <- function (x) {
  attr (x, "label", exact = TRUE)
}


#' @export
label.data.frame <- function (x) {
  lapply (x, label)
}


#' @rdname labels
#' @param ... Other arguments, for internal use.
#' @param value The variable label, or a list of (named) variable labels
#' @export
`label<-` <- function (x, ..., value) {
  UseMethod ("label<-")
}


#' @export
`label<-.declared` <- function (x, ..., value) {
  if (!is.null (value) && length (value) > 1) {
    stopError_ ("`value` should be a single character string or NULL.")
  }

  if (is.null (value)) {
    attr (x, "label") <- NULL
  }
  else {
    attr (x, "label") <- as.character (value)
  }

  return (x)
}


#' @export
`label<-.haven_labelled` <- function (x, ..., value) {
  if (!is.null (value) && length (value) > 1) {
    stopError_ ("`value` should be a single character string or NULL.")
  }

  if (is.null (value)) {
    attr (x, "label") <- NULL
  }
  else {
    attr (x, "label") <- as.character (value)
  }

  return (x)
}



# the following two functions are deliberately taken from
# package Hmisc to ensure functionality upon a namespace collision

#' @export
`label<-.default` <- function (x, ..., value) {
  if (is.list (value)) {
    stopError_ ("cannot assign a list to be a object label")
  }

  if (!is.null (value) && length (value) != 1L) {
    stopError_ ("value must be character vector of length 1")
  }

  attr (x, "label") <- value

  if (! inherits (x, "labelled")) {
    class (x) <- c ("labelled", class (x))
  }

  return (x)
}


#' @export
`label<-.data.frame` <- function (x, self = TRUE, ..., value) {
  if (is.list (value)) {
    self <- FALSE
  }

  if (self) {
    if (!is.null (value) && length (value) > 1) {
      stopError_ ("`value` should be a single character string or NULL.")
    }

    xc <- class (x)
    xx <- unclass (x)
    if (is.null (value)) {
      attr (xx, "label") <- NULL
    }
    else {
      attr (xx, "label") <- as.character (value)
    }

    class (xx) <- xc
    return (xx)
  }
  else {
    if (length (value) != length (x)) {
      stopError_ ("value must have the same length as x")
    }

    for (i in seq (along.with = x)) {
      label(x[[i]]) <- value[[i]]
    }
  }

  return (x)
}






#' @export
labels.declared <- function (
    object, prefixed = FALSE, print_as_df = TRUE, ...
  ) {
    labels <- attr (object, "labels", exact = TRUE)
    
    if (is.null(labels)) {
      return(NULL)
    }

    if (isTRUE(prefixed)) {
      names (labels) <- paste0 ("[", labels, "] ", names (labels))
    }

    attr (labels, "print_as_df") <- isTRUE (print_as_df)
      
    return (
      structure (labels, class = unique (c ("labels_df", class (labels))))
    )
}


#' @export
labels.haven_labelled <- function (
    object, prefixed = FALSE, print_as_df = TRUE, ...
  ) {
    
    labels <- attr (object, "labels", exact = TRUE)
    
    if (is.null(labels)) {
      return(NULL)
    }

    if (isTRUE(prefixed)) {
      names (labels) <- paste0 ("[", labels, "] ", names (labels))
    }

    attr (labels, "print_as_df") <- isTRUE (print_as_df)
      
    return (
      structure (labels, class = unique (c ("labels_df", class (labels))))
    )
}


#' @export
labels.data.frame <- function (object, ...) {
  lapply (object, labels, ...)
}




#' @rdname labels
#' @export
`labels<-` <- function (x, value) {
  UseMethod ("labels<-")
}


#' @export
`labels<-.default` <- function (x, value) {
  # do nothing
  return (x)
}


#' @export
`labels<-.declared` <- function (x, value) {
    attr (x, "labels") <- value
    return (x)
}


#' @export
`labels<-.haven_labelled` <- function (x, value) {
  attr (x, "labels") <- value
  return (x)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 12
#' @param what A character vector or a named list of functions, identifying
#' what to compute. Available character values are `"n"`, `"NA"`, `"mode"`,
#' `"mean"`, `"median"`, `"range"`, `"var"`, `"sd"`, `"iqr"`, `"min"`, and
#' `"max"`.
#' @details
#' For `wmeasures()`, the `"n"` measure counts valid values, while `"NA"` counts
#' all missing values, both declared and empty.
#'
#' When `wmeasures()` is evaluated in the context of a data frame, such as with
#' `admisc::using()`, `.` can be used as a placeholder for all variables in the
#' current dataset.
#' @export
`wmeasures` <- function (
    x, what = c ("n", "mode", "mean", "median", "range", "var", "sd"),
    wt = NULL, na.rm = TRUE, ...
) {

    if (identical (substitute (x), quote (.))) {
        x <- currentDataset_ (parent.frame ())
    }

    what <- parseWhat_ (what)

    if (is.data.frame (x)) {
        result <- do.call (rbind, lapply (
            x,
            function (y) {
                calculateMeasures_ (
                    y, what = what, wt = wt, na.rm = na.rm, ... = ...
                )
            }
        ))

        rownames (result) <- names (x)
        class (result) <- c ("wmeasures", class (result))
        return (result)
    }

    result <- calculateMeasures_ (
        x, what = what, wt = wt, na.rm = na.rm, ... = ...
    )

    class (result) <- c ("wmeasures", class (result))
    return (result)
}

`currentDataset_` <- function (envir) {

    objects <- as.list (envir, all.names = FALSE)
    is_column <- vapply (
        objects,
        function (x) {
            !is.function (x) && !is.environment (x)
        },
        TRUE
    )

    objects <- objects[is_column]

    if (!length (objects)) {
        stopError_ ("Could not find a current dataset.")
    }

    lengths <- vapply (objects, NROW, integer (1))

    if (length (unique (lengths)) != 1 || lengths[1] == 0) {
        stopError_ ("Could not find a current dataset.")
    }

    return (as.data.frame (objects, optional = TRUE))
}

`parseWhat_` <- function (what) {

    builtin <- list (
        n = function (x, wt = NULL, na.rm = TRUE, ...) sum (!is.na (x)),
        "NA" = function (x, wt = NULL, na.rm = TRUE, ...) sum (is.na (x)),
        mean = wmean,
        sd = wsd,
        median = wmedian,
        iqr = wIQR,
        var = wvar,
        mode = wmode,
        min = function (x, wt = NULL, na.rm = TRUE, ...) {
            x <- drop_declared_missing_ (x, wt = wt)
            min (x$x, na.rm = na.rm)
        },
        max = function (x, wt = NULL, na.rm = TRUE, ...) {
            x <- drop_declared_missing_ (x, wt = wt)
            max (x$x, na.rm = na.rm)
        }
    )

    if (is.character (what)) {
        key <- tolower (what)
        choices <- list (
            n = "n",
            na = "NA",
            mean = "mean",
            average = "mean",
            sd = "sd",
            stdev = "sd",
            median = "median",
            iqr = "iqr",
            var = "var",
            variance = "var",
            mode = "mode",
            min = "min",
            max = "max",
            range = c ("min", "max")
        )

        if (!all (key %in% names (choices))) {
            stopError_ ("Unknown measure name.")
        }

        what <- builtin[unlist (choices[key], use.names = FALSE)]
    }
    else if (is.function (what)) {
        what <- list (what)
    }

    if (!is.list (what) || !all (vapply (what, is.function, TRUE))) {
        stopError_ (
            "Argument `what` should be a character vector or a list of functions."
        )
    }

    nms <- names (what)

    if (is.null (nms)) {
        nms <- rep ("", length (what))
    }

    missing_names <- !nzchar (nms)

    if (any (missing_names)) {
        nms[missing_names] <- paste0 ("Measure", which (missing_names))
    }

    names (what) <- nms

    return (what)
}

`drop_declared_missing_` <- function (x, wt = NULL) {

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (length (na_index)) {
            x <- x[-na_index]
            wt <- wt[-na_index]
        }
    }

    attributes (x) <- NULL

    return (list (x = x, wt = wt))
}

`calculateMeasures_` <- function (
    x, what, wt = NULL, na.rm = TRUE, ...
) {

    result <- vapply (
        what,
        function (fun) {
            callMeasure_ (fun, x = x, wt = wt, na.rm = na.rm, ... = ...)
        },
        numeric (1)
    )

    return (result)
}

`callMeasure_` <- function (fun, ...) {

    arguments <- list (...)
    fmls <- names (formals (fun))

    if (!is.null (fmls) && !("..." %in% fmls)) {
        arguments <- arguments[names (arguments) %in% fmls]
    }

    return (do.call (fun, arguments))
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @export
`print.declared` <- function (x, ...) {
    x <- sanitize_na_index_ (x)
    label <- label (x)
    if (!is.null (label)) {
        label <- paste ("", label)
    }

    measurement <- attr (x, "measurement")
    # type <- likely_type (x)
    cat (
        paste0 (
            "<declared<",
            # ifelse (is.null (measurement), "", paste0 (", ", measurement)),
            # ifelse (is.null (type), "", paste0 (", ", type)),
            # likely_type (x),
            ifelse (
                isTRUE (attr (x, "date")),
                "Date",
                setdiff (class (x), "declared")[1]
            ),
            ">[", length (x), "]>",
            label,
            "\n"
        )
    )

    if (length (x) > 0) {
        print (noquote (format_declared (x)), ...)

        na_values <- attr (x, "na_values")
        if (!is.null (na_values)) {
            cat (paste0 (
                "Missing values: ", paste (na_values, collapse = ", "), "\n"
            ))
        }

        na_range <- attr (x, "na_range")
        if (!is.null (na_range)) {
            cat (paste0 (
                "Missing range:  [",
                paste (na_range, collapse = ", "),
                "]\n"
            ))
        }

        labels <- attr (x, "labels", exact = TRUE)

        if (length (labels) == 0) {
            return (invisible(x))
        }

        large <- length(labels) > 7

        if (large) {
            rest <- length(labels) - 7
            labels <- labels[1:7]
        }

        cat ("\nLabels:", "\n", sep = "")

        print (
            data.frame (
                value = unname (labels),
                label = names (labels),
                row.names = NULL
            ),
            row.names = FALSE
        )

        if (large) {
            cat(paste(
                "# ... plus",
                rest,
                "more labels, use labels() to print them all.\n"
            ))
        }

        return (invisible(x))
    }
}


#' @export
`print.labels_df` <- function(x, ...) {
    class (x) <- setdiff (class (x), "labels_df")
    print_as_df <- isTRUE (attr (x, "print_as_df"))
    attr(x, "print_as_df") <- NULL
    if (print_as_df) {
        print (
            data.frame (
                value = unname (x),
                label = names (x)
            ),
            row.names = FALSE
        )
    }
    else {
        print (x)
    }
}


#' @export
`print.wtable` <- function (x, force = FALSE, startend = TRUE, ...) {
    toprint <- attr (x, "toprint")
    xlabel <- attr (toprint, "xlabel")
    ylabel <- attr (toprint, "ylabel")
    attr (toprint, "xlabel") <- NULL
    attr (toprint, "ylabel") <- NULL
    achar <- "\u00c2"
    tick <- c ("\u00b4", "\u0060", "\u2018", "\u2019") # ticks and single quotes
    tick <- c (paste0 (achar, "'"), paste0 (achar, tick), tick)

    if (!all (is.na (x)) && x[1] != as.matrix (toprint)[1]) {
        # this means the original table was altered. e.g. proportions (tbl)
        class (x) <- setdiff (class (x), c ("wtable", "array"))
        names (dimnames (x)) <- NULL
        attr (x, "toprint") <- NULL
        attr (x, "missing") <- NULL
        if (is.null (dim (x))) {
            names (x) <- gsub (paste (tick, collapse = "|"), "'", names (x))
        }
        else {
            rownames (x) <- gsub (
                paste (tick, collapse = "|"), "'", rownames (x)
            )
        }
        if (length (dimnames (x)) == 2) {
            colnames (x) <- gsub (
                paste (tick, collapse = "|"), "'", colnames (x)
            )
        }
        print (x)
    }
    else {
        x <- toprint

        if (!is.null (xlabel)) {
            cat (xlabel, "\n")
        }

        if (!is.null (ylabel)) {
            cat (ylabel, "\n")
        }

        if (is.matrix (x)) { # crosstab

            rnms <- strsplit (rownames (x), split = "_-_")

            xlabels <- sapply (rnms, tail, 1)
            # max.nchar.xlabels <- max (nchar (encodeString (xlabels)))
            # for (i in seq (length (xlabels))) {
            #     if (nchar (xlabels[i]) < max.nchar.xlabels) {
            #         xlabels[i] <- padLeft_ (
            #             xlabels[i], max.nchar.xlabels - nchar (xlabels[i])
            #         )
            #     }
            # }

            if (attr (x, "xvalues")) {
                # -length (rnms), because of "Total" which does not have a value
                xvalues <- unlist (lapply (rnms[-length (rnms)], "[[", 1))

                # suppress duplicate labels (when label equals value)
                if (length (xlabels) > 0) {
                    idx <- seq_along (xvalues)
                    xlabels[idx] <- ifelse (xlabels[idx] == xvalues, "", xlabels[idx])
                }

                max.nchar.xvalues <- max (nchar (encodeString (xvalues)))

                for (i in seq (length (xvalues) - 1)) {
                    if (nchar (xvalues[i]) < max.nchar.xvalues) {
                        xvalues[i] <- padLeft_ (
                            xvalues[i], max.nchar.xvalues - nchar (xvalues[i])
                        )
                    }
                }

                xvalues <- c (xvalues, paste(rep(" ", max.nchar.xvalues), collapse = ""))
                rnms <- paste (xvalues, xlabels)
            }
            else {
                rnms <- xlabels
            }

            rownames (x) <- rnms

            cnms <- colnames (x)
            if (attr (x, "yvalues")) {
                cnms <- vapply (
                    strsplit (cnms, split = "_-_"),
                    function(v) {
                        if ( length(v) < 2) return(v)
                        val <- v[1]
                        lab <- v[length (v)]
                        if (identical (val, lab)) {
                            val
                        } else {
                            paste (val, lab)
                        }
                    },
                    character(1L)
                )
            } else {
                cnms <- sapply (
                    strsplit (cnms, split = "_-_"),
                    tail,
                    1
                )
            }

            max.nchar.cols <- max (nchar (c (encodeString (cnms), x)))
            for (i in seq (length (cnms))) {
                if (nchar (cnms[i]) < max.nchar.cols) {
                    cnms[i] <- padBoth_ (
                        cnms[i], max.nchar.cols - nchar (cnms[i])
                    )
                }
            }
            colnames (x) <- cnms

            for (i in seq (length (x))) {
                if (nchar (x[i]) < max.nchar.cols) {
                    x[i] <- padBoth_ (x[i], max.nchar.cols - nchar (x[i]))
                }
            }

            class (x) <- setdiff (class (x), c ("w_table", "array"))
            attr (x, "xvalues") <- NULL
            attr (x, "yvalues") <- NULL
            rownames (x) <- gsub (
                paste (tick, collapse = "|"), "'", rownames (x)
            )
            colnames (x) <- gsub (
                paste (tick, collapse = "|"), "'", colnames (x)
            )
            cat (ifelse (startend, "\n", ""))
            print (noquote (x))
            cat (ifelse (startend, "\n", ""))
        }
        else {
            show_values <- attr (x, "show_values")
            valid <- attr (x, "valid")
            dots <- list (...)
            if (is.element ("show_values", names (dots))) {
                show_values <- dots$show_values
            }

            values <- gsub (
                paste (tick, collapse = "|"), "'", attr (x, "values")
            )
            labels <- gsub (
                paste (tick, collapse = "|"), "'", attr (x, "labels")
            )
            na_values <- gsub (
                paste (tick, collapse = "|"), "'", attr (x, "na_values")
            )

            first_missing <- Inf
            if (any (is.element (values, na_values))) {
                first_missing <- which (is.element (values, na_values))[1]
            }

            if (first_missing == Inf && is.na (labels[length (labels)])) {
                first_missing <- length (labels)
            }

            rnms <- labels
            max.nchar.values <- ifelse (length (values) > 0, max (nchar (values)), 0)

            if (show_values) {
                labels[!is.na (labels)][values == labels[!is.na (labels)]] <- ""
                values <- formatC (
                    as.character (values),
                    digits = max.nchar.values - 1,
                    flag = " "
                )
                rnms[!is.na (labels)] <- paste (values, labels[!is.na (labels)])
            }


            if (max.nchar.values > 1) {
                rnms[is.na (labels)] <- padLeft_ ("NA", max.nchar.values - 2)
            } else {
                if (max.nchar.values == 1) {
                    rnms[!is.na (labels)] <- sapply (
                        rnms[!is.na(labels)],
                        padLeft_,
                        max.nchar.values - 1
                    )
                }
                rnms[is.na (labels)] <- "NA"
            }

            max.nchar.cases <- max (nchar (encodeString (rnms)))
            # rnms <- sprintf (paste0 ("% ", max.nchar.cases, "s"), rnms)
            for (i in seq (length (rnms))) {
                if (nchar (rnms[i]) < max.nchar.cases) {
                    rnms[i] <- padRight_ (
                        rnms[i], max.nchar.cases - nchar (rnms[i])
                    )
                    # rnms[i] <- paste (
                    #     c (
                    #      rep (" ", max.nchar.cases - nchar (rnms[i])), rnms[i]
                    #     ),
                    #     collapse = "", sep = ""
                    # )
                }
            }

            sums <- colSums (x[, 1:3])

            fres <- formatC (as.character (c (x$fre, sums[1])), format = "s")
            fres <- paste (
                sprintf (
                    paste ("%", max (3, nchar (sums[1])), "s", sep = ""),
                    fres
                ),
                ""
            )

            x$rel <- formatC (x$rel, digits = 3, format = "f")
            rel <- sprintf ("% 5s", x$rel)
            x$per <- formatC (x$per, digits = 1, format = "f")
            if (valid & is.element ("vld", names (x))) {
                x$vld <- formatC (x$vld, digits = 1, format = "f")
                vld <- sprintf (paste ("% 5s", sep = ""), x$vld)
            }
            per <- sprintf ("% 5s", c (x$per, sums[3]))
            cpd <- formatC (x$cpd, digits = 1, format = "f")
            cpd <- sprintf (paste ("% 5s", sep = ""), cpd)

            miseparator <- paste (
                c (
                    # rep (
                    #     " ",
                    #     ifelse (max.nchar.cases > 5, max.nchar.cases - 5, 0)
                    # ),
                    rep (
                        "-",
                        min (max.nchar.cases, 5) + 1 * (sums[1] >= 1000)
                    ),
                    "\n"
                ),
                collapse = ""
            )
            separator <- paste (
                c (
                    rep (" ", max.nchar.cases + 1),
                    rep ("-", nchar (sums[1])),
                    ifelse (
                        nchar (sums[1]) < 3,
                        paste (
                            rep ("-", 3 - nchar (sums[1])),
                            collapse = ""
                        ),
                        ""
                    ),
                    sprintf (
                        "-------------------%s\n",
                        ifelse (valid, "------", "")
                    )
                ),
                collapse = ""
            )

            if (nrow (x) > 100 & !force) {
                stopError_ (paste (
                    "It looks like a lot of categories. If you really want to",
                    "print it, use:\nprint(x, force = TRUE)"
                ))
            }

            cat (ifelse (startend, "\n", ""))

            cat (
                paste (
                    rep (
                        " ",
                        max.nchar.cases + ifelse (
                            nchar (sums[1]) > 4, nchar (sums[1]) - 4, 0
                        )
                    ),
                    collapse = ""
                ),
                sprintf (ifelse (
                    sums[1] < 1000,
                    "fre    rel   per   %scpd\n",
                    " fre    rel   per   %scpd\n"
                ), ifelse (valid, "vld   ", ""))
            )

            cat (separator)

            for (i in seq (nrow (x))) {
                if (first_missing == i) {
                    cat (miseparator)
                }

                if (valid) {
                    if (i < first_missing) {
                        cat (
                            rnms[i], fres[i], rel[i], per[i], vld[i], cpd[i],
                            "\n"
                        )
                    }
                    else {
                        cat (rnms[i], fres[i], rel[i], per[i], "\n")
                    }
                }
                else {
                    cat (rnms[i], fres[i], rel[i], per[i], cpd[i], "\n")
                }
            }

            cat (separator)

            cat (
                paste (
                    rep (" ", max.nchar.cases),
                    sep = ""
                ),
                " ",
                fres[length (fres)],
                " 1.000 100.0\n",
                sep = ""
            )

            cat (ifelse (startend, "\n", ""))
        }
    }
}


#' @export
`print.fobject` <- function (x, startend = TRUE, ...) {

    class (x) <- setdiff (class (x), "fobject")

    x <- matrix(
        if (possibleNumeric_ (x)) round(asNumeric_ (x), 3) else x,
        nrow = 1,
        dimnames = list ("", names (x))
    )

    nax <- is.na (x)

    pN <- apply (x, 2, possibleNumeric_)
    nms <- colnames (x)

    cx <- x

    for (c in seq (ncol(x))) {
        xc <- x[, c]
        max.nchar.nc <- max (nchar (xc), na.rm = TRUE)
        ndec <- 0

        if (pN[c]) {
            ndec <- min (numdec_ (xc), 3)
            x[, c] <- sprintf (
                paste0 ("%", max.nchar.nc, ".", ndec, "f"),
                asNumeric_ (xc)
            )
        }

        # if (possibleNumeric_ (nms[c])) {
        #     # since this is a column name, most likely it is a whole number
        #     # e.g. the value of a declared object instead of the label
        #     nmsc <- sprintf (
        #         paste0 ("%", max.nchar.nc, ".", ndec, "f"),
        #         asNumeric_ (nms[c])
        #     )

        #     if (grepl ("[.]", nmsc)) {
        #         nmsc <- paste (
        #             unlist (strsplit (nmsc, split = "[.]"))[1],
        #             paste (rep (" ", ndec), collapse = "")
        #         )
        #     }

        #     nms[c] <- nmsc
        # }
    }

    x[nax] <- ""


    max.nchars <- max (nchar (c (encodeString (nms), x)), na.rm = TRUE)

    for (i in seq (length (nms))) {
        if (nchar (nms[i]) < max.nchars) {
            nms[i] <- padBoth_ (nms[i], max.nchars - nchar (nms[i]))
        }
    }

    for (i in seq (length (x))) {
        if (nchar (x[i]) < max.nchars) {
            x[i] <- padBoth_ (x[i], max.nchars - nchar (x[i]))
        }
    }

    colnames (x) <- nms

    cat (ifelse (startend, "\n", ""))
    print (noquote (x))
    cat (ifelse (startend, "\n", ""))

}


#' @export
`print.wmeasures` <- function (x, startend = TRUE, digits = 3, ...) {

    result <- x
    class (x) <- setdiff (class (x), "wmeasures")

    if (is.null (dim (x))) {
        x <- matrix (x, nrow = 1, dimnames = list ("", names (x)))
    }

    nms <- colnames (x)
    x <- matrix (
        apply (x, 2, format_wmeasure_column_, digits = digits),
        nrow = nrow (x),
        dimnames = dimnames (x)
    )

    widths <- pmax (
        nchar (encodeString (nms)),
        apply (x, 2, function (y) max (nchar (y)))
    )

    for (i in seq_along (nms)) {
        nms[i] <- padBoth_ (nms[i], widths[i] - nchar (nms[i]))
        x[, i] <- mapply (
            padLeft_,
            x[, i],
            widths[i] - nchar (x[, i]),
            USE.NAMES = FALSE
        )
    }

    rnms <- rownames (x)
    width_rnms <- max (nchar (rnms))
    row_prefix <- width_rnms > 0

    cat (ifelse (startend, "\n", ""))
    cat (
        paste0 (
            paste (rep (" ", width_rnms), collapse = ""),
            ifelse (row_prefix, " ", ""),
            paste (nms, collapse = "  "),
            "\n"
        )
    )

    for (i in seq (nrow (x))) {
        cat (
            ifelse (
                row_prefix,
                paste0 (padRight_ (rnms[i], width_rnms - nchar (rnms[i])), " "),
                ""
            ),
            paste (x[i, ], collapse = "  "),
            "\n",
            sep = ""
        )
    }
    cat (ifelse (startend, "\n", ""))

    return (invisible (result))
}


`format_wmeasure_column_` <- function (x, digits = 3) {

    result <- rep ("", length (x))
    ok <- !is.na (x)

    if (!any (ok)) {
        return (result)
    }

    x <- round (as.numeric (x[ok]), digits)
    ndec <- numdec_ (x, each = TRUE, maxdec = digits)
    maxdec <- max (ndec, na.rm = TRUE)

    result[ok] <- sprintf (paste0 ("%.", maxdec, "f"), x)

    return (result)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname declared
#'
#' @export
`as.declared` <- function (x, ...) {
  UseMethod ("as.declared")
}


#' @export
`as.declared.default` <- function (x, ...) {
  x <- sanitize_na_index_ (x)
  dots <- list (...)
  interactive <- isTRUE (dots$interactive)
  labels <- attr (x, "labels", exact = TRUE)
  na_index <- attr (x, "na_index", exact = TRUE)
  na_values <- attr (x, "na_values", exact = TRUE)
  na_range <- attr (x, "na_range", exact = TRUE)
  label <- attr (x, "label", exact = TRUE)
  measurement <- attr (x, "measurement", exact = TRUE)
  decimals <- attr (x, "decimals", exact = TRUE)
  xdate <- inherits (x, "Date") || isTRUE (attr (x, "date", exact = TRUE))

  metadata_present <- any (!vapply (
    list (labels, na_index, na_values, na_range, label, measurement, decimals),
    is.null,
    logical (1)
  ))

  if (metadata_present && (is.atomic (x) || inherits (x, "Date"))) {
    normalized_na_index <- !is.null (na_index) && length (na_index) > 0

    if (normalized_na_index) {
      return (direct_declared(
        x,
        na_index = na_index,
        na_values = na_values,
        na_range = na_range,
        labels = labels,
        label = label,
        measurement = measurement,
        decimals = decimals,
        date = xdate
      ))
    }

    return (declared(
      x,
      labels = labels,
      na_values = na_values,
      na_range = na_range,
      label = label,
      measurement = measurement,
      decimals = decimals
    ))
  }

  if (interactive) {
    msg <- "There is no automatic class method conversion for this type of"
    if (!is.null (dots$vname_)) {
      msg <- paste0 (dots$vname_, ": ", msg, " variable.")
    }
    else {
      msg <- paste (msg, "object.")
    }
    message (msg)
  }

  return (x)
}


#' @export
`as.declared.haven_labelled` <- function (x, ...) {

  dots <- list (...)

  na_values <- attr (x, "na_values")
  na_range <- attr (x, "na_range")
  labels <- attr (x, "labels", exact = TRUE)
  label <- attr (x, "label", exact = TRUE)
  xdate <- is.element ("Date", class (x))

  #---------------------------------------
  # necessary for DDIwR::convert
  format_spss <- attr (x, "format.spss")
  format_stata <- attr (x, "format.stata")
  format_sas <- attr (x, "format.sas")
  #---------------------------------------

  declared_nas <- NULL

  if (!inherits (x, "haven_labelled_spss")) {
    na_index <- which (hasTag_ (x))
    if (length (na_index) > 0) {
      declared_nas <- getTag_ (x[na_index])
      x[na_index] <- NA
      names(na_index) <- declared_nas
    }

    attributes (x) <- NULL

    if (!is.null (labels)) {
      nms <- names (labels)
      tagged <- hasTag_ (labels)

      if (any (tagged)) {
        labels[tagged] <- getTag_ (labels[tagged])
        na_values <- sort (unname (labels[tagged]))
      }

      # labels <- coerceMode_ (labels)
      names (labels) <- nms
    }

    misvals <- unique (sort (c (na_values, declared_nas)))
  } else {
    misvals <- all_missing_values (unclass (x))
    na_index <- which (is.element (x, misvals))
    if (length (na_index) > 0) {
      names (na_index) <- x[na_index]
      x[na_index] <- NA
    }
  }

  attributes (x) <- NULL
  pnx <- possibleNumeric_ (x) | all (is.na (x))

  charlabels <- FALSE
  if (!is.null (labels)) {
    if (length (setdiff (labels, na_values)) > 0) {
      charlabels <- !possibleNumeric_ (setdiff(labels, na_values))
    }
  }

  if (length (na_index) > 0) {
    attr (x, "na_index") <- na_index
  }
  attr (x, "na_values") <- na_values
  attr (x, "na_range") <- na_range
  attr (x, "labels") <- labels
  attr (x, "label") <- label
  attr (x, "format.spss") <- format_spss
  attr (x, "format.stata") <- format_stata
  attr (x, "format.sas") <- format_sas
  attr (x, "measurement") <- check_measurement (dots$measurement)
  attr (x, "date") <- xdate
  class(x) <- unique (c ("declared", class (x)))

  return (x)
}


#' @export
`as.declared.factor` <- function (x, ...) {
  return (declared (x, ... = ...))
}


#' @export
`as.declared.data.frame` <- function (x, ..., interactive = FALSE) {
  if (isFALSE (interactive)) {
    x[] <- lapply (x, as.declared, interactive = FALSE, ... = ...)
  }
  else {
    nms <- names (x)
    for (i in seq (length (nms))) {
      x[[i]] <- as.declared (x[[i]], vname_ = nms[i], ... = ...)
    }
  }

  return (x)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @export
`as.character.declared` <- function (
    x, drop_na = TRUE, values = FALSE, nolabels = FALSE, ...
) {

    x <- sanitize_na_index_ (x)
    allabels <- names_values (x, drop_na = drop_na)
    labels <- labels (x)
    attrx <- attributes (x)

    if (isFALSE (drop_na)) {
        x <- undeclare (x)
    }
    attributes (x) <- NULL

    if (isTRUE (attrx$date)) {
        if (is.null (attrx$na_index) & !is.null(labels)) {
            # in Date objects, only the missing values can possibly have labels
            attrx$na_index <- rep(1, length(labels)) # positions don't matter
            names (attrx$na_index) <- labels # the labels do matter
        }
        missing <- is.element (allabels, names(attrx$na_index))
        nms <- names (allabels)
        if (any(!missing)) {
            nms[!missing] <- as.character (as.Date (as.numeric (nms[!missing])))
        }
        allabels[!missing] <- as.character (as.Date (allabels[!missing]))
        names (allabels) <- nms

        xmissing <- is.element (x, allabels[missing])
        x[!xmissing] <- as.character (as.Date (x[!xmissing]))
    }

    if (isTRUE (values)) {
        return (as.character (x))
    }

    x <- names (allabels)[match (x, allabels)]

    if (isTRUE (nolabels)) {
        x[!is.element (x, names (labels))] <- NA
    }

    return (x)
}

#' @export
`[.declared` <- function (x, i, ...) {
  x <- sanitize_na_index_ (x)
  attrx <- attributes (x)
  x <- undeclare (x)
  x <- NextMethod()
  # attrx$label, if not existing, takes from attrx$labels
  # attrx[["label"]] is something like attr (x, "label", exact = TRUE)
  declared (
    x, attrx[["labels"]], attrx$na_values, attrx$na_range, attrx[["label"]],
    decimals = attrx$decimals
  )
}

#' @export
`[<-.declared` <- function (x, i, value) {
  x <- sanitize_na_index_ (x)
  attrx <- attributes (x)
  value <- undeclare (value)
  x <- undeclare (x)
  xdate <- isTRUE (attrx$date)
  attributes (x) <- NULL
  x <- NextMethod()
  if (xdate) {
    x <- as.Date (x)
  }
  declared (
    x, attrx[["labels"]], attrx$na_values, attrx$na_range, attrx[["label"]],
    decimals = attrx$decimals
  )
}

#' @export
`[.labels_df` <- function (x, i, ...) {
  result <- NextMethod()
  attr(result, "print_as_df") <- attr(x, "print_as_df")
  class(result) <- class(x)
  return(result)
}

#' @export
`c.declared` <- function (...) {
  dots <- lapply (list (...), sanitize_na_index_)
  declared <- unlist (lapply (dots, is.declared))
  na_values <- sort (unique (unlist (
    lapply (dots, function (x) attr (x, "na_values"))
  )))

  labels <- unlist (lapply (dots, function (x) {
    attr (x, "labels", exact = TRUE)
  }))

  duplicates <- duplicated (labels)

  if (length (wduplicates <- which (duplicates)) > 0) {
    for (i in wduplicates) {
      if (length (unique (names (labels[labels == labels[i]]))) > 1) {
        stopError_ ("Labels must be unique.")
      }
    }
  }

  labels <- sort (labels[!duplicates])

  na_range <- lapply (dots, function (x) attr (x, "na_range", exact = TRUE))
  nulls <- unlist (lapply (na_range, is.null))

  decimals <- unique (unlist (lapply (dots, function (x) {
    attr (x, "decimals", exact = TRUE)
  })))
  if (length (decimals) == 0) {
    decimals <- NULL
  }
  else {
    decimals <- max (decimals)
  }

  if (all (nulls)) {
    na_range <- NULL
  }
  else {
    if (sum (nulls) == length (na_range) - 1) {
      na_range <- unlist (na_range)
    }
    else {
      compatible <- logical (length (na_range))
      if (!is.null (na_range)) {
        for (i in seq (1, length (na_range) - 1)) {
          nai <- na_range[[i]]
          if (is.null (nai)) {
            compatible[i] <- TRUE
          }
          else {
            for (j in seq (2, length (na_range))) {
              naj <- na_range[[j]]
              if (is.null (naj)) {
                compatible[j] <- TRUE
              }
              else {
                if (
                  any (
                    is.element (seq (nai[1], nai[2]), seq (naj[1], naj[2]))
                  ) > 0
                ) {
                  compatible[i] <- TRUE
                  compatible[j] <- TRUE
                }
              }
            }
          }
        }
      }

      if (any (!compatible)) {
        stopError_ ("Incompatible NA ranges.")
      }

      na_range <- range(unlist (na_range))
    }
  }

  dots <- unlist (lapply (dots, function (x) {
    if (is.declared (x)) x <- undeclare (x)
    attributes (x) <- NULL
    return (x)
  }))

  return (declared (
    dots,
    labels = labels,
    na_values = na_values,
    na_range = na_range,
    label = attr (dots[[which (declared)[1]]], "label", exact = TRUE),
    decimals = decimals
  ))
}

#' @export
`names<-.declared` <- function (x, value) {
  attr (x, "names") <- value
  x
}

#' @export
`sort.declared` <- function (x, decreasing = FALSE, ...) {
  return (x[order_declared (x, decreasing = decreasing, ... = ...)])
}

#' @export
`duplicated.declared` <- function (x, incomparables = FALSE, ...) {
  x <- unclass (undeclare (x))
  NextMethod()
}

#' @export
`unique.declared` <- function (x, incomparables = FALSE, ...) {
  x[!duplicated (x)]
}

#' @export
`head.declared` <- function (x, n = 6L, ...) {
  lx <- length (x)
  if (n < 0) {
    n <- lx - abs (n)
  }
  n <- min (n, lx)
  if (n < 1) {
    return (x[0])
  }
  x[seq (n)]
}

#' @export
`head.labels_df` <- function (x, n = 6L, ...) {
  lx <- length (x)
  if (n < 0) {
    n <- lx - abs (n)
  }
  n <- min (n, lx)
  if (n < 1) {
    result <- x[0]
  } else {
    result <- x[seq (n)]
  }
  attr(result, "print_as_df") <- attr(x, "print_as_df")
  class(result) <- class(x)
  return(result)
}

#' @export
`tail.declared` <- function (x, n = 6L, ...) {
  if (n < 1) {
    n <- 6L
  }
  lx <- length (x)
  n <- min (n, lx)
  x[seq (lx - n + 1, lx)]
}

#' @export
`tail.labels_df` <- function (x, n = 6L, ...) {
  if (n < 1) {
    n <- 6L
  }
  lx <- length (x)
  n <- min (n, lx)
  result <- x[seq (lx - n + 1, lx)]

  attr(result, "print_as_df") <- attr(x, "print_as_df")
  class(result) <- class(x)
  return(result)
}

#' @export
`na.omit.declared` <- function (object, ...)  {
  object <- sanitize_na_index_ (object)
  attrx <- attributes (object)
  attrx$na_index <- NULL
  object <- unclass (object)
  object <- NextMethod()
  attrx$na.action <- attr (object, "na.action")
  nms <- attrx$names
  if (!is.null (nms) && !is.null (attrx$na.action)) {
    nms <- nms[-attr (object, "na.action")]
    attrx$names <- nms
  }
  attributes (object) <- attrx
  return (object)
}

#' @export
`na.fail.declared` <- function (object, ...)  {
  object <- sanitize_na_index_ (object)
  object <- unclass (object)
  if (isTRUE (attr (object, "date"))) {
    attributes (object) <- NULL
    object <- as.Date (object)
  }
  NextMethod()
}

#' @export
`na.exclude.declared` <- function (object, ...)  {
  object <- sanitize_na_index_ (object)
  attrx <- attributes (object)
  attrx$na_index <- NULL
  object <- unclass (object)
  object <- NextMethod()
  attrx$na.action <- attr (object, "na.action")
  nms <- attrx$names
  if (!is.null (nms) && !is.null (attrx$na.action)) {
    nms <- nms[-attr (object, "na.action")]
    attrx$names <- nms
  }
  attributes (object) <- attrx
  return (object)
}

#' @export
`mean.declared` <- function (x, ...) {
  x <- sanitize_na_index_ (x)
  xdate <- isTRUE (attr (x, "date"))
  na_index <- attr (x, "na_index")
  if (!is.null (na_index)) {
    x <- x[-na_index]
  }
  x <- unclass (x)
  if (xdate) {
    attributes (x) <- NULL
    x <- as.Date (x)
  }
  NextMethod()
}


# DO NOT use backticks for this function name
# it might break the roxygen2 documentation generating process

#' @method weighted.mean declared
#' @importFrom stats weighted.mean
#' @export
weighted.mean.declared <- function (x, w, ..., na.rm = FALSE) {
  x <- sanitize_na_index_ (x)
  xdate <- isTRUE (attr (x, "date"))
  na_index <- attr (x, "na_index")
  if (!is.null (na_index)) {
    x <- x[-na_index]
  }
  x <- unclass (x)
  if (xdate) {
    attributes (x) <- NULL
    x <- as.Date (x)
  }
  NextMethod()
}

#' @export
`median.declared` <- function (x, na.rm = FALSE, ...) {
  x <- sanitize_na_index_ (x)
  xdate <- isTRUE (attr (x, "date"))
  na_index <- attr (x, "na_index")
  if (!is.null (na_index)) {
    x <- x[-na_index]
  }
  x <- unclass (x)
  if (xdate) {
    attributes (x) <- NULL
    x <- as.Date (x)
  }
  NextMethod()
}

#' @export
`summary.declared` <- function (object, ...) {
  object <- unclass (object)
  if (isTRUE (attr (object, "date"))) {
    attributes (object) <- NULL
    object <- as.Date (object)
  }
  NextMethod()
}

#' @method all.equal declared
#' @export
`all.equal.declared` <- function (target, current, ...) {
  target <- sanitize_na_index_ (target)
  na_index <- attr (target, "na_index")
  target <- undeclare (target, drop = TRUE)
  if (is.declared (current)) {
    current <- undeclare (current, drop = TRUE)
  }

  allna <- TRUE

  if (!is.null (na_index)) {
    allna <- all.equal(target[na_index], current[na_index])
    target <- target[-na_index]
    current <- current[-na_index]
  }

  allv <- all.equal(target, current)

  if (isTRUE (allv)) {
    if (isTRUE (allna)) {
      return (TRUE)
    }
    return (paste ("Declared mising values", tolower(allna)))
  }

  return (allv)
}


# Math operations

#' @export
`abs.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("abs")(x)
}

#' @export
`sign.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("sign")(x)
}

#' @export
`sqrt.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("sqrt")(x)
}

#' @export
`floor.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("floor")(x)
}

#' @export
`ceiling.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("ceiling")(x)
}

#' @export
`trunc.declared` <- function (x, ...) {
  x <- check_date (x)
  .Primitive ("trunc")(x, ...)
}

#' @export
`round.declared` <- function (x, digits = 0) {
  x <- check_date (x)
  .Primitive ("round")(x, digits)
}

#' @export
`signif.declared` <- function (x, digits = 0) {
  x <- check_date (x)
  .Primitive ("signif")(x, digits)
}

#' @export
`exp.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("exp")(x)
}

#' @export
`log.declared` <- function (x, base = exp(1)) {
  x <- check_date (x)
  .Primitive ("log")(x, base)
}

#' @export
`expm1.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("expm1")(x)
}

#' @export
`log1p.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("log1p")(x)
}

#' @export
`cos.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("cos")(x)
}

#' @export
`sin.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("sin")(x)
}

#' @export
`tan.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("tan")(x)
}

#' @export
`cospi.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("cospi")(x)
}

#' @export
`sinpi.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("sinpi")(x)
}

#' @export
`tanpi.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("tanpi")(x)
}

#' @export
`acos.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("acos")(x)
}

#' @export
`asin.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("asin")(x)
}

#' @export
`atan.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("atan")(x)
}

#' @export
`lgamma.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("lgamma")(x)
}

#' @export
`gamma.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("gamma")(x)
}

#' @export
`digamma.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("digamma")(x)
}

#' @export
`trigamma.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("trigamma")(x)
}

#' @export
`cumsum.declared` <- function (x) {
  x <- sanitize_na_index_ (x)
  na_index <- attr (x, "na_index")
  x <- check_date (x)
  if (is.null (na_index)) {
    return(cumsum (x))
  }
  x[-na_index] <- cumsum (x[-na_index])
  return(x)
}

#' @export
`cumprod.declared` <- function (x) {
  x <- sanitize_na_index_ (x)
  na_index <- attr (x, "na_index")
  x <- check_date (x)
  if (is.null (na_index)) {
    return(cumprod (x))
  }
  x[-na_index] <- cumprod (x[-na_index])
  return(x)
}

#' @export
`cummax.declared` <- function (x) {
  x <- sanitize_na_index_ (x)
  na_index <- attr (x, "na_index")
  x <- check_date (x)
  if (is.null (na_index)) {
    return(cummax (x))
  }
  x[-na_index] <- cummax (x[-na_index])
  return(x)
}

#' @export
`cummin.declared` <- function (x) {
  x <- sanitize_na_index_ (x)
  na_index <- attr (x, "na_index")
  x <- check_date (x)
  if (is.null (na_index)) {
    return(cummin (x))
  }
  x[-na_index] <- cummin (x[-na_index])
  return(x)
}



# Arithmetic operations
#' @export
`+.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("+")(e1, e2)
}

#' @export
`-.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("-")(e1, e2)
}

#' @export
`*.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("*")(e1, e2)
}

#' @export
`/.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("/")(e1, e2)
}

#' @export
`^.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("^")(e1, e2)
}

#' @export
`%%.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("%%")(e1, e2)
}

#' @export
`%/%.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("%/%")(e1, e2)
}

#' @export
`&.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("&")(e1, e2)
}

#' @export
`|.declared` <- function (e1, e2) {
  e1 <- check_date (e1)
  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- check_date (e2)
    }
  }
  .Primitive ("|")(e1, e2)
}

#' @export
`!.declared` <- function (x) {
  x <- check_date (x)
  .Primitive ("!")(x)
}

#' @export
`==.declared` <- function (e1, e2) {
  le1 <- attr (e1, "labels", exact = TRUE)
  e1 <- undeclare (e1, drop = TRUE)

  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- undeclare (e2, drop = TRUE)
    }

    if (
      length (e2) == 1 && is.element (e2, names (le1)) && !is.element (e2, e1)
    ) {
      e2 <- le1[names (le1) == e2]
    }
  }

  .Primitive ("==")(e1, e2)
}

#' @export
`!=.declared` <- function (e1, e2) {e1 <- unclass (undeclare (e1))
le1 <- attr (e1, "labels", exact = TRUE)
e1 <- undeclare (e1)
attributes (e1) <- NULL

if (!missing(e2)) {
  if (is.declared (e2)) {
    e2 <- undeclare (e2)
    attributes (e2) <- NULL
  }

  if (
    length (e2) == 1 && is.element (e2, names (le1)) && !is.element (e2, e1)
  ) {
    e2 <- le1[names (le1) == e2]
  }
}


.Primitive ("!=")(e1, e2)
}

#' @export
`<.declared` <- function (e1, e2) {
  e1 <- undeclare (e1)
  attributes (e1) <- NULL

  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- undeclare (e2)
      attributes (e2) <- NULL
    }
  }

  .Primitive ("<")(e1, e2)
}

#' @export
`<=.declared` <- function (e1, e2) {
  e1 <- undeclare (e1)
  attributes (e1) <- NULL

  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- undeclare (e2)
      attributes (e2) <- NULL
    }
  }

  .Primitive ("<=")(e1, e2)
}

#' @export
`>=.declared` <- function (e1, e2) {
  e1 <- undeclare (e1)
  attributes (e1) <- NULL

  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- undeclare (e2)
      attributes (e2) <- NULL
    }
  }

  .Primitive (">=")(e1, e2)
}

#' @export
`>.declared` <- function (e1, e2) {
  e1 <- undeclare (e1)
  attributes (e1) <- NULL

  if (!missing(e2)) {
    if (is.declared (e2)) {
      e2 <- undeclare (e2)
      attributes (e2) <- NULL
    }
  }

  .Primitive (">")(e1, e2)
}

#' @export
`Arg.declared` <- function (z) {
  z <- check_date (z)
  .Primitive ("Arg")(z)
}

#' @export
`Conj.declared` <- function (z) {
  z <- check_date (z)
  .Primitive ("Conj")(z)
}

#' @export
`Im.declared` <- function (z) {
  z <- check_date (z)
  .Primitive ("Im")(z)
}

#' @export
`Mod.declared` <- function (z) {
  z <- check_date (z)
  .Primitive ("Mod")(z)
}

#' @export
`Re.declared` <- function (z) {
  z <- check_date (z)
  .Primitive ("Re")(z)
}

#' @export
`format.declared` <- function (x, ..., justify = "none") {
  if (any (is.na (x)) || !is.null (attr (x, "decimals", exact = TRUE))) {
    dots <- list (...)
    digits <- if (!is.null (dots$digits)) dots$digits else getOption ("digits")
    return (format_declared (x, digits = digits))
  }

  class (x) <- setdiff (class (x), "declared")
  return (format (x, ..., justify = justify))
}

#' @export
`Summary.declared` <- function (..., na.rm = FALSE) {
  dots <- lapply (list (...), function (x) {
    if (inherits (x, "declared")) {
      x <- sanitize_na_index_ (x)
      na_index <- attr (x, "na_index")
      if (!is.null (na_index)) {
        x <- x[-na_index]
      }
      xdate <- isTRUE (attr (x, "date"))
      attributes (x) <- NULL
      if (xdate) {
        x <- as.Date (x)
      }
    }
    return (x)
  })

  return (do.call (.Generic, c (dots, na.rm = na.rm)))
}


#' @export
`anyDuplicated.declared` <- function (x, incomparables = FALSE, ...) {
  x <- unclass (undeclare (x))
  return (anyDuplicated (x, incomparables = incomparables, ...))
}

#' @export
`rep.declared` <- function (x, ...) {
  x <- sanitize_na_index_ (x)
  attrx <- attributes (x)
  rep_val <- rep (undeclare (x, drop = TRUE), ...)
  return (
    declared (
      rep_val,
      labels = attrx$labels,
      na_values = attrx$na_values,
      na_range = attrx$na_range,
      label = attrx[["label"]],
      decimals = attrx$decimals
    )
  )
}

#' @export
`as.Date.declared` <- function (x, ...) {
  x <- sanitize_na_index_ (x)
  attrx <- attributes (x)
  date_vals <- as.Date (undeclare (x, drop = TRUE), ...)
  return (
    declared (
      date_vals,
      labels = attrx$labels,
      na_values = attrx$na_values,
      na_range = attrx$na_range,
      label = attrx[["label"]],
      decimals = attrx$decimals
    )
  )
}

#' @export
`xtfrm.declared` <- function (x) {
  return (xtfrm_declared (x))
}

#' @export
`cosh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("cosh")(x))
}

#' @export
`sinh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("sinh")(x))
}

#' @export
`tanh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("tanh")(x))
}

#' @export
`acosh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("acosh")(x))
}

#' @export
`asinh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("asinh")(x))
}

#' @export
`atanh.declared` <- function (x) {
  x <- check_date (x)
  return (.Primitive ("atanh")(x))
}


# TODO:
# cut() ?
# diff() ?

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @name measurement
#' @family labelling functions
#' @title Get / Set measurement levels for declared objects
#'
#' @description
#' Functions to extract information about the measurement levels of a variable
#' (if already present), or to specify such measurement levels.
#'
#' @details
#' This function creates an attribute called `"measurement"` to a declared
#' object. This is an optional feature, at this point for purely aesthetic
#' reasons, but it might become useful in the future to (automatically)
#' determine if a declared object is suitable for a certain statistical
#' analysis, for instance regression requires quantitative variables, while some
#' declared objects are certainly categorical despite using numbers to denote
#' categories.
#'
#' It distinguishes between `"categorical"` and `"quantitative"` types of
#' variables, and additionally recognizes `"nominal"` and `"ordinal"` as
#' categorical, and similarly recognizes `"interval"`, `"ratio"`,
#' `"discrete"` and `"continuous"` as quantitative.
#'
#' The words `"qualitative"` is treated as a synonym for `"categorical"`,
#' and the words `"metric"` and `"numeric"` are treated as synonyms for
#' `"quantitative"`, respectively.
#'
#' @return
#' A character vector.
#'
#' @examples
#' x <- declared(
#'     c(-2, 1:5, -1),
#'     labels = c(Good = 1, Bad = 5, DK = -1),
#'     na_values = c(-1, -2),
#'     label = "Test variable"
#' )
#'
#' x
#'
#' measurement(x)
#'
#' # automatically recognized as categorical
#' measurement(x) <- "ordinal"
#'
#' measurement(x)
#'
#' # the same with
#' measurement(x) <- "categorical, ordinal"
#'
#' set.seed(1890)
#' x <- declared(
#'     sample(c(18:90, -91), 20, replace = TRUE),
#'     labels = c("No answer" = -91),
#'     na_values = -91,
#'     label = "Respondent's age"
#' )
#'
#' # automatically recognized as quantitative
#' measurement(x) <- "discrete"
#'
#' measurement(x)
#'
#' # the same with
#' measurement(x) <- "metric, discrete"
#'
#' @param x A declared vector.
#' @export
`measurement` <- function (x) {
    UseMethod ("measurement")
}


#' @export
`measurement.default` <- function (x) {
    # do nothing
    NULL
}


#' @export
`measurement.declared` <- function (x) {
    m <- attr (x, "measurement")
    if (is.null (m)) {
        m <- "Unspecified"
        l_m <- likely_measurement(x)
        if (!identical (l_m, "")) {
            m <- paste0 (m, ", but likely ", paste (l_m, collapse = " "))
        }
    }

    return (m)
}


#' @rdname measurement
#' @param value A single character string of measurement levels,
#' separated by commas.
#' @export
`measurement<-` <- function (x, value) {
    UseMethod ("measurement<-")
}


#' @export
`measurement<-.default` <- function (x, value) {
    # do nothing
    x
}


#' @export
`measurement<-.declared` <- function (x, value) {
    attr (x, "measurement") <- check_measurement (value)
    return (x)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 7
#' @param probs Numeric vector of probabilities with values in \[0,1\]
#' @details
#' The function `wquantile()` extensively borrowed ideas from packages
#' **`stats`** and **`Hmisc`**, to ensure a constant interpolation that would
#' produce the same quantiles if no weights are provided or if all weights are
#' equal to 1.
#'
#' Other arguments can be passed to the stats function `quantile()` via the
#' three dots `...` argument, and their extensive explanation is found in the
#' corresponding stats function's help page.
#' @export
`wquantile` <- function (
    x, wt = NULL, probs = seq (0, 1, 0.25), na.rm = TRUE, ...
) {

    metacall <- as.list (match.call ())

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (is.null (x) || !is.atomic (x) || !is.numeric (x)) {
        stopError_ ("'x' should be an atomic numerical vector.")
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (length (na_index)) {
            x <- x[-na_index]
            wt <- wt[-na_index] # if wt is NULL, the result is still NULL
        }
        attributes (x) <- NULL
    }

    if (is.null (wt)) {
        qs <- quantile (x, probs = probs, na.rm = na.rm, ... = ...)
        class (qs) <- c ("fobject", class (qs))
        return (qs)
    }

    if (
        !is.null (wt) && !(
            is.atomic (wt) && all (is.finite (na.omit (wt)))
        )
    ) {
        stopError_ ("'wt' should be an atomic vector with finite values.")
    }

    if (length (x) != length (wt)) {
        stopError_ ("Lengths of 'x' and 'wt' differ.")
    }

    ok <- !is.na (x + wt)

    if (na.rm) {
        x <- x[ok]
        wt <- wt[ok]
    }
    else if (any (!ok)) {
        stopError_ ("Missing values and NaN's not allowed if `na.rm' is FALSE")
    }

    if (any ((p.ok <- !is.na (probs)) & (probs < 0 | probs > 1))) {
        stopError_ ("probs outside [0,1]")
    }

    if (na.p <- any (!p.ok)) {
        o.pr <- probs
        probs <- probs[p.ok]
        probs <- pmax (0, pmin (1, probs))
    }

    sumwt <- sum (wt)
    wt <- cumsum (tapply (wt, x, sum, na.rm = TRUE))
    x <- sort (unique (x))

    index <- 1 + (sumwt - 1) * probs

    left <- pmax (floor (index), 1)
    right <- pmin (left + 1, sumwt)

    index <- index%%1

    both <- approx(
        wt, x, xout = c (left, right),
        method = "constant", rule = 2, f = 1
    )$y
    lp <- length (probs)
    qs <- (1 - index) * both[seq (1, lp)] + index * both[-seq (1, lp)]

    qs <- coerceMode_ (qs)

    names (qs) <- paste0 (format (100 * probs, trim = TRUE), "%")

    class (qs) <- c ("fobject", class (qs))
    return (qs)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_quantile` <- function (...) {
    wquantile(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @name drop_undeclare
#' @title Drop information / undeclare labelled objects
#' @description
#' A function to obtain a version of the object with all information about
#' declared missing values, dropped
#' @details
#' The function `undeclare()` replaces the NA entries into their original
#' numeric values, and drops all attributes related to missing values:
#' `na_values`, `na_range` and `na_index`, and it preserves the labels referring
#' to the missing values.
#'
#' The result can be a regular vector (dropping all attributes, including the
#' class "declared") by activating the argument `drop`.
#'
#' Function `drop_na()` transforms the declared missing values in regular empty
#' NAs, and the labels referring to the missing values are deleted by default.
#'
#' Function `drop()` deletes all attributes.
#' @family labelling functions
#' @return A declared labelled object.
#' @examples
#' x <- declared(
#'     c(-2, 1:5, -1),
#'     labels = c("Good" = 1, "Bad" = 5, "DK" = -1),
#'     na_values = c(-1, -2),
#'     label = "Test variable"
#' )
#'
#' x
#'
#' undeclare(x)
#'
#' drop_na(x)
#'
#' drop(x)
#'
#' undeclare(x, drop = TRUE)
#'
#' # similar to:
#' drop(undeclare(x))
#' @param x A labelled object with declared missing values
#' @param drop Logical, drop all attributes
#' @param drop_labels Logical, drop the labels for the declared missing values
#' @param ... Other internal arguments
#' @export
`undeclare` <- function (x, drop = FALSE, ...) {
  UseMethod ("undeclare")
}


#' @export
`undeclare.default` <- function (x, drop = FALSE, ...) {
  # do nothing
  return (x)
}


#' @export
`undeclare.declared` <- function (x, drop = FALSE, ...) {
  x <- sanitize_na_index_ (x)
  na_index <- attr (x, "na_index")
  attrx <- attributes (x)

  # this is necessary to replace those values
  # (because of the "[<-.declared" method)
  attributes (x) <- NULL # or x <- unclass (x), but I find this cleaner

  if (!is.null (na_index)) {
    values <- names (na_index)

    if (possibleNumeric_ (values)) {
      if (is.integer (x)) {
        values <- as.integer (values)
      } else if (is.numeric (x)) {
        values <- as.numeric (values)
      }
    } else {
      x <- as.character (x)
    }

    x[na_index] <- values
  }

  attrx$na_index <- NULL
  attrx$na_values <- NULL
  attrx$na_range <- NULL

  if (isFALSE (drop)) {
    attrx$class <- c("declared", class(x))
    attributes (x) <- attrx
  }

  return (x)
}


#' @export
`undeclare.data.frame` <- function (x, drop = FALSE, ...) {
  declared <- vapply (x, is.declared, logical (1))
  x[declared] <- lapply (x[declared], undeclare, drop = drop)

  return (x)
}


#' @rdname drop_undeclare
#' @export
`drop_na` <- function (x, drop_labels = TRUE) {
  UseMethod ("drop_na")
}


#' @export
`drop_na.default` <- function (x, drop_labels = TRUE) {
  # do nothing
  return(x)
}


#' @export
`drop_na.haven_labelled_spss` <- function (x, drop_labels = TRUE) {
  attrx <- attributes (x)
  x[is.element (x, attrx$na_values)] <- NA
  if (isTRUE (drop_labels)) {
    attrx$labels <- attrx$labels[!is.element (attrx$labels, attrx$na_values)]
  }
  attrx$na_index <- NULL
  attrx$na_values <- NULL
  attrx$na_range <- NULL
  attributes (x) <- attrx
  return (x)
}


#' @export
`drop_na.declared` <- function (x, drop_labels = TRUE) {
  x <- sanitize_na_index_ (x)
  attrx <- attributes (x)
  if (isTRUE (drop_labels)) {
    attrx$labels <- attrx$labels[!is.element (attrx$labels, attrx$na_values)]
  }
  attrx$na_index <- NULL
  attrx$na_values <- NULL
  attrx$na_range <- NULL
  attributes (x) <- attrx
  return (x)
}


#' @export
`drop_na.data.frame` <- function (x, drop_labels = TRUE) {
  declared <- vapply (x, is.declared, logical (1))
  x[declared] <- lapply (x[declared], drop_na, drop_labels = drop_labels)
  return (x)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 3
#' @param ... Further arguments passed to or from other methods.
#' @export
`wmedian` <- function (x, wt = NULL, na.rm = TRUE, ...)
{
    metacall <- as.list (match.call ())

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (length (na_index)) {
            x <- x[-na_index]
            wt <- wt[-na_index] # if wt is NULL, the result is still NULL
        }
        attributes (x) <- NULL # if passed to wquantile, to bypass this process
    }

    if (is.null (wt)) {
        return (median (x, na.rm = na.rm, ... = ...))
    }

    x <- wquantile (
        x, wt = wt, probs = 0.5, na.rm = na.rm, ... = ...
    )

    attributes (x) <- NULL

    return (x)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_median` <- function (...) {
    wmedian(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 4
#' @details
#' The function `wmode()` returns the weighted mode of a variable. Unlike the
#' other functions where the prefix `w` signals a weighted version of the
#' base function with the same name, this has nothing to do with the base
#' function `mode()` which refers to the storage mode or type of an R object.
#' @export
`wmode` <- function (x, wt = NULL) {

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (
        is.null (x) || !is.atomic (x) || !(
            is.numeric (x) || is.complex (x) || is.logical (x)
        )
    ) {
        warning ("'x' should be a numerical / logical vector: returning NA")
        return (NA_real_)
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (length (na_index)) {
            x <- x[-na_index]
            wt <- wt[-na_index] # if wt is NULL, the result is still NULL
        }
        attributes (x) <- NULL
    }

    if (
        !is.null (wt) && !(
            is.atomic (wt) && all (is.finite (na.omit (wt)))
        )
    ) {
        stopError_ ("'wt' should be an atomic vector with finite values.")
    }

    tbl <- wtable (x, wt = wt)
    wm <- which.max (tbl)

    fmode <- names (tbl[wm])
    if (possibleNumeric_ (fmode)) {
        fmode <- asNumeric_ (fmode)
    }

    if (length (tbl[tbl == max (tbl)]) > 1) {
        message ("Multiple modes detected, only the first is returned.")
    }

    return (fmode)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_mode` <- function (...) {
    wmode(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @title
#' Test the presence of empty (undeclared) missing values
#'
#' @description
#' Functions that indicate which elements are empty `NA` missing values, in
#' contrast to declared missing values.
#'
#' @details
#' All missing values, declared or undeclared, as stored as regular `NA`
#' values, therefore the base function `is_na()` does not differentiate
#' between them.
#'
#' These functions are specifically adapted to objects of class `"declared"`,
#' to return a truth value only for those elements that are completely missing
#' with no reason.
#'
#' @return
#' A logical vector.
#'
#' @examples
#'
#' x <- declared(
#'     c(1:2, -91),
#'     labels = c(Good = 1, Bad = 2, Missing = -91),
#'     na_values = -91
#' )
#'
#' x
#'
#' is.empty(x) # FALSE FALSE FALSE
#'
#' anyNAempty(x) # FALSE
#'
#' x <- c(x, NA)
#'
#' is.empty(x) # FALSE FALSE FALSE  TRUE
#'
#' anyNAempty(x) # TRUE
#'
#' @param x A vector
#'
#' @export
`is.empty` <- function (x) {
  if (is.null (x) || !is.atomic (x)) {
    stopError_ ("'x' should be an atomic vector.")
  }

  if (is.declared (x)) {
    x <- sanitize_na_index_ (x)
  }

  empty <- is.na (x)

  if (is.declared (x)) {
    na_index <- attr (x, "na_index")
    if (!is.null (na_index)) {
      empty[na_index] <- FALSE
    }
  }

  return (empty)
}

#' @rdname is.empty
#' @export
`anyNAempty` <- function (x) {
  return (anyNA (undeclare (x)))
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 9
#' @export
`wfivenum` <- function (
    x, wt = NULL, na.rm = FALSE
) {
    metacall <- as.list (match.call ())

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (
        is.null (x) || !(
            is.atomic (x) && is.numeric (x)
        )
    ) {
        stopError_ ("'x' should be an atomic numerical vector.")
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (length (na_index)) {
            x <- x[-na_index]
            wt <- wt[-na_index] # if wt is NULL, the result is still NULL
        }
        attributes (x) <- NULL
    }

    if (is.null(wt)) {
        fvn <- fivenum (x, na.rm = na.rm)
        names (fvn) <- c ("Min", "Q1", "Q2", "Q3", "Max")
        class (fvn) <- c ("fobject", class (fvn))
        return (fvn)
    }

    if (
        !is.null (wt) && !(
            is.atomic (wt) && all (is.finite (na.omit (wt)))
        )
    ) {
        stopError_ ("'wt' should be an atomic vector with finite values.")
    }

    if (length (x) != length (wt)) {
        stopError_ ("Lengths of 'x' and 'wt' differ.")
    }

    ok <- !is.na (x + wt)

    if (na.rm) {
        x <- x[ok]
        wt <- wt[ok]
    }
    else if (any (!ok)) {
        stopError_ ("Missing values and NaN's not allowed if `na.rm' is FALSE")
    }

    sumwt <- sum (wt)

    if (any (wt < 0) || sumwt == 0) {
        stopError_ ("'wt' must be non-negative and not all zero")
    }

    fvn <- fivenum (x * wt, na.rm = na.rm)
    names (fvn) <- c ("Min", "Q1", "Q2", "Q3", "Max")
    class (fvn) <- c ("fobject", class (fvn))
    return (fvn)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_fivenum` <- function (...) {
    wfivenum(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @title declared internal functions
#' @description Functions to be used internally, only by developers and
#' contributors.
#' @name declared_internal
NULL

#' @rdname declared_internal
#' @keywords internal
#' @export
`format_declared` <- function (x, digits = getOption ("digits")) {
    if (is.null (x) || !is.atomic (x)) {
        stopError_ ("`x` has to be a vector.")
    }

    x <- sanitize_na_index_ (x)

    if (inherits (x, "Date") || isTRUE (attr (x, "date"))) {
        class (x) <- "Date"
        out <- as.character (x)
    } else if (is.numeric (x) && !is.null (attr (x, "decimals"))) {
        out <- format_decimals_ (unclass (x), attr (x, "decimals"))
    } else {
        out <- format (unclass (x), digits = digits)
    }

    na_index <- attr (x, "na_index")

    if (!is.null (na_index)) {
        out[na_index] <- paste0 ("NA(", names (na_index), ")")
    }

    # format again to make sure all elements have same width
    return (format (out, justify = "right"))
}

`format_decimals_` <- function (x, decimals) {
    decimals <- check_decimals (decimals)
    out <- formatC (x, digits = decimals, format = "f")
    if (decimals > 0) {
        out <- sub ("\\.?0+$", "", out)
    }
    out[out == "-0"] <- "0"
    return (out)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`order_declared` <- function (
    x, na.last = NA, decreasing = FALSE, method = c ("auto", "shell", "radix"),
    empty.last = na.last, ...) {

    if (!is.declared (x)) {
        stopError_ ("`x` has to be a vector of class `declared`.")
    }

    x <- sanitize_na_index_ (x)

    if (!identical (empty.last, NA)) {
        if (!(isTRUE (empty.last) | isFALSE (empty.last))) {
        stopError_ ("Argument `empty.last` should be either TRUE or FALSE.")
        }
    }

    method <- match.arg (method)

    x_indexes <- seq_along(x)

    na_index <- attr (x, "na_index")
    na_declared <- logical (length (x))
    na_declared[na_index] <- TRUE
    na_empty <- is.empty (x)

    declared_indexes <- c ()

    if (any (na_declared)) {
        x <- undeclare (x)
        nms <- names (na_index)
        if (possibleNumeric_ (nms)) {
        nms <- asNumeric_ (nms)
        }
        declared_indexes <- unname (
            na_index[order (nms, decreasing = decreasing, method = method)]
        )
    }

    attributes (x) <- NULL
    x_indexes <- x_indexes[!(is.na (x) | na_declared)]
    x <- x[!(is.na (x) | na_declared)]

    res <- c ()

    if (isFALSE (na.last)) {
        if (isFALSE (empty.last)) {
        res <- c (which (na_empty), declared_indexes)
        }

        if (isTRUE (empty.last)) {
        res <- c (declared_indexes, which (na_empty))
        }
    }


    res <- c (
        res,
        x_indexes[order (unclass (x), decreasing = decreasing, method = method)]
    )


    if (isTRUE (na.last)) {
        if (isTRUE (empty.last)) {
        res <- c (res, declared_indexes, which (na_empty))
        }

        if (isFALSE (empty.last)) {
        res <- c (res, which (na_empty), declared_indexes)
        }
    }

    return (res)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`value_labels` <- function (...) {
    .Deprecated(msg = "Function value_labels() is deprecated, use labels()\n")
    labels (...)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`variable_label` <- function (...) {
    .Deprecated(msg = "Function variable_label() is deprecated, use label()\n")
    label(...)
}


`likely_type` <- function (x) {
    type <- NULL
    others <- setdiff (class (x), "declared")

    if (length (others) > 0) {
        type <- others[1]
    }

    if (identical(type, "numeric")) {
        if (!anyTagged_ (x) && (is.integer (x) || wholeNumeric_ (x))) {
            type <- "integer"
        }
    }

    ## all declared objects have a second class
    # if (!is.null (type)) {
        return (paste0 ("<", type, ">"))
    # }

    # return (type)
}


`check_measurement` <- function (x) {

    if (is.null (x)) {
        return (x)
    }

    x <- trimstr_ (tolower(unlist (strsplit (x, split = ","))))

    mlevels <- c (
        "categorical", "nominal", "ordinal",
        "quantitative", "interval", "ratio", "discrete", "continuous"
    )

    if (any (wx <- x == "qualitative")) {
        x[wx] <- "categorical"
    }

    if (any (wx <- x == "metric")) {
        x[wx] <- "quantitative"
    }

    if (any (wx <- x == "numeric")) {
        x[wx] <- "quantitative"
    }

    position <- pmatch (x, mlevels)

    if (any (is.na (position))) {
        stopError_ ("Unknown measurement level.")
    }

    position <- sort (position)

    first <- unique (ifelse (position < 4, 1, 4))

    if (length (first) > 1) {
        stopError_ (
            paste (
                "Measurement can not be categorical",
                "and quantitative at the same time."
            )
        )
    }
    cpos <- setdiff (position, first)

    if (length (cpos) > 1) {

        if (first == 4) { # quantitative
            # check for combinations of interval / ratio and
            # discrete / continuous
            if (any (cpos < 7)) {
                ir <- cpos[cpos < 7]
                if (length (ir) > 1) {
                stopError_ (sprintf (
                    "Measurement can not be both %s and %s at the same time.",
                    mlevels[ir[1]],
                    mlevels[ir[2]]
                ))
                }
            }

            if (any (cpos > 6)) {
                dc <- cpos[cpos > 6]
                if (length (dc) > 1) {
                stopError_ (sprintf (
                    "Measurement can not be both %s and %s at the same time.",
                    mlevels[dc[1]],
                    mlevels[dc[2]]
                ))
                }
            }
        }
        else {
            stopError_ (
                sprintf (
                    "Measurement can not be both %s and %s at the same time.",
                    mlevels[position[1]],
                    mlevels[position[2]]
                )
            )
        }

        return (
            paste (mlevels[unique (sort (c (first, cpos)))], collapse = ", ")
        )
    }
    else {
        return (paste (c (mlevels[first], mlevels[cpos]), collapse = ", "))
    }
}

`check_decimals` <- function (x) {
    if (is.null (x)) {
        return (x)
    }

    if (
        !is.numeric (x) ||
        length (x) != 1 ||
        is.na (x) ||
        x < 0 ||
        x != floor (x)
    ) {
        stopError_ ("`decimals` should be a single non-negative integer.")
    }

    return (as.integer (x))
}


`likely_measurement` <- function (x) {

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
    }

    labels <- attr (x, "labels", exact = TRUE)
    na_values <- attr (x, "na_values", exact = TRUE)
    na_range <- attr (x, "na_range", exact = TRUE)

    x <- undeclare (x, drop = TRUE)
    numx <- possibleNumeric_ (x)

    if (length (labels) > 0) {

        # possibly a categorical variable
        # but even numeric variables can have labels (for missing values)
        # check unique labels except the missing values
        except_na <- setdiff (labels, na_values)
        numlabels <- possibleNumeric_ (labels)

        if (length (except_na) > 0 && !is.null (na_range)) {
            if (numlabels) {
                except_na <- asNumeric_ (except_na)
                except_na <- except_na[
                    except_na < na_range[1] | except_na > na_range[2]
                ]
            } else {
                # in principle this should not be possible, but just in case
                return ("") # character, so cannot determine the measurement level
            }
        }

        if (length (except_na) == 0) {
            if (numlabels & numx) {
                return ("quantitative")
            }
            else {
                # character, but cannot determine the measurement level
                return ("")
            }
        }

        return ("categorical")
    }

    if (numx) {
        return ("quantitative")
    }

    return ("") # character, but cannot determine the measurement level
}


`all_missing_values` <- function (
    x, na_values = NULL, na_range = NULL, labels = NULL
) {

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
    }

    ##########
    # Arguments na_values, na_range and labels can either be provided
    # by the user or, if the input is a haven_labelled(_spss) objects
    # they might already be in the attributes
    if (is.null (na_values)) {
        na_values <- attr (x, "na_values")
    }

    if (is.null (na_range)) {
        na_range <- attr (x, "na_range")
    }

    if (is.null (labels)) {
        labels <- attr (x, "labels", exact = TRUE)
    }
    ##########


    misvals <- c ()

    if (is.null (na_values) & is.null (na_range)) {
        return (misvals)
    }

    if (!is.null (na_values)) {
        misvals <- sort (na_values)
    }

    if (is.numeric (x)) {
        if (!is.null (labels)) {
            x <- c (x, unname (unclass (labels)))
        }

        if (!is.null (na_range)) {
            na_range <- range (na_range)
            if (length (unique (na_range)) != 2) {
                stopError_ ("`na_range` must have two unique values.")
            }

            uniques <- sort (unique (x[x >= na_range[1] & x <= na_range[2]]))
            if (length (uniques) == 0) {
                uniques <- na_range
            } else {
                uniques <- sort (unique (c (uniques, na_range)))
            }

            misvals <- sort (unique (c (misvals, uniques)))
        }
    }

    return (misvals)
}


#' @rdname declared_internal
#' @export
`names_values` <- function (x, drop_na = FALSE, observed = TRUE) {

    if (!inherits (x, "declared") & !inherits (x, "haven_labelled_spss")) {
        stopError_ (
            "The input should be a declared / haven_labelled_spss vector."
        )
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
    }

    attrx <- attributes (x)
    na_values <- getElement(attrx, "na_values")
    labels <- getElement(attrx, "labels")

    if (drop_na) {
        attrx$na_index <- NULL
        attrx$na_values <- NULL

        if (!is.null (labels)) {
            labels <- labels[!is.element (labels, na_values)]
            attrx$labels <- labels
        }
        attributes (x) <- NULL
    }
    else {
        x <- undeclare (x, drop = TRUE)
    }

    x <- c (x, unname (labels))
    x <- x[!duplicated (x)]
    xmis <- logical (length (x))

    na_values <- getElement(attrx, "na_values")
    na_range <- getElement(attrx, "na_range")


    if (!is.null (na_values)) {
        xmis <- xmis | is.element (x, na_values)
    }

    if (!is.null (na_range)) {
        xmis <- xmis | (x >= na_range[1] & x <= na_range[2])
    }


    xnotmis <- sort (x[!xmis])
    xmis <- sort (x[xmis])
    if (all (is.element (xmis, na_values))) {
        xmis <- na_values
    }

    if (isFALSE (observed) && is.numeric (labels)) {
        lnotmis <- length(xnotmis)
        labels_notmis <- labels[!is.element(labels, na_values)]
        if (labels_notmis[2] < 16) {
            # there should be an upper limit for tables of frequencies
            if (
                length (labels_notmis) == 2 &&
                xnotmis[1] == labels_notmis[1] &&
                xnotmis[lnotmis] == labels_notmis[2] &&
                lnotmis != (labels_notmis[2] + labels_notmis[1] == 0)
            ) {
                # Likert type response scale, with labels only for the
                # first and last values
                xnotmis <- seq(labels_notmis[1], labels_notmis[2])
            }
        }
    }

    wd <- which(duplicated(labels))
    if (length(wd) > 0) {
        labels <- labels[-wd]
    }

    if (length (xmis) > 0) {
        names (xmis) <- xmis
        for (i in seq (length (xmis))) {
            if (any (isel <- labels == xmis[i])) {
                names (xmis)[i] <- names (labels)[isel]
            }
        }
    }


    names (xnotmis) <- xnotmis
    if (length (xnotmis) > 0) {
        nms <- names (labels)
        for (i in seq (length (xnotmis))) {
            if (any (isel <- labels == xnotmis[i])) {
                names (xnotmis)[i] <- ifelse (
                    nzchar(nms[isel]), nms[isel], xnotmis[i]
                )
            }
        }
    }

    result <- c (xnotmis, xmis)
    attr (result, 'missing') <- unname (xmis)

    return (result)
}



# The following functions are copied from package admisc
# to achieve zero dependency.
`stopError_` <- function (message, enter = "\n") {

    message <- paste0 (
        "Error: ",
        unlist (
            strsplit (message, split = "\\n")
        )
    )

    for (i in seq (length (message))) {
        message[i] <- gsub (
            "Error: ",
            ifelse (i > 1, "       ", ""),
            paste (
                strwrap (message[i], exdent = 7),
                collapse = "\n"
            )
        )
    }

    cat (enter)

    stop (
        simpleError(
            paste0 (
                paste (message, collapse = "\n"),
                enter, enter
            )
        )
    )
}


`coerceMode_` <- function (x) {

    if (is.null (x) || !is.atomic (x)) {
        stopError_ ("The input is not atomic.")
    }

    if (
        !is.numeric (x) &&
        (possibleNumeric_ (x) || all (is.na (x)))
    ) {
        x <- asNumeric_ (x)
    }

    if (
        !is.integer (x) &&
        wholeNumeric_ (x) &&
        # some whole numbers might be too big to be represented in memory
        # as integers, in which case a warning will be captured
        # else, if nothing is captured (the result is null) everything is ok
        is.null (tryCatchWEM_ (as.integer (x)))
    ) {
        x <- as.integer (x)
    }

    return (x)
}


`possibleNumeric_` <- function (x, each = FALSE) {

    result <- rep (NA, length (x))
    nax <- is.na (x)

    if (all (nax)) {
        if (each) {
            return (result)
        }

        return (FALSE)
    }

    if (is.logical(x)) {
        if (each) {
            result <- logical (length (x))
            result[nax] <- NA
            return (result)
        }

        return (FALSE)
    }

    if (inherits (x, "haven_labelled") || inherits(x, "declared")) {
        num <- Recall (unclass (x), each = each)

        labels <- attr (x, "labels", exact = TRUE)
        if (!is.null (labels) && !each && num) {
            return (Recall (labels))
        }

        return (num)
    }

    if (is.numeric (x)) {
        if (each) {
            result[!nax] <- TRUE
            return (result)
        }

        return (TRUE)
    }

    if (is.factor (x)) {
        x <- as.character (x)
    }

    x <- gsub (
        "\u00a0", # multibyte space
        " ",
        gsub (
            "\u009d", # weird zero length character
            "",
            x
        )
    )
    multibyte <- grepl ("[^!-~ ]", x)

    if (any (multibyte)) {
        result[multibyte] <- FALSE
    }

    if (sum (nax) < length (x)) {
        eachx <- suppressWarnings (as.numeric (x[!nax & !multibyte]))
        result[!nax & !multibyte] <- !is.na(eachx)
    }

    if (each | length(x) == 1) {
        return (result)
    }

    return (all (result[!nax]))
}


`asNumeric_` <- function (x, levels = TRUE) {
    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        attributes (x) <- NULL

        if (!is.null (na_index)) {
            # Non-numeric missing codes naturally become NA after coercion.
            x[na_index] <- suppressWarnings (as.numeric (names (na_index)))
        }

        return (Recall (x, levels = levels))
    }

    if (is.numeric (x)) {
        return (x)
    }

    if (is.factor (x)) {
        if (isTRUE (levels)) {
            return (suppressWarnings (as.numeric (levels (x)))[x])
        }
        return (as.numeric (x))
    }

    if (!all (is.na (x))) {
        x <- gsub ("\u00a0", " ", x) # multibyte space
    }

    result <- rep (NA, length (x))
    multibyte <- logical (length (x))
    if (!all (is.na (x))) {
        multibyte <- grepl ("[^!-~ ]", x)
    }

    attributes (x) <- NULL

    result[!multibyte] <- suppressWarnings (as.numeric (x[!multibyte]))

    return (result)
}


`wholeNumeric_` <- function (x, each = FALSE) {
    if (inherits (x, "haven_labelled") || inherits (x, "declared")) {
        return (Recall (unclass (x), each = each))
    }

    if (!possibleNumeric_ (x) & !each) {
        return (FALSE)
    }

    result <- logical (length (x))
    isna <- is.na (x)
    result[isna] <- NA

    if (all (isna) || is.logical (x)) {
        # each is certainly TRUE because if they are all missing or all logical
        # it would not be numeric, which means the only condition continuing the
        # function on line 636 is each = TRUE otherwise line 637 would stop
        return (result)
    }

    x <- asNumeric_ (x)
    # some characters might be recoded to NA when coerced to numeric
    isnax <- is.na (x)

    result[!isna & isnax] <- FALSE
    isna <- isna | isnax
    x <- x[!isna]

    result[!isna] <- abs (x - round (x)) < .Machine$double.eps^0.5

    if (each) {
        return (result)
    }

    return (all (result[!isna]))
}


`tryCatchWEM_` <- function (expr, capture = FALSE) {
    env <- new.env()
    env$toreturn <- list ()
    output <- withVisible (withCallingHandlers (
        tryCatch (expr, error = function (e) {
            env$toreturn$error <- e$message
            NULL
        }),
        warning = function (w) {
            env$toreturn$warning <- c (env$toreturn$warning, w$message)
            invokeRestart ("muffleWarning")
        },
        message = function (m) {
            env$toreturn$message <- paste (
                env$toreturn$message,
                m$message,
                sep = ""
            )
            invokeRestart ("muffleMessage")
        }
    ))

    if (capture && output$visible && !is.null (output$value)) {
        env$toreturn$output <- capture.output (output$value)
        env$toreturn$value <- output$value
    }

    if (length (env$toreturn) > 0) {
        return (env$toreturn)
    }
}


`padLeft_` <- function (x, n) {
    paste (c (rep (" ", n), x), collapse = "", sep = "")
}


`padRight_` <- function (x, n) {
    paste (c (x, rep (" ", n)), collapse = "", sep = "")
}


`padBoth_` <- function (x, n) {
    n1 <- ceiling(n/2)
    n2 <- floor (n/2)
    paste (c (rep (" ", n1), x, rep (" ", n2)), collapse = "", sep = "")
}


# `unlockEnvironment_` <- function (env) {
#     .Call ("_unlockEnvironment", env, PACKAGE = "declared")
# }


#' @rdname declared_internal
#' @keywords internal
#' @export
`makeTag_` <- function (...) {
    x <- as.character (c (...))

    x <- .Call ("_tag", x, PACKAGE = "declared")
    class (x) <- "double"

    return (x)
}


#' @rdname declared_internal
#' @keywords internal
#' @export
`hasTag_` <- function (x, tag = NULL) {
    if (!is.double (x)) {
        return (logical (length (x)))
    }

    if (
        !is.null (tag) && !(
            is.atomic (tag) && length (tag) == 1 && !is.na (tag)
        )
    ) {
        stopError_ ("`tag` should be a vector of length 1.")
    }

    if (!is.null (tag)) {
        tag <- as.character (tag)
    }

    return (.Call ("_hasTag", x, tag, PACKAGE = "declared"))
}


#' @rdname declared_internal
#' @keywords internal
#' @export
`getTag_` <- function (x) {
    if (is.double (x)) {
        x <- .Call ("_getTag", x, PACKAGE = "declared")
        if (!any (is.na (suppressWarnings (as.numeric (na.omit (x)))))) {
            x <- as.numeric (x)
        }
        return (x)
    }
    else {
        # stopError_ ("Unsuitable input to extract a tagged value.")
        return (rep (NA, length (x)))
    }
}


#' @rdname declared_internal
#' @keywords internal
#' @export
`anyTagged_` <- function (x) {
    if (is.data.frame (x)) {
        i <- 1
        tagged <- FALSE
        while (!tagged & i <= ncol (x)) {
            tagged <- Recall (x[[i]])
            i <- i + 1
        }
        return (tagged)
    }

    if (is.double (x)) {
        return (.Call ("_anyTagged", x, PACKAGE = "declared"))
    }

    return (FALSE)
}


`numdec_` <- function (x, each = FALSE, na.rm = TRUE, maxdec = 15) {

    maxdec <- min (15, maxdec)

    pN <- possibleNumeric_ (x, each = TRUE)

    # sum (pN), maybe each = TRUE and it's a vector
    if (sum (na.omit (pN)) == 0) {
        stopError_ ("'x' should contain at least one (possibly) numeric value.")
    }

    # asNumeric is important here because the (possible) number might arrive
    # as character through coercion, for instance c ("A", 1e-04)
    if (is.character (x)) {
        x <- asNumeric_ (x)
    }

    result <- rep (NA, length (x))
    wpN <- which (pN)

    x <- abs (x[wpN])
    x <- x - floor (x) # all numbers are now 0.something

    x <- sub ("0\\.", "",
        sub ("0+$", "", # erase trailing zeros
            format (x, scientific = FALSE, digits = max (7, maxdec))
        )
    )

    if (any (w9 <- grepl ("999999", x))) {
        # A floating point number like 234.1 might have been represented as
        # 0.0999999999999943 (after subtracting the floor)
        x[w9] <- sub (
            # last 0 becomes 1
            "0+", "1",
            # retains everthing <up to> the sequence
            sub ("(*)999999.*", "\\1", x[w9])
        )
    }

    if (any (w0 <- grepl ("000000", x))) {
        # ex. 0.00000000000000001, for all practical purposes this is equal to 0
        x[w0] <- sub ("(*)000000.*", "\\1", x[w0])
    }

    result[wpN] <- nchar (x)

    if (each) {
        return (pmin (result, maxdec))
    }

    return (min (maxdec, max (result, na.rm = na.rm)))
}


`trimstr_` <- function (x, what = " ", side = "both") {
    if (is.element (what, c ("*", "+"))) {
        what <- paste ("\\", what, sep = "")
    }

    what <- ifelse (
        identical (what, " "),
        paste0 ("[[:space:]|", "\u00a0", "]"), # plus the multibyte space
        what
    )

    pattern <- switch(side,
        both = paste ("^", what, "+|", what, "+$", sep = ""),
        left = paste ("^", what, "+", sep = ""),
        right = paste (what, "+$", sep = "")
    )

    gsub (pattern, "", x)
}


`getName_` <- function(x, object = FALSE) {
    result <- rep ("", length (x))
    x <- as.vector (gsub ("1-", "", gsub ("[[:space:]]", "", x)))

    condsplit <- unlist (strsplit (x, split = ""))

    startpos <- 0
    keycode <- ""

    if (any (condsplit == "]")) {
        startpos <- max (which (condsplit == "]"))
        keycode <- "]"
    }


    if (any (condsplit == "$")) {
        sp <- max (which (condsplit == "$"))
        if (sp > startpos) {
            startpos <- sp
            keycode <- "$"
        }
    }


    if (identical (keycode, "$")) {
        if (object) {
            return (substring (x, 1, min (which (condsplit == "$")) - 1))
        }

        # else
        result <- substring (x, startpos + 1)

    }
    else if (identical (keycode, "]")) {
        # ex. dd[,c("A","B")]
        # or dd[,c(A,B)] if the quotes have been removed before this function

        objname <- substring (x, 1, min (which (condsplit == "[")) - 1)

        if (object) {
            return (objname)
        }

        nms <- character(0)
        for (target in c ("names", "colnames")) {
        for (n in 1:2) {
            if (length (nms) == 0) {
                testnms <- tryCatchWEM_ (
                    nms <- eval.parent (
                        parse (
                            text = paste (target, "(", objname, ")", sep = "")
                        ),
                        n = n
                    )
                )
            }
        }
        }

        # else
        # keycode is "]"
        # this is a matrix or a list
        # determine where the indexing starts
        stindex <- max (which (condsplit == "["))

        stopindex <- ifelse (
            identical (condsplit[stindex - 1], "["),
            stindex - 2,
            stindex - 1
        )

        # ptn = possibly the name
        # ,c("A","B") or c(A, B)
        ptn <- gsub ("]", "", substr (x, stindex + 1, startpos))

        if (substring (ptn, 1, 1) == ",") {
            ptn <- substring (ptn, 2)
        }

        if (substring (ptn, 1, 2) == "c(") {
            # "A","B" or A,B
            ptn <- substring (ptn, 3, nchar(ptn) - 1)
        }

        ptn <- gsub ("'|\"|]|\ ", "", ptn)

        ptn <- unlist (strsplit (ptn, split = ","))
        if (length (ptn) == 1) {
            # try for something like [, 1:2]
            ptn <- unlist (strsplit (ptn, split = ":"))
        }

        # determine if what remains is a number or a name
        if (possibleNumeric_ (ptn)) {
            # it's a number (an index)
            # see if it has column names

            if (length (nms) > 0) {
                result <- nms[as.numeric (ptn)]
            }
        }
        else if (length (nms) > 0) {
            if (all (is.element (ptn, nms))) {
                return (ptn)
            }
        }
    }
    else {
        result <- x
    }

    return (gsub (",|'|\"|[[:space:]]", "", result))
}


check_date <- function (x) {
    xdate <- isTRUE (attr (x, "date"))
    attributes (x) <- NULL
    if (xdate) {
        x <- as.Date (x)
    }
    return (x)
}


`xtfrm_declared` <- function (x, decreasing = FALSE, na.last = TRUE, empty.last = na.last) {
    x <- sanitize_na_index_ (x)
    na_index <- attr (x, "na_index")
    na_declared <- logical (length (x))
    na_declared[na_index] <- TRUE
    na_empty <- is.empty (x)
    valid <- !is.na (x)

    z <- numeric (length (x))

    if (is.na (na.last)) {
        z[na_declared | na_empty] <- NA
        if (any (valid)) {
            val <- undeclare (x, drop = TRUE)
            r_v <- xtfrm (val)
            r_v <- r_v - min (r_v[valid]) + 1
            z[valid] <- r_v[valid]
        }
        return (z)
    }

    if (any (valid)) {
        val <- undeclare (x, drop = TRUE)
        r_v <- xtfrm (val)
        r_v <- r_v - min (r_v[valid]) + 1
        z[valid] <- r_v[valid]
    }

    if (any (na_declared)) {
        nms <- names (na_index)
        if (possibleNumeric_ (nms)) {
            nms <- asNumeric_ (nms)
        }
        r_m_vals <- xtfrm (nms)
        r_m_vals <- r_m_vals - min (r_m_vals) + 1
        z[na_index] <- r_m_vals
    }

    if (any (na_empty)) {
        z[na_empty] <- 1
    }

    if (isTRUE (na.last)) {
        if (isTRUE (empty.last)) {
            g_v <- 1
            g_m <- 2
            g_e <- 3
        }
        else {
            g_v <- 1
            g_e <- 2
            g_m <- 3
        }
    }
    else {
        if (isTRUE (empty.last)) {
            g_m <- 1
            g_e <- 2
            g_v <- 3
        }
        else {
            g_e <- 1
            g_m <- 2
            g_v <- 3
        }
    }

    shift <- 10 + length (x)

    if (isFALSE (decreasing)) {
        if (any (valid)) {
            z[valid] <- z[valid] + (g_v - 1) * shift
        }
        if (any (na_declared)) {
            z[na_declared] <- z[na_declared] + (g_m - 1) * shift
        }
        if (any (na_empty)) {
            z[na_empty] <- z[na_empty] + (g_e - 1) * shift
        }
    }
    else {
        if (any (valid)) {
            z[valid] <- z[valid] + (3 - g_v) * shift
        }
        if (any (na_declared)) {
            z[na_declared] <- z[na_declared] + (3 - g_m) * shift
        }
        if (any (na_empty)) {
            z[na_empty] <- z[na_empty] + (3 - g_e) * shift
        }
    }

    return (z)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 2
#'
#' @param trim A fraction (0 to 0.5) of observations to be trimmed from each end
#' of x before the mean is computed. Values of trim outside that range are
#' taken as the nearest endpoint.
#'
#' @param na.rm Logical, should the empty missing values be removed?
#'
#' @export
`wmean` <- function (
    x, wt = NULL, trim = 0, na.rm = TRUE
) {

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (
        is.null (x) || !is.atomic (x) || !(
            is.numeric (x) || is.complex (x) || is.logical (x)
        )
    ) {
        warning ("'x' should be a numerical / logical vector: returning NA")
        return (NA_real_)
    }

    if (inherits (x, "declared")) {
        x <- sanitize_na_index_ (x)
        na_index <- attr (x, "na_index")
        if (!is.null (na_index)) {
            wt <- wt[-na_index]
            x <- x[-na_index]
        }
    }

    if (is.null (wt)) {
        return (mean (x, na.rm = na.rm))
    }

    if (
        !is.null (wt) && !(
            is.atomic (wt) && all (is.finite (na.omit (wt)))
        )
    ) {
        stopError_ ("'wt' should be an atomic vector with finite values.")
    }

    if (length (x) != length (wt)) {
        stopError_ ("Lengths of 'x' and 'wt' differ.")
    }

    ok <- !is.na (x + wt)

    if (na.rm) {
        x <- x[ok]
        wt <- wt[ok]
    }
    else if (any (!ok)) {
        return (NA_real_)
    }

    sumwt <- sum (wt)

    if (any (wt < 0) || sumwt == 0) {
        stopError_ ("'wt' must be non-negative and not all zero")
    }

    n <- length (x)

    if (trim > 0 & n) {
        if (is.complex (x)) {
            stopError_ ("Trimmed means are not defined for complex data")
        }

        if (trim >= 0.5) {
            return (wmedian (x, wt = wt))
        }

        lo <- floor (n * trim) + 1
        hi <- n + 1 - lo

        # when this will be implemented in the base package:
        # ox <- sort.list (x, partial = unique (c (lo, hi)))
        lohi <- order (wt * x)[lo:hi]

        x <- x[lohi]
        wt <- wt[lohi]
        sumwt <- sum (wt)
    }

    return (sum (wt * x)/sumwt)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_mean` <- function (...) {
    wmean(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @title Coerce to haven / labelled objects
#' @description
#' Convert declared labelled objects to haven labelled objects
#' @details
#' This is a function that reverses the process of `as.declared()`, making
#' a round trip between `declared` and `haven_labelled_spss` classes.
#'
#' @return A labelled vector of class "haven_labelled_spss".
#' @examples
#'
#' x <- declared(
#'     c(1:5, -1),
#'     labels = c(Good = 1, Bad = 5, DK = -1),
#'     na_values = -1
#' )
#'
#' x
#'
#' as.haven(x)
#' @param x A declared labelled vector
#' @param ... Other arguments used by various methods
#' @export

#' @name as.haven
#' @export
# See https://stackoverflow.com/questions/18512528/
#     how-to-export-s3-method-so-it-is-available-in-namespace
`as.haven` <- function (x, ...) {
    UseMethod ("as.haven")
}

#' @export
`as.haven.default` <- function (x, ...) {

    dots <- list (...)
    interactive <- !isFALSE(dots$interactive)

    if (interactive) {
        msg <- "There is no automatic class method conversion for this type of"
        if (!is.null (dots$vname_)) {
            msg <- paste0 (dots$vname_, ": ", msg, " variable.")
        }
        else {
            msg <- paste (msg, "object.")
        }
        message (msg)
    }

    return (x)
}

#' @export
`as.haven.haven_labelled` <- function (x, ...) {
    # do nothing
    return (x)
}

#' @export
`as.haven.haven_labelled_spss` <- function (x, ...) {
    # do nothing
    return (x)
}



#' @export
`as.haven.declared` <- function (x, ...) {
    x <- sanitize_na_index_ (x)
    attrx <- attributes (x)
    attrx$class <- c ("haven_labelled", "vctrs_vctr")

    declared_is_integer <- is.integer (x)
    xdate <- inherits(x, "Date")

    # this is necessary to replace those values
    # (because of the "[<-.declared" method)
    attributes (x) <- NULL # or x <- unclass (x), but I find this cleaner

    spss <- TRUE
    if (possibleNumeric_ (x) || all (is.na (x))) { # this excludes dates
        x <- as.double (x)
        attr (x, "class") <- "double"
    }

    na_index <- attrx$na_index
    has_na_index <- !is.null (na_index)

    na_values <- attrx$na_values
    has_na_values <- !is.null (na_values)

    na_range <- attrx$na_range
    has_na_range <- !is.null (na_range)
    pN_na_range <- possibleNumeric_ (na_range)

    labels <- attrx$labels
    has_labels <- !is.null(labels)

    tagged_na_values <- NULL
    pN_na_values <- possibleNumeric_ (na_values)

    if (has_na_values && pN_na_values) {
        na_values <- asNumeric_ (na_values)
    }

    pN_labels <- FALSE
    if (has_labels) {
        pN_labels <- TRUE
        if (length (setdiff (labels, na_values)) > 0) {
            pN_labels <- possibleNumeric_ (setdiff (labels, na_values))
            if (!pN_labels) {
                x <- as.character (x)
            }
        }
    }


    if (has_na_index) {
        na_index_values <- names (na_index)

        if (pN_na_values | pN_na_range) {
            na_index_values <- asNumeric_ (na_index_values)
        } else if (is.numeric (x)) {
            if (length (setdiff (na_values, c (letters, LETTERS))) > 0) {
                stopError_ ("Tagged NAs can only be created for single letter declared missing values.")
            }

            spss <- FALSE
            attrx$na_values <- NULL
            tagged_na_values <- makeTag_ (na_values)
            pN_na_values <- TRUE # tagged_na_values are now numeric
            na_index_values <- tagged_na_values[match(na_index_values, na_values)]
            attrx$class <- setdiff(attrx$class, "integer")
        }

        x[na_index] <- na_index_values
    }

    if (
        xdate | is.numeric (x) &
        (pN_labels | !has_labels) &
        (pN_na_values | is.null(na_values))
    ) {

        if (has_labels) {
            copy_labels <- numeric (length (labels))
            not_missing <- !is.element(labels, na_values)

            copy_labels[not_missing] <- asNumeric_ (labels[not_missing])

            if (has_na_values) {
                labels_missing <- setdiff(match(labels, na_values), NA)

                if (is.null(tagged_na_values)) {
                    copy_labels[!not_missing] <- asNumeric_ (na_values[labels_missing])
                }
                else {
                    copy_labels[!not_missing] <- tagged_na_values[labels_missing]
                }
            }

            names (copy_labels) <- names (attrx$labels)
            attrx$labels <- copy_labels
        }

        if (has_na_values & spss) {
            na_values <- as.numeric (na_values)
            names (na_values) <- names (attrx$na_values)
            attrx$na_values <- na_values
        }
    } else {
        x <- as.character (x)

        if (!is.null (na_values)) {
            na_values <- as.character (na_values)
            names (na_values) <- names (attrx$na_values)
            attrx$na_values <- na_values
        }

        if (!is.null (labels)) {
            labels <- as.character (labels)
            names (labels) <- names (attrx$labels)
            attrx$labels <- labels
        }
    }

    attrx$na_index <- NULL

    if (spss) { # deactivated above if tagged NAs
        attrx$class <- c ("haven_labelled_spss", attrx$class)

        if (declared_is_integer) {
            x <- as.integer (x)

            if (has_na_values) {
                # class (attrx$na_values) <- "integer"
                attrx$na_values <- as.integer (attrx$na_values)
            }

            if (has_labels) {
                nms <- names (attrx$labels)
                attrx$labels <- as.integer (attrx$labels)
                names (attrx$labels) <- nms
            }

            if (has_na_range && all(attrx$na_range > -Inf) && all(attrx$na_range < Inf)) {
                attrx$na_range <- as.integer (attrx$na_range)
            }
        }
    }

    attrx$class <- c (attrx$class, class (x))
    attrx$date <- NULL
    attributes (x) <- attrx
    return (x)
}

#' @export
`as.haven.data.frame` <- function (
    x, ..., only_declared = TRUE, interactive = FALSE
) {
    if (only_declared) {
        xdeclared <- vapply (x, is.declared, logical (1))
        if (isFALSE (interactive)) {
            x[xdeclared] <- lapply (
                x[xdeclared], as.haven, interactive = FALSE, ... = ...
            )
        }
        else {
            nms <- names (x)[xdeclared]
            for (i in seq (length (nms))) {
                x[[nms[i]]] <- as.haven (
                    x[[nms[i]]], vname_ = nms[i], ... = ...
                )
            }
        }
    } else {
        if (isFALSE (interactive)) {
            x[] <- lapply (x, as.haven, interactive = FALSE, ... = ...)
        }
        else {
            nms <- names (x)
            for (i in seq (length (nms))) {
                x[[i]] <- as.haven (x[[i]], vname_ = nms[i], ... = ...)
            }
        }
    }

    class (x) <- c ("tbl", "tbl_df", "data.frame")
    return (x)
}





# Dynamically exported, see onLoad.R
# using eval (parse ()) to avoid the huge dependency tree of vctrs, haven,
# labelled and pillar these functions will be registered when or if the package
# haven is loaded

`as_factor.declared` <- function (
    x, levels = c ("default", "labels", "values", "both"), ordered = FALSE, ...
) {
    eval(parse(text = paste(
        "haven::as_factor (",
        "    as.haven (x),",
        "    levels = levels,",
        "    ordered = ordered,",
        "    ... = ...",
        ")"
    )))
}

`zap_labels.declared` <- function (x) {
    attr (x, "labels") <- NULL
    attr (x, "na_index") <- NULL
    attr (x, "na_values") <- NULL
    attr (x, "na_range") <- NULL
    attr (x, "date") <- NULL
    class (x) <- NULL

    return (x)
}

`zap_missing.declared` <- function (x) {
    attr (x, "na_index") <- NULL
    attr (x, "na_values") <- NULL
    attr (x, "na_range") <- NULL
    attr (x, "date") <- NULL

    return (x)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname declared
#'
#' @export
`is.declared` <- function (x) {
  inherits (x, "declared")
}

#' @rdname declared
#' @export
`anyNAdeclared` <- function (x) {
  if (is.declared (x)) {
    x <- sanitize_na_index_ (x)
    if (length(attr(x, "na_index")) > 0) {
      return (TRUE)
    }
  }

  return (FALSE)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' Weighted summaries
#' @name weighted
#' @order 1
#'
#' @title Compute weighted summaries for declared objects
#'
#' @description
#' Functions to compute weighted tables or summaries, based on a vector of
#' frequency weights. These are reimplementations of various existing functions,
#' adapted to objects of class \code{"declared"} (see Details below)
#'
#' @returns
#' A vector of (weighted) values.
#'
#' @details
#' A frequency table is usually performed for a categorical variable, displaying
#' the frequencies of the respective categories. Note that general variables
#' containing text are not necessarily factors, despite having a small number of
#' characters.
#'
#' A general table of frequencies, using the base function `table()`, ignores
#' the defined missing values (which are all stored as NAs). The
#' reimplementation of this function in `wtable()` takes care of this detail,
#' and presents frequencies for each separately defined missing values. Similar
#' reimplementations for the other functions have the same underlying objective.
#'
#' It is also possible to perform a frequency table for numerical variables, if
#' the number of values is limited (an arbitrary and debatable upper limit of 15
#' is used). An example of such variable can be the number of children, where
#' each value can be interpreted as a class, containing a single value (for
#' instance 0 meaning the category of people with no children).
#'
#' Objects of class `declared` are not pure categorical variables (R factors)
#' but they are nevertheless interpreted as if they were factors, to allow
#' producing frequency tables. Given the high similarity with package
#' **`haven`**, objects of class `haven_labelled_spss` are automatically
#' coerced to objects of class `declared` and treated accordingly.
#'
#' The argument `values` makes sense only when the input is of family class
#' `declared`, otherwise for regular (base R) factors the values are
#' just a sequence of numbers.
#'
#' The later introduced argument `observed` is useful in situations when a
#' variable has a very large number of potential values, and a smaller subset of
#' actually observed ones. As an example, the variable \dQuote{Occupation} has
#' hundreds of possible values in the ISCO08 codelist, and not all of them might
#' be actually observed. When activated, this argument restricts the printed
#' frequency table to the subset of observed values only.
#'
#' The argument `wt` refers only to frequency weights. Users should be
#' aware of the differences between frequency weights, analytic weights,
#' probability weights, design weights, post-stratification weights etc. For
#' purposes of inferential testing, Thomas Lumley's package **`survey`**
#' should be employed.
#'
#' If no frequency weights are provided, the result is identical to the
#' corresponding base functions.
#'
#' For all functions, the argument `na.rm` refers to the empty missing values
#' and its default is set to TRUE. The declared missing values are automatically
#' eliminated from the summary statistics, even if this argument is deactivated.
#'
#' @examples
#' set.seed(215)
#'
#' # a pure categorical variable
#' x <- factor(sample(letters[1:5], 215, replace = TRUE))
#' wtable(x)
#'
#'
#' # simulate number of children
#' x <- sample(0:4, 215, replace = TRUE)
#' wtable(x)
#'
#' # simulate a Likert type response scale from 1 to 7
#' values <- sample(c(1:7, -91), 215, replace = TRUE)
#' x <- declared(values, labels = c("Good" = 1, "Bad" = 7))
#' wtable(x)
#'
#'
#' # Defining missing values
#' missing_values(x) <- -91
#' wtable(x)
#'
#'
#' # Defined missing values with labels
#' values <- sample(c(1:7, -91, NA), 215, replace = TRUE)
#' x <- declared(
#'     values,
#'     labels = c("Good" = 1, "Bad" = 7, "Don't know" = -91),
#'     na_values = -91
#' )
#'
#' wtable(x)
#'
#' # Including the values in the table of frequencies
#' wtable(x, values = TRUE)
#'
#'
#' # An example involving multiple variables
#' DF <- data.frame(
#'     Area = declared(
#'         sample(1:2, 215, replace = TRUE, prob = c(0.45, 0.55)),
#'         labels = c(Rural = 1, Urban = 2)
#'     ),
#'     Gender = declared(
#'         sample(1:2, 215, replace = TRUE, prob = c(0.55, 0.45)),
#'         labels = c(Males = 1, Females = 2)
#'     ),
#'     Age = sample(18:90, 215, replace = TRUE),
#'     Children = sample(0:5, 215, replace = TRUE)
#' )
#'
#' wtable(DF$Gender)
#'
#' wsd(DF$Age)
#'
#' wmeasures(DF[c("Age", "Children")], what = c("n", "mean", "sd"))
#'
#'
#' # Weighting: observed proportions
#' op <- proportions(with(DF, table(Gender, Area)))
#'
#' # Theoretical proportions: 53% Rural, and 50.2% Females
#' tp <- rep(c(0.53, 0.47), each = 2) * rep(c(0.498, 0.502), 2)
#'
#' # Corrections by strata
#' fweights <- tp / op
#'
#' DF$fweight <- fweights[match(10 * DF$Area + DF$Gender, c(11, 12, 21, 22))]
#'
#' with(DF, wtable(Gender, wt = fweight))
#'
#' with(DF, wmean(Age, wt = fweight))
#'
#' with(DF, wquantile(Age, wt = fweight))
#' @author Adrian Dusa
#'
#' @param x A numeric vector for summaries, or declared / factor for frequency
#' tables
#'
#' @param y An optional variable, to create crosstabs; must have the same length
#' as x
#'
#' @param wt A numeric vector of frequency weights
#'
#' @param values Logical, print the values in the table rows
#'
#' @param valid Logical, print separate percent distribution for valid values,
#' if any missing values are present; for cross tables, use valid values only
#'
#' @param observed Logical, print the observed categories only
#'
#' @param vlabel Logical, print the variable label, if existing
#' @export
`wtable` <- function (
    x, y = NULL, wt = NULL, values = TRUE, valid = TRUE, observed = TRUE,
    vlabel = FALSE
) {

    funargs <- lapply(
        lapply(
            match.call(), deparse)[-1],
            function(x) gsub("'|\"|[[:space:]]", "", x
        )
    )

    nmx <- getName_ (funargs$x)

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    if (is.null (x) || !is.atomic (x)) {
        stopError_ ("'x' should be an atomic vector.")
    }

    xvallab <- yvallab <- NULL
    xna_values <- yna_values <- NULL
    xvalues <- yvalues <- TRUE
    missing_table <- NULL
    crosstab <- !is.null (y)

    if (!crosstab) {
        valid <- isTRUE (valid) && any (is.na (x))
    }

    xlabel <- attr (x, "label", exact = TRUE)
    allnax <- all (is.na (x))

    if (inherits (x, "declared")) {
        allnax <- all (is.empty (x))
        if (!allnax) {
            # names_values () arranges missing values at the end
            xvallab <- names_values (
                x,
                drop_na = isTRUE (valid) & crosstab,
                observed = observed
            )

            xna_values <- attr (xvallab, "missing")
            # x <- factor (as.character (x), levels = names (xvallab))
            # sometimes (e.g. ISCO codifications in ESS) there are identical labels
            # with different values, and factor () complains with overlapping levels

            xvalues <- !identical (names (xvallab), as.character (xvallab))
            # print (head(paste (as.character (x), undeclare (x), sep = "_-_")))
            x <- factor (
                paste (
                    undeclare (x, drop = TRUE),
                    as.character (undeclare (x)),
                    sep = "_-_"
                ),
                levels = paste (xvallab, names (xvallab), sep = "_-_")
            )
        }
        else {
            xvalues <- FALSE
            x <- undeclare (x, drop = TRUE)
        }
    }
    else {
        xvalues <- FALSE
        if (!allnax) {
            lvls <- levels (as.factor (x))
            xvallab <- seq (length (lvls))
            names (xvallab) <- lvls
        }
    }

    xy <- list (x = x)
    if (crosstab) {
        if (!is.atomic (y)) {
            stopError_ ("'y' should be an atomic vector.")
        }

        if (length (x) != length (y)) {
            stopError_ ("Lengths of 'x' and 'y' differ.")
        }

        ylabel <- attr (y, "label", exact = TRUE)
        allnay <- all (is.na (y))

        nmy <- getName_ (funargs$y)
        dimnms <- c(nmx, nmy)

        ncharx <- nchar (nmx)
        nchary <- nchar (nmy)
        if (length(c (ncharx + nchary)) > 0) {
            nchars <- max (nchar (c (nmx, nmy)))
        }

        if (length(ncharx) > 0 && ncharx > 0) {
            nmx <- padLeft_ (nmx, nchars - ncharx)
        }

        if (length(nchary) > 0 && nchary > 0) {
            nmy <- padLeft_ (nmy, nchars - nchary)
        }

        if (inherits (y, "declared")) {
            allnay <- all (is.empty (y))
            if (!allnay) {
                yvallab <- names_values (
                    y,
                    drop_na = crosstab && isTRUE (valid),
                    observed = observed
                )
                yna_values <- attr (yvallab, "missing")

                y <- factor (
                    paste (
                        undeclare (y, drop = TRUE),
                        as.character (undeclare (y)),
                        sep = "_-_"
                    ),
                    levels = paste (yvallab, names (yvallab), sep = "_-_")
                )
            }
            else {
                yvalues <- FALSE
                y <- undeclare (y, drop = TRUE)
            }
        }
        else {
            yvalues <- FALSE
            if (!allnay) {
                lvls <- levels (as.factor (y))
                yvallab <- seq (length (lvls))
                names (yvallab) <- lvls
            }
        }

        xy$y <- y
    }

    if (is.null (wt)) {
        wt <- rep (1, length (x))
    }


    if (
        !is.null (wt) && !(
            is.atomic (wt) && all (is.finite (na.omit (wt)))
        )
    ) {
        stopError_ ("'wt' should be an atomic vector with finite values.")
    }

    if (length (x) != length (wt)) {
        stopError_ ("Lengths of 'x' and 'wt' differ.")
    }


    orig <- round (tapply (wt, xy, sum, na.rm = TRUE), 0)
    orig[is.na (orig)] <- 0
    tbl <- as.matrix(orig)
    dimnames (tbl) <- unname (dimnames (tbl))

    rs <- rowSums (tbl)
    cs <- colSums (tbl)

    if (isTRUE (observed)) {
        if (crosstab) {
            xvallab <- xvallab[rs > 0]
            yvallab <- yvallab[cs > 0]
            orig <- orig[rs > 0, , drop = FALSE]
            orig <- orig[, cs > 0, drop = FALSE]
            tbl <- tbl[rs > 0, , drop = FALSE]
            tbl <- tbl[, cs > 0, drop = FALSE]
            rs <- rs[rs > 0]
            cs <- cs[cs > 0]
        }
        else {
            orig <- orig[rs > 0]
            tbl <- tbl[rs > 0, , drop = FALSE]
            xvallab <- xvallab[rs > 0]
        }
    }

    if (crosstab) {

        toprint <- orig <- tbl
        names (dimnames (orig))  <- dimnms

        toprint <- rbind (toprint, Total = colSums (toprint))
        toprint <- cbind (toprint, Total = rowSums (toprint))

        if (isTRUE (vlabel)) {
            if (!is.null (xlabel) & !is.null (ylabel)) {
                attr (toprint, "xlabel") <- paste(nmx, xlabel, sep = ": ")
                attr (toprint, "ylabel") <- paste(nmy, ylabel, sep = ": ")
            } else {
                message("Variable label(s) not available.")
            }
        }
        attr (toprint, "xvalues") <- isTRUE (values) & xvalues
        attr (toprint, "yvalues") <- isTRUE (values) & yvalues

        # class (toprint) <- c ("wtable", "matrix")
    }
    else {
        only_na <- length (x) > 0 && isTRUE (allnax)
        labels <- NULL
        if (nrow(tbl) > 0) {
            labels <- rownames (tbl)
            labels <- sapply (strsplit (labels, split = "_-_"), tail, 1)
        }

        if (any (is.na (x))) {
            tbl <- c (tbl, sum (is.na (x)))
            labels <- c (labels, NA)
        }

        if (isTRUE (observed)) {
            tbl <- tbl[tbl > 0]
        }

        missing <- rep (FALSE, length (tbl))
        if (length (xna_values) > 0) {
            missing[seq_along (xvallab)] <- is.element (xvallab, xna_values)
        }
        missing <- missing | is.na (labels)
        missing_table <- tbl[missing]
        if (length (missing_table) > 0) {
            names (missing_table) <- labels[missing]
            names (missing_table)[is.na (names (missing_table))] <- "NA"
        }

        toprint <- data.frame (fre = tbl)

        toprint$rel <- proportions (toprint$fre)
        toprint$per <- toprint$rel * 100

        if (valid & (length (missing) > 0 | any (is.na (labels)))) {
            vld <- toprint$fre
            nalabels <- is.element (xvallab, xna_values)
            vld[nalabels] <- NA
            vld[is.na (labels)] <- NA

            lna <- seq (length (nalabels))

            vld[seq (sum (!nalabels))] <- 100 * proportions (
                vld[seq (sum (!nalabels))]
            )

            toprint$vld <- NA
            toprint$vld[seq (length (vld))] <- vld
            toprint$cpd <- NA
            toprint$cpd[seq (length (vld))] <- cumsum (vld)
        }
        else {
            valid <- FALSE
            toprint$cpd <- cumsum (toprint$per)
        }

        if (isTRUE (vlabel)) {
            attr (toprint, "xlabel") <- paste(nmx, xlabel, sep = ": ")
        }
        attr (toprint, "labels") <- labels
        attr (toprint, "values") <- as.vector (xvallab)
        attr (toprint, "show_values") <- values & xvalues
        attr (toprint, "na_values") <- xna_values
        attr (toprint, "valid") <- valid

        orig <- tbl[!missing]
        if (length (xvallab) > 0) {
            xvallab <- xvallab[!is.element (xvallab, xna_values)]
        }

        if (only_na) {
            orig <- array (numeric (0), dim = c (0))
        }
    }

    if (is.matrix(orig)) {
        if (length (xvallab)) {
            rownames (orig) <- names (xvallab)
        }
        if (length (yvallab)) {
            colnames (orig) <- names (yvallab)
        }
    }
    else {
        orig <- array (orig, dim = length (orig))
        if (length (xvallab)) {
            dimnames (orig) <- list (names (xvallab))
        }
    }



    attr (orig, "toprint") <- toprint
    if (length (missing_table) > 0) {
        attr (orig, "missing") <- missing_table
    }
    class (orig) <- c ("wtable", "table")
    return (orig)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_table` <- function (...) {
    wtable(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @keywords internal
"_PACKAGE"
#'
#' @name declared_package
#'
#' @title Functions for Declared Missing Values
#'
#' @description A set of functions to declare labels and missing values, coupled
#' with associated functions to create (weighted) tables of frequencies and
#' various other summary measures.
#' Some of the base functions are rewritten to make use of the specific
#' information about the missing values, most importantly to distinguish between
#' empty and declared missing values.
#' Many functions have a similar functionality with the corresponding ones from
#' packages "haven" and "labelled". The aim is to ensure as much compatibility
#' as possible with these packages, while offering an alternative in the objects
#' of class "declared".
#'
#' @author Adrian Dusa
#'
#' Maintainer: Adrian Dusa (dusa.adrian@unibuc.ro)
#'
#' @details
#' \tabular{ll}{
#'   Package: \tab declared\cr
#'   Type: \tab Package\cr
#'   Version: \tab 0.27\cr
#'   Date: \tab 2026-08-22\cr
#'   License: \tab GPL-v3\cr
#' }
#'
#' @importFrom stats na.omit na.fail na.exclude median
#' @importFrom utils head tail capture.output
#' @useDynLib declared, .registration = TRUE
#'
NULL

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' @rdname weighted
#' @order 11
#' @export
`wsummary` <- function (x, wt = NULL, ...) {

    wsum <- wquantile (x, wt = wt, ... = ...)
    wmean <- wmean (x, wt = wt, ... = ...)
    wsum <- c (wsum[1:3], wmean, wsum[4:5])

    if (inherits (x, "haven_labelled")) {
        x <- as.declared (x)
    }

    nas <- wtable (is.empty (x), wt = wt)
    rnms <- rownames (nas)
    if (is.null (rnms)) {
        rnms <- names (nas)
    }

    names (wsum) <- c ("Min.", "1st Qu.", "Median", "Mean", "3rd Qu.", "Max.")

    if (any (rnms == "TRUE")) {
        wsum <- c (
            wsum,
            "NA's" = attr (nas, "toprint")$fre[which (rnms == "TRUE")[1]]
        )
    }

    class (wsum) <- c ("fobject", class (wsum))
    return (wsum)
}

#' @rdname declared_internal
#' @keywords internal
#' @export
`w_summary` <- function (...) {
    wsummary(...)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

`na_values.declared` <- function (x, value) {
    missing_values (x)
}

`na_values<-.declared` <- function (x, value) {
    x <- sanitize_na_index_ (x)
    declared (
        x = undeclare (x),
        labels = attr (x, "labels", exact = TRUE),
        na_values = value,
        na_range = attr (x, "na_range", exact = TRUE),
        label = attr (x, "label", exact = TRUE),
        decimals = attr (x, "decimals", exact = TRUE)
    )
}

`na_range.declared` <- function (x, value) {
    missing_range (x)
}

`na_range<-.declared` <- function (x, value) {
    x <- sanitize_na_index_ (x)
    declared (
        x = undeclare (x),
        labels = attr (x, "labels", exact = TRUE),
        na_values = attr (x, "na_values", exact = TRUE),
        na_range = value,
        label = attr (x, "label", exact = TRUE),
        decimals = attr (x, "decimals", exact = TRUE)
    )
}

`val_labels.declared` <- function (x, prefixed = FALSE) {
    labels <- attr (x, "labels", exact = TRUE)
    if (prefixed) {
        names (labels) <- paste0 ("[", labels, "] ", names (labels))
    }

    return (labels)
}

`val_labels<-.declared` <- function (x, value) {
    attr (x, "labels") <- value
    return (x)
}

`var_label.declared` <- function (x) {
    attr (x, "label", exact = TRUE)
}

`var_label<-.declared` <- function (x, value) {
    if (!is.null (value) && length (value) > 1) {
        stopError_ ("`value` should be a single character string or NULL.")
    }

    if (is.null (value)) {
        attr (x, "label") <- NULL
    }
    else {
        attr (x, "label") <- as.character (value)
    }

    return (x)
}

`drop_unused_value_labels.declared` <- function (x) {
    labels <- labels (x)
    labels (x) <- labels[is.element (labels, unique (undeclare (x)))]
    return (x)
}

`val_label.declared` <- function (x, v, prefixed = FALSE) {
    if (length (v) != 1) {
        stop (
            "`v` should be a single value", call. = FALSE, domain = "R-labelled"
        )
    }

    labels <- labels (x)

    if (is.element (v, labels)) {
        if (prefixed) {
            return (paste0 ("[", v, "] ", names (labels)[labels == v]))
        }
        else {
            return (names (labels)[labels == v])
        }
    }
    else {
        return (NULL)
    }
}

`val_label<-.declared` <- function (x, v, value) {
    if (length (v) != 1) {
        stop (
            "`v` should be a single value", call. = FALSE, domain = "R-labelled"
        )
    }

    if (length (value) > 1) {
        stop ("`value` should be a single character string or NULL",
        call. = FALSE, domain = "R-labelled")
    }

    labels <- labels (x)

    if (is.null (value)) {
        if (is.element (v, labels)) {
            labels <- labels[labels != v]
        }
    }
    else {
        if (is.element (v, labels)) {
            names (labels)[labels == v] <- value
        }
        else {
            names (v) <- value
            labels <- c (labels, v)
        }
    }

    if (length (labels) == 0)
        labels <- NULL

    labels (x) <- labels
    x
}

`sort_val_labels.declared` <- function (x, according_to = c ("values",
    "labels"), decreasing = FALSE) {

    according_to <- match.arg (according_to)

    labels <- labels (x)

    if (!is.null (labels)) {
        if (according_to == "values") {
            labels <- sort (labels, decreasing = decreasing)
        }

        if (according_to == "labels") {
            labels <- labels[order (names (labels), decreasing = decreasing)]
        }

        labels (x) <- labels
    }

    return (x)
}

`nolabel_to_na.declared` <- function (x) {
    nolabel <- setdiff (
        na.omit (unique (undeclare (x, drop = TRUE))),
        labels (x)
    )
    if (length (nolabel) > 0) {
        x[is.element (x, nolabel)] <- NA
    }
    x
}

`val_labels_to_na.declared` <- function (x) {
    labels <- labels (x)

    if (length (labels) > 0) {
        x[is.element (x, labels)] <- NA
    }

    labels (x) <- NULL
    return (x)
}

`remove_labels.declared` <- function (
    x, user_na_to_na = FALSE, keep_var_label = FALSE
) {
    if (!keep_var_label) {
        label(x) <- NULL
    }

    labels (x) <- NULL
    attr (x, "format.spss") <- NULL
    return (x)
}

`remove_user_na.declared` <- function (x, user_na_to_na = FALSE) {
    attr (x, "na_index") <- NULL
    missing_values (x) <- NULL
    missing_range (x) <- NULL
    return (x)
}

`to_factor.declared` <- function (x, levels = c ("labels", "values",
    "prefixed"), ordered = FALSE, nolabel_to_na = FALSE,
    sort_levels = c ("auto", "none", "labels", "values"), decreasing = FALSE,
    drop_unused_labels = FALSE, user_na_to_na = FALSE, strict = FALSE,
    unclass = FALSE,
    ...) {

    x <- as.haven (x)
    levels <- match.arg (levels)
    sort_levels <- match.arg (sort_levels)
    eval(parse(text = paste(
        "labelled::to_factor (",
        "    x, levels = levels, ordered = ordered, sort_levels = sort_levels,",
        "    nolabel_to_na = nolabel_to_na, decreasing = decreasing,",
        "    drop_unused_labels = drop_unused_labels, strict = strict,",
        "    user_na_to_na = user_na_to_na, unclass = unclass, ... = ...",
        ")"
    )))
}

`to_character.declared` <- function (x, levels = c ("labels", "values",
    "prefixed"), nolabel_to_na = FALSE, user_na_to_na = FALSE, ...) {
    x <- as.haven (x)
    levels <- match.arg (levels)
    eval(parse(text = paste(
        "labelled::to_character (",
        "    x, levels = levels, nolabel_to_na = nolabel_to_na,",
        "    user_na_to_na = user_na_to_na, ... = ...",
        ")"
    )))
}

`copy_labels.declared` <- function (from, to, .strict = TRUE) {
    from <- sanitize_na_index_ (from)
    if (mode (from) != mode (to) & .strict) {
        stop (
            "`from` and `to` should be of same type",
            call. = FALSE, domain = "R-labelled"
        )
    }

    label(to) <- label(from)

    if (mode (from) == mode (to)) {
        labels (to) <- labels (from)
        missing_range (to) <- missing_range (from)
        missing_values (to) <- missing_values (from)
    }

    return (to)
}

# Copyright (c) 2022 - 2026, Adrian Dusa
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, in whole or in part, are permitted provided that the
# following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * The names of its contributors may NOT be used to endorse or promote
#       products derived from this software without specific prior written
#       permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ADRIAN DUSA BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#' Validate the declared missing values index
#'
#' Check whether every position stored in the `na_index` attribute identifies a
#' genuine missing value in the vector. This can detect when external code has
#' changed a vector without refreshing its declared missing values metadata.
#'
#' The check examines only the positions in `na_index` and stops at the first
#' invalid position. A missing or empty `na_index` attribute is valid.
#'
#' @param x An atomic vector, possibly containing an `na_index` attribute.
#' @return A logical value. `TRUE` when every indexed value is a genuine `NA`;
#'   otherwise `FALSE`.
#' @examples
#' x <- declared(c(1, -1, 3), na_values = -1)
#'
#' valid_na_index(x)
#'
#' attr(x, "na_index") <- 1
#' valid_na_index(x)
#' @export
`valid_na_index` <- function (x) {
  if (!is.atomic (x)) {
    return (FALSE)
  }

  index <- attr (x, "na_index")

  if (is.null (index)) {
    return (TRUE)
  }

  return (.Call (
    "_allIndexedNA", x, index, PACKAGE = "declared"
  ))
}


`sanitize_na_index_` <- function (x) {
  if (is.atomic (x) && !valid_na_index (x)) {
    # The positional map cannot be repaired selectively once it is stale.
    attr (x, "na_index") <- NULL
    attr (x, "na_values") <- NULL
    attr (x, "na_range") <- NULL
  }

  return (x)
}

---
title: "a. Motivation for the declared package"
output: 
  rmarkdown::html_vignette:
   highlight: "declared.theme"
   css: "declared.css"
vignette: >
  %\VignetteIndexEntry{a. Motivation for the declared package}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options: 
  markdown: 
    wrap: 80
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(declared)
```

The R ecosystem already has some very good packages that deal with labelled
objects. In particular, the inter-connected packages **haven** and **labelled**
provide all the functionality most users would ever need.

As nice and useful as these packages are, it has become apparent they have some
fundamental design features that run, in some situations, against users'
expectations. This often relates to the treatment of declared missing values
that are instrumental for the social sciences.

The following minimal example (adapted from the vignette in package **haven**)
illustrates the situation:

```{r}
library(haven)
x1 <- labelled_spss(
  x = c(1:5, -91),
  labels = c("Missing" = -91),
  na_value = -91
)
```

The printed objects from this package nicely display some properties:

```{r}
x1
```

There are 5 normal (non-missing) values (supposedly, they represent the number
of children), and one declared missing value coded `-91`. This value *acts* as a
missing value, but it is different from a regular missing value in R, coded `NA`.
The latter stands for any missing information (something like an empty cell)
regardless of the reason.

On the other hand, the cell is *not* empty, but the value `-91` is not a valid
value either. It cannot possibly represent -91 children in the household, but
for instance, it could have meant the respondent did not want to respond. It is
properly identified as missing with:

```{r}
is.na(x1)
```

But when calculating a mean, for instance, the normal expectation is that value
99 would not play any role in the calculations (since it should be *missing*).
However:

```{r}
mean(x1)
```

This means the value 99 did play an active role despite being identified as
"missing". In an ideal world, the expected mean would be 3, or at best, employ
the argument `na.rm = TRUE`{.R} if the result is `NA` because of the
declared missing value.

A solution to this problem is offered by package **labelled**, which has a
function called `user_na_to_na()`{.R}:

```{r}
library(labelled)
mean(
  user_na_to_na(x1),
  na.rm = TRUE
)
```


## The declared solution

While solving the problem, this above solution forces two additional operations:

-   converting the (already) declared user missing values, and

-   employing the `na.rm` argument.

This should not be necessary, especially if (and it is extremely likely that)
users may forget the declared missing values are not actually missing values.
This scenario is quite possible, as many users previously using other software
like SPSS or Stata where nothing else should be done after declaring the missing
values may not realize more is needed.

To solve this problem, the **declared** package creates a similar object, where
declared missing values are stored (hence interpreted as) regular `NA`
missing values in R.

```{r}
library(declared)
x2 <- declared(
  x = c(1:5, -91),
  labels = c("Missing" = -91),
  na_value = -91
)
x2
```

The print method makes it obvious the value `-91` is not a regular number but an
actual missing value. More importantly, this type of storage circumvents the
need to convert user-defined missing values to regular `NA` values since
they are already stored as regular NA values. The average value is calculated
simply as follows:

```{r}
mean(x2)
```

Notice that neither `user_na_to_na()`{.R}, nor employing `na.rm = TRUE`{.R} are
necessary, and, despite being stored as an `NA` value, the value `99` is not
equivalent to an *empty cell*. The information still exists, but it is simply
ignored in the calculations.

At first glance, providing a class method for this function seems unnecessary
because activating the argument `na.rm` will return the correct result.
Explaining the importance of the class method requires a discussion about the
base R decision to have this argument deactivated by default. This is most
likely to alert users about possible problems in the data since a default value
of `TRUE` would obscure such problems; the mean is calculated irrespective
of potentially problematic `NA` values.

This is where differentiating between empty and declared missing values proves
valuable. The declared missing values are neither problematic nor signal
potential problems in the data, given that once a reason is declared, it is
already known why a particular value is missing.

The genuinely problematic values are the empty `NA` values, and the custom
class method still allows identifying such values if they exist:

```{r}
mean(c(x2, NA))
mean(c(x2, NA), na.rm = TRUE)
```

Since all declared values are stored as regular NA values, the base function
`is.na()`{.R}, as well as all related functions such as `anyNA()`{.R} etc., are
unaware and can not differentiate between empty and declared missing values:

```{r}
is.na(c(x2, NA))
```

To overcome this situation, package **declared** complementary provides an
additional function to account for the difference:

```{r}
is.empty(c(x2, NA))
```

All missing values, empty and declared, play well with the NA oriented, base
functions such as `na.omit()`{.R} or `na.exclude()`{.R}:

```{r}
na.omit(x2)
```

It should be made obvious the excellent packages **haven** and **labelled** are
not inherently doing a bad thing: the very same result is obtained, just via a
different route. Package **declared** was created as an alternative to the
design philosophy of these packages, with a fundamental difference: instead of
interpreting existing values as missing, the missing values are interpreted as
existing.

It does so by storing an additional attribute containing the positions (indexes)
of the regular `NA` values in the object, which should be treated as
missing, and, even more so, to be interpreted as a particular missing response
category, as specified in the value labels attribute.

---
title: "b. Added value"
output:
  rmarkdown::html_vignette:
   highlight: "declared.theme"
   css: "declared.css"
vignette: >
  %\VignetteIndexEntry{b. Added value}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 80
---

```{r vignettesetup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(haven)
set.seed(12345)
```


```{r}
library(declared)

x2 <- declared(
  x = c(1:5, -91),
  labels = c("Missing" = -91),
  na_value = -91
)
```

The proposed method to declare missing values is unique in the R ecosystem.
Differentiating between empty and declared missing values opens the door to a
new set of challenges for which base R does not have built-in functionality.

For instance, the declared missing values can be compared against both the
original values and their labels:

```{r}
x2 == -91
x2 == "Missing"
```

Similar methods have been added to the primitive functions `"<"`{.R}, `">"`{.R}
and `"!="`{.R} etc., to allow a fully functional collection of subsetting
possibilities.

Combining on this type of vector creates an object of the same class:

```{r}
x2 <- c(x2, -91)
x2
```

Most functions are designed to be as similar as possible, for instance,
`labels()`{.R} to retrieve or add/change value labels:

```{r}
labels(x2) <- c("Does not know" = -92, "Not responded" = -91)
x2
```

The value `-92` is now properly labelled, and can further be declared missing.
Such declarations do not necessarily have to use the main function
`declared()`{.R}, due to the separate functions `missing_values()`{.R} and
`missing_range()`{.R}:

```{r}
missing_values(x2) <- c(-91, -92)
missing_range(x2) <- c(-91, -95)
x2
```

To ease the smooth inter-operation with packages **haven** and **labelled**, the
following functions are of interest: `undeclare()`{.R}, `as.haven()`{.R} and
`as.declared()`{.R}.

The function `undeclare()`{.R} replaces the NAs with their declared missing
values. The result is still an object of class `"declared"`{.R}, but all missing
values (and missing range) are stripped off the vector, and values are presented
as they have been collected. All other attributes of interest (variable and
value labels) are retained and printed accordingly. Activating the argument
`drop`{.R} eliminates all classes and attributes, returning a regular R object:

```{r undeclare}
undeclare(x2, drop = TRUE)
```

The function `as.haven()`{.R} coerces the resulting object to the class
`"haven_labelled_spss"`{.R}, and the function `as.declared()`{.R} reverses the
process:

```{r haven}
xh <- as.haven(x2)
xh

as.declared(xh)
```

The missing values are properly formatted, even inside the base data frame:

```{r dataframe}
dfm <- data.frame(x1 = letters[1:7], x2)
dfm
```

If users prefer a tibble instead of a data frame, the objects of class
"`declared"`{.R} are properly formatted in a similar way to those from package
**haven**:

```{r tibble}
tibble::as_tibble(dfm)
```

Special challenges are associated with sorting and ordering the declared
objects, where not all missing values are treated the same.

```{r sort}
x3 <- declared(
  x = c(1:5, -91, NA, -92, -91),
  na_value = c(-92, -91)
)
sort(x3, na.last = TRUE)
```

Sorting in decreasing order applies the same order to the missing values:

```{r}
sort(x3, na.last = TRUE, decreasing = TRUE)
```

This custom function benefits from an additional argument `empty.last`{.R}
(internally passed to the ordering function), to allow sorting within the
missing values:

```{r}
sort(x3, na.last = TRUE, decreasing = TRUE, empty.last = FALSE)
```

All types of variables (categorical and numerical) can have declared missing
values. There are situations when values are not missing randomly but with a
specific reason. In social research, respondents often can not or do not want to
provide an answer for a certain question, be it categories of opinions or pure
numerical answers like age, income, etc.

Base R has a clear distinction between numerical and categorical variables. For
the latter, R provides a special type of object called `factor`{.R}. The
following object simulates such a categorical variable, for instance
political orientation:

```{r}
x4 <- declared(
  x = c(1:3, -91),
  labels = c("Left" = 1, "Middle" = 2, "Right" = 3, "Apolitic" = -91),
  na_value = -91,
  label = "Respondent's political orientation"
)

x4
```

Such a variable could in principle be constructed directly as a factor:

```{r}
x5 <- factor(
  c("Left", "Middle", "Right", "Apolitic"),
  levels = c("Left", "Middle", "Right", "Apolitic")
)

x5
```

The base factors provide no possibility to assign specific values for the
specific categories, and most importantly, they do not differentiate between
valid values and (declared) missing values. To make the functionality consistent
with the base treatment of `NA` values, coercing to factors defaults to
dropping the declared missing values because of the default value of the
argument `drop_na`{.R}:

```{r}
as.factor(x4)

# essentially acting as:
as.factor(drop_na(x4))
```

Converting the declared missing values to factor levels is something for which
the function `undeclare()`{.R} proves useful, and switching between factors and 
declared objects, preserving the declared missing values, is now straightforward:

```{r}
as.factor(undeclare(x4))
```

This is almost identical, but differs from `x5` with respect to the level
orders. It happens because the missing value -91, coerced to a valid value,
becomes the first in the order of the categories while normally, a categorical
variable displays the valid values first, and the missing values last.

If the original order is important, the argument `drop_na` can be deactivated:

```{r}
as.factor(x4, drop_na = FALSE)
```



The reverse process is also possible, to convert / coerce factors to declared
labelled objects:


```{r}
as.declared(x5, na_values = 4)
```


When the declared objects are constructed as categorical variables to replace
the base factors, the function `as.character()`{.R} extracts the categories
in a similar way to factors, via a dedicated class method for declared objects:

```{r}
as.character(x4)
```

Similarly, the function `as.character()`{.R} drops the declared missing values
by default, something that can be prevented by either of the following two
commands:

```{r}
as.character(undeclare(x4))

as.character(x4, drop_na = FALSE)
```

The base factors store the categories using numeric values, which is also the
most often scenario for the declared objects. But this is not at all mandatory,
as the declared objects are also able to ingest character vectors:

```{r}
x6 <- declared(
  x = sample(
    c("a", "b", "c", "z"),
    20,
    replace = TRUE
  ),
  labels = c("Left" = "a", "Middle" = "b", "Right" = "c", "Apolitic" = "z"),
  na_values = "z"
)

x6
```

Either as a character, numeric or categorical, it is possible to declare and use
special types of missing values, employing this new object type of class
`"declared"`{.R}.

Factors and `haven`{.R} objects have default coercion methods, but not all types
of objects can be automatically coerced to this class. To meet this possibility,
the main functions `declared()`{.R}, `as.declared()`{.R}, `undeclare()`{.R} and
`drop_na()`{.R} are all generic, allowing full  flexibility for any other
packages to create custom (coercion) methods for different object classes, thus
facilitating and encouraging a widespread use.

---
title: "c. Weighting (declared) values"
output:
  rmarkdown::html_vignette:
   highlight: "declared.theme"
   css: "declared.css"
vignette: >
  %\VignetteIndexEntry{c. Weighting (declared) values}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 80
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


For the examples in this vignette, the following data frame is created:

```{r}
library(declared)

n <- 1234
set.seed(n)
dfm <- data.frame(
  Area = declared(
    sample(1:2, n, replace = TRUE, prob = c(0.45, 0.55)),
    labels = c("Rural" = 1, "Urban" = 2)
  ),
  Gender = declared(
    sample(1:2, n, replace = TRUE, prob = c(0.55, 0.45)),
    labels = c("Males" = 1, "Females" = 2)
  ),
  Opinion = declared(
    sample(c(1:5, NA, -91), n, replace = TRUE),
    labels = c(
      "Very bad" = 1, "Bad" = 2, "Neither" = 3,
      "Good" = 4, "Very good" = 5, "Don't know" = -91
    ),
    na_values = -91
  ),
  Age = sample(18:90, n, replace = TRUE),
  Children = sample(0:5, n, replace = TRUE)
)
```

One of the most interesting applications to make use of the declared missing
values are the tables of frequencies. The base function `table()`{.R} ignores
missing values by default, but they can be revealed by using the `useNA`{.R}
argument:

```{r}
table(dfm$Opinion, useNA = "ifany")
```

However, it does not differentiate between empty and declared missing values.
Since "Opinion" is the equivalent of a categorical variable, this can be
improved through a custom built coercion to the base `factor`{.R} class:

```{r}
table(as.factor(undeclare(dfm$Opinion)), useNA = "ifany")
```

The dedicated function `w_table()`{.R} does the same thing by automatically
recognizing objects of class `"declared"`{.R}, additionally printing more
detailed information:

```{r}
w_table(dfm$Opinion, values = TRUE)
```

The prefix `w_`{.R} from the function name stands for "weighted", this being
another example of functionality where the declared missing values play a
different role than the empty, base NA missing values.

It is important to differentiate between frequency weights, on one hand, and
other probability based, post-stratification weights on one other, the later
being thoroughly treated by the specialized package **survey**. The `w_`{.R}
family of functions are solely dealing with frequency weights, to allow
corrections in descriptive statistics, such as the tables of frequencies and
other similar descriptive measures for both categorical and numeric variables.

To exemplify, a frequency weights variable is constructed, to correct for the
distributions of gender by males and females, as well as the theoretical
distribution by residential areas differentiating between urban and rural
settlements.

```{r}
# Observed proportions
op <- with(dfm, proportions(table(Gender, Area)))

# Theoretical / population proportions:
# 53% Rural, and 50% Females
weights <- rep(c(0.53, 0.47), each = 2) * rep(0.5, 4) / op

dfm$fweight <- weights[
  match(10 * dfm$Area + dfm$Gender, c(11, 12, 21, 22))
]
```

The updated frequency table, this time using the frequency weights, can be
constructed by passing the weights to the argument `wt`{.R}:

```{r}
with(dfm, w_table(Opinion, wt = fweight, values = TRUE))
```

Except for the empty NA values, for which the weights cannot be applied, almost
all other frequencies (including the one for the declared missing value -91) are
now updated by applying the weights. This shows that, despite being interpreted
as "missing" values, the declared ones can and should also be weighted, with a
very useful result. Other versions of weighted frequencies do exist in R, but a
custom one was needed to identify (and weight) the declared missing values.

In the same spirit, many other similar functions are provided such as
`w_mean()`{.R}, `w_var()`{.R}, `w_sd()`{.R} etc., and the list will likely grow
in the future. They are similar to the base package counterparts, with a single
difference: the argument `na.rm` is activated by default, with or without
weighting. This is an informed decision about which users are alerted in the
functions' respective help pages.

The package **declared** was built with the specific intention to provide a
lightweight, zero dependency resource in the R ecosystem. It contains an already
extensive, robust and ready to use functionality that duly takes into account
the difference between empty and declared missing values.

It extends base R and opens up data analysis possibilities without precedent. By
providing generic classes for all its objects and functions, package
**declared** is easily extensible to any type of object, for both creation and
coercion to class `"declared"`{.R}.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/missing_range.R, R/missing_values.R
\name{missing_range}
\alias{missing_range}
\alias{missing_range<-}
\alias{missing_values}
\alias{missing_values<-}
\title{Get / Declare missing values}
\usage{
missing_range(x)

missing_range(x) <- value

missing_values(x)

missing_values(x) <- value
}
\arguments{
\item{x}{A vector.}

\item{value}{Any vector of values that should be declared as missing
(for \code{missing_values}) or a numeric vector of length two giving the
(inclusive) extents of the range of missing values (for \code{missing_range}).}
}
\value{
\code{missing_values()} will return a vector of one or more values.

\code{missing_range()} will return a numeric vector of length 2.
}
\description{
Functions to extract information about the declared missing values, or to
declare such values if they are present in the data.
}
\examples{
x <- declared(c(-2, 1:5, -1),
    labels = c(Good = 1, Bad = 5, DK = -1, NotApplicable = -2),
    na_values = c(-1, -2)
)
x

missing_values(x)

missing_range(x) <- c(-10, -7)

missing_range(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/valid_na_index.R
\name{valid_na_index}
\alias{valid_na_index}
\title{Validate the declared missing values index}
\usage{
valid_na_index(x)
}
\arguments{
\item{x}{An atomic vector, possibly containing an \code{na_index} attribute.}
}
\value{
A logical value. \code{TRUE} when every indexed value is a genuine \code{NA};
otherwise \code{FALSE}.
}
\description{
Check whether every position stored in the \code{na_index} attribute identifies a
genuine missing value in the vector. This can detect when external code has
changed a vector without refreshing its declared missing values metadata.
}
\details{
The check examines only the positions in \code{na_index} and stops at the first
invalid position. A missing or empty \code{na_index} attribute is valid.
}
\examples{
x <- declared(c(1, -1, 3), na_values = -1)

valid_na_index(x)

attr(x, "na_index") <- 1
valid_na_index(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/haven.R
\name{as.haven}
\alias{as.haven}
\title{Coerce to haven / labelled objects}
\usage{
as.haven(x, ...)
}
\arguments{
\item{x}{A declared labelled vector}

\item{...}{Other arguments used by various methods}
}
\value{
A labelled vector of class "haven_labelled_spss".
}
\description{
Convert declared labelled objects to haven labelled objects
}
\details{
This is a function that reverses the process of \code{as.declared()}, making
a round trip between \code{declared} and \code{haven_labelled_spss} classes.
}
\examples{

x <- declared(
    c(1:5, -1),
    labels = c(Good = 1, Bad = 5, DK = -1),
    na_values = -1
)

x

as.haven(x)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is.empty.R
\name{is.empty}
\alias{is.empty}
\alias{anyNAempty}
\title{Test the presence of empty (undeclared) missing values}
\usage{
is.empty(x)

anyNAempty(x)
}
\arguments{
\item{x}{A vector}
}
\value{
A logical vector.
}
\description{
Functions that indicate which elements are empty \code{NA} missing values, in
contrast to declared missing values.
}
\details{
All missing values, declared or undeclared, as stored as regular \code{NA}
values, therefore the base function \code{is_na()} does not differentiate
between them.

These functions are specifically adapted to objects of class \code{"declared"},
to return a truth value only for those elements that are completely missing
with no reason.
}
\examples{

x <- declared(
    c(1:2, -91),
    labels = c(Good = 1, Bad = 2, Missing = -91),
    na_values = -91
)

x

is.empty(x) # FALSE FALSE FALSE

anyNAempty(x) # FALSE

x <- c(x, NA)

is.empty(x) # FALSE FALSE FALSE  TRUE

anyNAempty(x) # TRUE

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/labels.R
\name{labels}
\alias{labels}
\alias{label}
\alias{label<-}
\alias{labels<-}
\title{Get / Declare value labels}
\usage{
label(x)

label(x, ...) <- value

labels(x) <- value
}
\arguments{
\item{x}{Any vector of values that should be declared as missing
(for \code{labels}) or a numeric vector of length two giving the (inclusive)
extents of the range of missing values (for \code{label}).}

\item{...}{Other arguments, for internal use.}

\item{value}{The variable label, or a list of (named) variable labels}
}
\value{
\code{labels()} will return a named vector.

\code{label()} will return a single character string.
}
\description{
Functions to extract information about the declared variable / value labels,
or to declare such values if they are present in the data.
}
\details{
The function \code{labels()} is a adaptation of the base function to the objects
of class \code{declared}. In addition to the regular arguments, it has two
additional (logical) arguments called \code{prefixed} (FALSE by default), to
retrieve the value labels prefixed with their values, and \code{print_as_df}
(TRUE by default) to print the result as a data frame.
}
\examples{
x <- declared(
    c(-2, 1:5, -1),
    labels = c("Good" = 1, "Bad" = 5, "DK" = -1),
    na_values = c(-1, -2),
    label = "Test variable"
)
x

labels(x)

labels(x, prefixed = TRUE)

labels(x) <- c("Good" = 1, "Bad" = 5, "DK" = -1, "Not applicable" = -2)

label(x)

label(x) <- "This is a proper label"

x

}
\seealso{
Other labelling functions:
\code{\link{drop_undeclare}},
\code{\link[=measurement]{measurement()}}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.declared.R, R/declared.R, R/is.declared.R
\name{as.declared}
\alias{as.declared}
\alias{declared}
\alias{is.declared}
\alias{anyNAdeclared}
\title{Labelled vectors with declared missing values}
\usage{
as.declared(x, ...)

declared(
  x,
  labels = NULL,
  na_values = NULL,
  na_range = NULL,
  label = NULL,
  measurement = NULL,
  decimals = NULL,
  llevels = FALSE,
  ...
)

is.declared(x)

anyNAdeclared(x)
}
\arguments{
\item{x}{A numeric vector to label, or a declared labelled vector
(for \code{undeclare})}

\item{...}{Other arguments used by various other methods}

\item{labels}{A named vector or \code{NULL}. The vector should be the same type
as \code{x}. Unlike factors, labels don't need to be exhaustive: only a fraction
of the values might be labelled}

\item{na_values}{A vector of values that should also be considered as missing}

\item{na_range}{A numeric vector of length two giving the (inclusive) extents
of the range. Use \code{-Inf} and \code{Inf} if you want the range to be
open ended}

\item{label}{A short, human-readable description of the vector}

\item{measurement}{Optional, user specified measurement level}

\item{decimals}{Optional, maximum number of decimal places used for display
only}

\item{llevels}{Logical, when \code{x} is a factor only use those levels that have
labels}
}
\value{
\code{declared()} and \code{as.declared()} return labelled vector of class
"declared". When applied to a data frame, \code{as.declared()} will return a
corresponding data frame with declared variables. \code{is.declared()} and
\code{anyNAdeclared()} return a logical value.
}
\description{
The labelled vectors are mainly used to analyse social science data,
and the missing values declaration is an important step in the analysis.
}
\details{
The \code{declared} objects are very similar to the \code{haven_labelled_spss} objects
from package \bold{haven}. It has exactly the same arguments, but it features
a fundamental difference in the treatment of (declared) missing values.

In package \bold{haven}, existing values are treated as if they were missing.
By contrast, in package \bold{declared} the NA values are treated as if they
were existing values.

This difference is fundamental and points to an inconsistency in package
\bold{haven}: while existing values can be identified as missing using the
function \code{is.na()}, they are in fact present in the vector and other
packages (most importantly the base ones) do not know these values should be
treated as missing.

Consequently, the existing values are interpreted as missing only by package
\bold{haven}. Statistical procedures will use those values as if they were
valid values.

Package \bold{declared} approaches the problem in exactly the opposite way:
instead of treating existing values as missing, it treats (certain) NA values
as existing. It does that by storing an attribute containing the indices of
those NA values which are to be treated as declared missing values, and it
refreshes this attribute each time the declared object is changed.

This is a trade off and has important implications when subsetting datasets:
all declared variables get this attribute refreshed, which consumes some time
depending on the number of variables in the data.

The generic function \code{as.declared()} attempts to coerce only the compatible
types of objects, namely \code{haven_labelled} and \code{factor}s. Dedicated class
methods can be written for any other type of object, and users are free to
write their own. To end of with a declared object, additional metadata is
needed such as value labels, which values should be treated as missing etc.

The measurement level is optional and, for the moment, purely aesthetic. It
might however be useful to (automatically) determine if a declared object is
suitable for a certain statistical analysis, for instance regression requires
quantitative variables, while some declared objects are certainly categorical
despite using numbers to denote categories.

It distinguishes between \code{"categorical"} and \code{"quantitative"} types of
variables, and additionally recognizes \code{"nominal"} and \code{"ordinal"} as
categorical, and similarly recognizes \code{"interval"}, \code{"ratio"},
\code{"discrete"} and \code{"continuous"} as quantitative.
}
\examples{

x <- declared(
    c(1:5, -1),
    labels = c(Good = 1, Bad = 5, DK = -1),
    na_values = -1
)

x

is.na(x)

x > 0

x == -1

# Values are actually placeholder for categories,
# so labels work as if they were factors:
x == "DK"


# when newly added values are already declared as missing,
# they are automatically coerced
c(x, 2, -1)

# switch NAs with their original values
undeclare(x)

as.character(x)

# Returning values instead of categories
as.character(x, values = TRUE)

anyNAdeclared(x) # contains declared missing values

anyNAdeclared(c(1:5, NA)) # no declared missing values
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/measurement.R
\name{measurement}
\alias{measurement}
\alias{measurement<-}
\title{Get / Set measurement levels for declared objects}
\usage{
measurement(x)

measurement(x) <- value
}
\arguments{
\item{x}{A declared vector.}

\item{value}{A single character string of measurement levels,
separated by commas.}
}
\value{
A character vector.
}
\description{
Functions to extract information about the measurement levels of a variable
(if already present), or to specify such measurement levels.
}
\details{
This function creates an attribute called \code{"measurement"} to a declared
object. This is an optional feature, at this point for purely aesthetic
reasons, but it might become useful in the future to (automatically)
determine if a declared object is suitable for a certain statistical
analysis, for instance regression requires quantitative variables, while some
declared objects are certainly categorical despite using numbers to denote
categories.

It distinguishes between \code{"categorical"} and \code{"quantitative"} types of
variables, and additionally recognizes \code{"nominal"} and \code{"ordinal"} as
categorical, and similarly recognizes \code{"interval"}, \code{"ratio"},
\code{"discrete"} and \code{"continuous"} as quantitative.

The words \code{"qualitative"} is treated as a synonym for \code{"categorical"},
and the words \code{"metric"} and \code{"numeric"} are treated as synonyms for
\code{"quantitative"}, respectively.
}
\examples{
x <- declared(
    c(-2, 1:5, -1),
    labels = c(Good = 1, Bad = 5, DK = -1),
    na_values = c(-1, -2),
    label = "Test variable"
)

x

measurement(x)

# automatically recognized as categorical
measurement(x) <- "ordinal"

measurement(x)

# the same with
measurement(x) <- "categorical, ordinal"

set.seed(1890)
x <- declared(
    sample(c(18:90, -91), 20, replace = TRUE),
    labels = c("No answer" = -91),
    na_values = -91,
    label = "Respondent's age"
)

# automatically recognized as quantitative
measurement(x) <- "discrete"

measurement(x)

# the same with
measurement(x) <- "metric, discrete"

}
\seealso{
Other labelling functions:
\code{\link{drop_undeclare}},
\code{\link{labels}}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wtable.R, R/wmean.R, R/wmedian.R, R/wmode.R,
%   R/wvar.R, R/wsd.R, R/wquantile.R, R/wIQR.R, R/wfivenum.R, R/wstandardize.R,
%   R/wsummary.R, R/wmeasures.R
\name{weighted}
\alias{weighted}
\alias{wtable}
\alias{wmean}
\alias{wmedian}
\alias{wmode}
\alias{wvar}
\alias{wsd}
\alias{wquantile}
\alias{wIQR}
\alias{wfivenum}
\alias{wstandardize}
\alias{wsummary}
\alias{wmeasures}
\title{Compute weighted summaries for declared objects}
\usage{
wtable(
  x,
  y = NULL,
  wt = NULL,
  values = TRUE,
  valid = TRUE,
  observed = TRUE,
  vlabel = FALSE
)

wmean(x, wt = NULL, trim = 0, na.rm = TRUE)

wmedian(x, wt = NULL, na.rm = TRUE, ...)

wmode(x, wt = NULL)

wvar(x, wt = NULL, method = NULL, na.rm = TRUE)

wsd(x, wt = NULL, method = NULL, na.rm = TRUE)

wquantile(x, wt = NULL, probs = seq(0, 1, 0.25), na.rm = TRUE, ...)

wIQR(x, wt = NULL, na.rm = FALSE, ...)

wfivenum(x, wt = NULL, na.rm = FALSE)

wstandardize(x, wt = NULL, na.rm = TRUE)

wsummary(x, wt = NULL, ...)

wmeasures(
  x,
  what = c("n", "mode", "mean", "median", "range", "var", "sd"),
  wt = NULL,
  na.rm = TRUE,
  ...
)
}
\arguments{
\item{x}{A numeric vector for summaries, or declared / factor for frequency
tables}

\item{y}{An optional variable, to create crosstabs; must have the same length
as x}

\item{wt}{A numeric vector of frequency weights}

\item{values}{Logical, print the values in the table rows}

\item{valid}{Logical, print separate percent distribution for valid values,
if any missing values are present; for cross tables, use valid values only}

\item{observed}{Logical, print the observed categories only}

\item{vlabel}{Logical, print the variable label, if existing}

\item{trim}{A fraction (0 to 0.5) of observations to be trimmed from each end
of x before the mean is computed. Values of trim outside that range are
taken as the nearest endpoint.}

\item{na.rm}{Logical, should the empty missing values be removed?}

\item{...}{Further arguments passed to or from other methods.}

\item{method}{Character, specifying how the result is scaled, see 'Details'
below.}

\item{probs}{Numeric vector of probabilities with values in [0,1]}

\item{what}{A character vector or a named list of functions, identifying
what to compute. Available character values are \code{"n"}, \code{"NA"}, \code{"mode"},
\code{"mean"}, \code{"median"}, \code{"range"}, \code{"var"}, \code{"sd"}, \code{"iqr"}, \code{"min"}, and
\code{"max"}.}
}
\value{
A vector of (weighted) values.
}
\description{
Functions to compute weighted tables or summaries, based on a vector of
frequency weights. These are reimplementations of various existing functions,
adapted to objects of class \code{"declared"} (see Details below)
}
\details{
Weighted summaries

A frequency table is usually performed for a categorical variable, displaying
the frequencies of the respective categories. Note that general variables
containing text are not necessarily factors, despite having a small number of
characters.

A general table of frequencies, using the base function \code{table()}, ignores
the defined missing values (which are all stored as NAs). The
reimplementation of this function in \code{wtable()} takes care of this detail,
and presents frequencies for each separately defined missing values. Similar
reimplementations for the other functions have the same underlying objective.

It is also possible to perform a frequency table for numerical variables, if
the number of values is limited (an arbitrary and debatable upper limit of 15
is used). An example of such variable can be the number of children, where
each value can be interpreted as a class, containing a single value (for
instance 0 meaning the category of people with no children).

Objects of class \code{declared} are not pure categorical variables (R factors)
but they are nevertheless interpreted as if they were factors, to allow
producing frequency tables. Given the high similarity with package
\strong{\code{haven}}, objects of class \code{haven_labelled_spss} are automatically
coerced to objects of class \code{declared} and treated accordingly.

The argument \code{values} makes sense only when the input is of family class
\code{declared}, otherwise for regular (base R) factors the values are
just a sequence of numbers.

The later introduced argument \code{observed} is useful in situations when a
variable has a very large number of potential values, and a smaller subset of
actually observed ones. As an example, the variable \dQuote{Occupation} has
hundreds of possible values in the ISCO08 codelist, and not all of them might
be actually observed. When activated, this argument restricts the printed
frequency table to the subset of observed values only.

The argument \code{wt} refers only to frequency weights. Users should be
aware of the differences between frequency weights, analytic weights,
probability weights, design weights, post-stratification weights etc. For
purposes of inferential testing, Thomas Lumley's package \strong{\code{survey}}
should be employed.

If no frequency weights are provided, the result is identical to the
corresponding base functions.

For all functions, the argument \code{na.rm} refers to the empty missing values
and its default is set to TRUE. The declared missing values are automatically
eliminated from the summary statistics, even if this argument is deactivated.

The function \code{wmode()} returns the weighted mode of a variable. Unlike the
other functions where the prefix \code{w} signals a weighted version of the
base function with the same name, this has nothing to do with the base
function \code{mode()} which refers to the storage mode or type of an R object.

For \code{wvar()} and \code{wsd()}, the argument \code{method} can be one of \code{"unbiased"}
or \code{"ML"}. The former produces an unbiased estimate using Bessel's
correction, while the latter produces the maximum likelihood estimate for a
Gaussian distribution.

The function \code{wquantile()} extensively borrowed ideas from packages
\strong{\code{stats}} and \strong{\code{Hmisc}}, to ensure a constant interpolation that would
produce the same quantiles if no weights are provided or if all weights are
equal to 1.

Other arguments can be passed to the stats function \code{quantile()} via the
three dots \code{...} argument, and their extensive explanation is found in the
corresponding stats function's help page.

For \code{wmeasures()}, the \code{"n"} measure counts valid values, while \code{"NA"} counts
all missing values, both declared and empty.

When \code{wmeasures()} is evaluated in the context of a data frame, such as with
\code{admisc::using()}, \code{.} can be used as a placeholder for all variables in the
current dataset.
}
\examples{
set.seed(215)

# a pure categorical variable
x <- factor(sample(letters[1:5], 215, replace = TRUE))
wtable(x)


# simulate number of children
x <- sample(0:4, 215, replace = TRUE)
wtable(x)

# simulate a Likert type response scale from 1 to 7
values <- sample(c(1:7, -91), 215, replace = TRUE)
x <- declared(values, labels = c("Good" = 1, "Bad" = 7))
wtable(x)


# Defining missing values
missing_values(x) <- -91
wtable(x)


# Defined missing values with labels
values <- sample(c(1:7, -91, NA), 215, replace = TRUE)
x <- declared(
    values,
    labels = c("Good" = 1, "Bad" = 7, "Don't know" = -91),
    na_values = -91
)

wtable(x)

# Including the values in the table of frequencies
wtable(x, values = TRUE)


# An example involving multiple variables
DF <- data.frame(
    Area = declared(
        sample(1:2, 215, replace = TRUE, prob = c(0.45, 0.55)),
        labels = c(Rural = 1, Urban = 2)
    ),
    Gender = declared(
        sample(1:2, 215, replace = TRUE, prob = c(0.55, 0.45)),
        labels = c(Males = 1, Females = 2)
    ),
    Age = sample(18:90, 215, replace = TRUE),
    Children = sample(0:5, 215, replace = TRUE)
)

wtable(DF$Gender)

wsd(DF$Age)

wmeasures(DF[c("Age", "Children")], what = c("n", "mean", "sd"))


# Weighting: observed proportions
op <- proportions(with(DF, table(Gender, Area)))

# Theoretical proportions: 53\% Rural, and 50.2\% Females
tp <- rep(c(0.53, 0.47), each = 2) * rep(c(0.498, 0.502), 2)

# Corrections by strata
fweights <- tp / op

DF$fweight <- fweights[match(10 * DF$Area + DF$Gender, c(11, 12, 21, 22))]

with(DF, wtable(Gender, wt = fweight))

with(DF, wmean(Age, wt = fweight))

with(DF, wquantile(Age, wt = fweight))
}
\author{
Adrian Dusa
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/declared_package.R
\name{declared_package}
\alias{declared_package}
\title{Functions for Declared Missing Values}
\description{
A set of functions to declare labels and missing values, coupled
with associated functions to create (weighted) tables of frequencies and
various other summary measures.
Some of the base functions are rewritten to make use of the specific
information about the missing values, most importantly to distinguish between
empty and declared missing values.
Many functions have a similar functionality with the corresponding ones from
packages "haven" and "labelled". The aim is to ensure as much compatibility
as possible with these packages, while offering an alternative in the objects
of class "declared".
}
\details{
\tabular{ll}{
Package: \tab declared\cr
Type: \tab Package\cr
    Version: \tab 0.27\cr
    Date:    \tab 2026-08-22\cr
License: \tab GPL-v3\cr
}
}
\author{
Adrian Dusa

Maintainer: Adrian Dusa (dusa.adrian@unibuc.ro)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_undeclare.R
\name{drop_undeclare}
\alias{drop_undeclare}
\alias{undeclare}
\alias{drop_na}
\title{Drop information / undeclare labelled objects}
\usage{
undeclare(x, drop = FALSE, ...)

drop_na(x, drop_labels = TRUE)
}
\arguments{
\item{x}{A labelled object with declared missing values}

\item{drop}{Logical, drop all attributes}

\item{...}{Other internal arguments}

\item{drop_labels}{Logical, drop the labels for the declared missing values}
}
\value{
A declared labelled object.
}
\description{
A function to obtain a version of the object with all information about
declared missing values, dropped
}
\details{
The function \code{undeclare()} replaces the NA entries into their original
numeric values, and drops all attributes related to missing values:
\code{na_values}, \code{na_range} and \code{na_index}, and it preserves the labels referring
to the missing values.

The result can be a regular vector (dropping all attributes, including the
class "declared") by activating the argument \code{drop}.

Function \code{drop_na()} transforms the declared missing values in regular empty
NAs, and the labels referring to the missing values are deleted by default.

Function \code{drop()} deletes all attributes.
}
\examples{
x <- declared(
    c(-2, 1:5, -1),
    labels = c("Good" = 1, "Bad" = 5, "DK" = -1),
    na_values = c(-1, -2),
    label = "Test variable"
)

x

undeclare(x)

drop_na(x)

drop(x)

undeclare(x, drop = TRUE)

# similar to:
drop(undeclare(x))
}
\seealso{
Other labelling functions:
\code{\link{labels}},
\code{\link[=measurement]{measurement()}}
}
\concept{labelling functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/declared.R, R/internals.R, R/wIQR.R,
%   R/wfivenum.R, R/wmean.R, R/wmedian.R, R/wmode.R, R/wquantile.R, R/wsd.R,
%   R/wstandardize.R, R/wsummary.R, R/wtable.R, R/wvar.R
\name{direct_declared}
\alias{direct_declared}
\alias{declared_internal}
\alias{format_declared}
\alias{order_declared}
\alias{value_labels}
\alias{variable_label}
\alias{names_values}
\alias{makeTag_}
\alias{hasTag_}
\alias{getTag_}
\alias{anyTagged_}
\alias{w_IQR}
\alias{w_fivenum}
\alias{w_mean}
\alias{w_median}
\alias{w_mode}
\alias{w_quantile}
\alias{w_sd}
\alias{w_standardize}
\alias{w_summary}
\alias{w_table}
\alias{w_var}
\title{declared internal functions}
\usage{
direct_declared(
  x,
  na_index = NULL,
  na_values = NULL,
  na_range = NULL,
  labels = NULL,
  label = NULL,
  measurement = NULL,
  decimals = NULL,
  date = inherits(x, "Date")
)

format_declared(x, digits = getOption("digits"))

order_declared(
  x,
  na.last = NA,
  decreasing = FALSE,
  method = c("auto", "shell", "radix"),
  empty.last = na.last,
  ...
)

value_labels(...)

variable_label(...)

names_values(x, drop_na = FALSE, observed = TRUE)

makeTag_(...)

hasTag_(x, tag = NULL)

getTag_(x)

anyTagged_(x)

w_IQR(...)

w_fivenum(...)

w_mean(...)

w_median(...)

w_mode(...)

w_quantile(...)

w_sd(...)

w_standardize(...)

w_summary(...)

w_table(...)

w_var(...)
}
\description{
Functions to be used internally, only by developers and
contributors.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/declared_package.R
\docType{package}
\name{declared-package}
\alias{declared-package}
\title{declared: Functions for Declared Missing Values}
\description{
A zero dependency package containing functions to declare labels and missing values, coupled with associated functions to create (weighted) tables of frequencies and various other summary measures. Some of the base functions have been rewritten to make use of the specific information about the missing values, most importantly to distinguish between empty and declared NA values. Some functions have similar functionality with the corresponding ones from packages "haven" and "labelled". The aim is to ensure as much compatibility as possible with these packages, while offering an alternative in the objects of class "declared".
}
\seealso{
Useful links:
\itemize{
  \item \url{https://dusadrian.github.io/declared/}
  \item \url{https://github.com/dusadrian/declared}
  \item Report bugs at \url{https://github.com/dusadrian/declared/issues}
}

}
\author{
\strong{Maintainer}: Adrian Dusa \email{dusa.adrian@unibuc.ro} (\href{https://orcid.org/0000-0002-3525-9253}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Adrian Dusa \email{dusa.adrian@unibuc.ro} (\href{https://orcid.org/0000-0002-3525-9253}{ORCID}) [copyright holder]
}

}
\keyword{internal}
